static char rcsid[] = "$Id: bclient.c,v 1.11 1995/05/31 19:56:55 hardy Exp $";
/* 
 *  File:         bclient.c
 *  Description:  icp (binary) client.
 *  Author:       John Noll, USC
 *                Dean Povey, University of Queensland
 *  Created:      Wed Mar 30 18:10:25 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include "stack.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "proto.h"
#include "icp_lib.h"
#include "comm.h"
#include "fdstat.h"

#ifdef SUPPORT_ICPV1_4
#include "proto_v1.4.h"
#include "icpv1.4_lib.h"
#endif

void print_hit(char *buf);

int theUdpConnection = -1;
Stack *free_pkt_list;

int opt_v=0;

void usage(progname) 
     char *progname;
{
  printf("****** bclient: usage: %s [-sqra][-c hostname][-ut port] url\n", 
	progname);
}

/* Globals */
int do_mail=0;
int do_reuse=0;
time_t cached_starttime = 0L;
int debug_level = 0;
unsigned long nconn = 0L;
time_t next_cleaning = 0L;
char *tmp_error_buf = NULL;
char *config_file;


main(argc,argv) 
     int argc;
     char **argv;
{
  int c, udp, conn, len;
  int  op = ICP_OP_QUERY;
  int options = 0;
  char url[8192], buf[8192];
  char *hostname;
  u_num32 authenticator[ICP_AUTH_SIZE];
  int ttl=128; /* Default ttl */
  char *pointer=NULL;
  time_t timestamp = time(NULL);
  int size = -1;
  int UDPPort = 3130;
  int TCPPort = 3129;
  int useICPv1_4=0;
  struct in_addr hostid;
  extern char *optarg;
  extern int optind;


  tmp_error_buf = xcalloc(1, 8192);
  comm_init();
  fdstat_init(PREOPEN_FD);

  config_file = xcalloc(1, 1024);
  strncpy(config_file, "/work/dean/harvest/lib/cache/cached.conf", 1024);

  hostname = (char*) malloc(256);
  sprintf(hostname, comm_hostname());

  if (argc < 2) {
    usage(argv[0]);
    exit(1);
  } else if (argc >= 2) {
    strcpy(url, argv[argc-1]);

    while ((c = getopt(argc, argv, "c:u:t:oqradhP:T:S:L?")) != -1)
      switch (c) {
		case 'q':
			op  = ICP_OP_QUERY;
			break;
		case 'c':
			hostname = optarg;
			break;
		case 'u':
			UDPPort = atoi(optarg);
			break;
		case 't':
			TCPPort = atoi(optarg);
			break;
		case 'o':
#ifndef SUPPORT_ICPV1_4	
			fprintf(stderr, "ICP v1.4 not supported\n");
			exit(1);
#else
			useICPv1_4=1;
#endif
			break;
		case 'r':
			options = ICP_OPT_RECURSE;
			op = ICP_OP_QUERY;
			break;
		case 'a':
			op = ICP_OP_ADD;
			break;
		case 'P':
			pointer = strdup(optarg);
			break;
		case 'd':
			op = ICP_OP_DELETE;
			break;
		case '?':
		case 'h':
		default:
	usage(argv[0]);
	exit(1);
	break;
      }
  }

  /* Send request. */
  switch (op) {
  case ICP_OP_QUERY:   {
    struct sockaddr_in from;
    int from_len, type_len = sizeof(int), stype = 0;
    int response_len;

    if((udp = comm_open(COMM_BLOCKING | COMM_DGRAM, 0, 0)) < 0) {
      perror("****** bclient: error: creating udp socket\n");
      exit(1);
    }

    if (getsockopt(udp, SOL_SOCKET, SO_TYPE, (char *)&stype, &type_len)) {
      perror("****** bclient: error: gettting socket type");
    }

#ifdef SUPPORT_ICPV1_4
	if (useICPv1_4)
    	icpv1_4_query(udp, 0, authenticator, hostid, url, hostname, UDPPort);
	else
#endif
    	icp_query(udp, 0, authenticator, hostid, url, hostname, UDPPort, options);

	response_len = comm_udp_recv(udp, buf, 4096, &from, &from_len);
	return (print_result(buf));
  }
    break;

  case ICP_OP_ADD: {
    struct sockaddr_in from;
    int from_len, type_len = sizeof(int), stype = 0;
    int response_len;

	if (!pointer)
	{
		printf("Must specify a pointer\n");
		exit(1);
	}

    if((udp = comm_open(COMM_BLOCKING | COMM_DGRAM, 0, 0)) < 0) {
      perror("****** bclient: error: creating udp socket\n");
      exit(1);
    }

    if (getsockopt(udp, SOL_SOCKET, SO_TYPE, (char *)&stype, &type_len)) {
      perror("****** bclient: error: gettting socket type");
    }
    printf("****** bclient: socket type: %s (%d)\n", 
	   stype == SOCK_STREAM ? "Stream" : "Dgram", stype);

#ifdef SUPPORT_ICPV1_4
	if (useICPv1_4)
	{
		printf("Add not valid in ICPv1.4\n");
		exit(1);
	}
	else
#endif
	/* Advertise a pointer addition */
    icp_add(udp, 0, authenticator, hostid, url, pointer, hostname, UDPPort, 
			timestamp, ttl, size);

	}
	break;

  case ICP_OP_DELETE: {
    struct sockaddr_in from;
    int from_len, type_len = sizeof(int), stype = 0;
    int response_len;

    if((udp = comm_open(COMM_BLOCKING | COMM_DGRAM, 0, 0)) < 0) {
      perror("****** bclient: error: creating udp socket\n");
      exit(1);
    }

    if (getsockopt(udp, SOL_SOCKET, SO_TYPE, (char *)&stype, &type_len)) {
      perror("****** bclient: error: gettting socket type");
    }
#ifdef SUPPORT_ICPV1_4
	if (useICPv1_4)
	{
		printf("Delete not valid in ICPv1.4\n");
		exit(1);
	}
	else
#endif
	/* Advertise a delete */
    icp_delete(udp, 0, authenticator, hostid, url, pointer, hostname, UDPPort);

	}
	break;
	
  default:
    printf("****** bclient: unknown operation: %d\n");
    exit(1);
    break;
  }
}

int print_result(char *buf)
{
	static char *opcodes[]=
	{
		"INVALID",
		"QUERY",
		"HIT",
		"MISS",
		"ERR",
		"SEND",
		"SENDA",
		"DATABEG",
		"DATA",
		"DATAEND",
		"SECHO",
		"ADD",
		"DELETE"
	};

	icp_common_t *headerp = (icp_common_t *)buf;

#ifdef SUPPORT_ICPV1_4
	icpv1_4_common_t *v1_4_headerp = (icpv1_4_common_t *)buf;
#endif

	if (headerp->version == 3)
	{
		if (opt_v)
		{
		printf("Opcode        : %s\n", opcodes[headerp->opcode]);
		printf("Version       : %d\n", headerp->version);
		printf("Length        : %d\n", headerp->length);
		printf("Request Number: %d\n", headerp->reqnum);
		printf("Host ID (hex) : %08x:%08x:%08x:%08x\n",  
			headerp->padding[0],
			headerp->padding[1],
			headerp->padding[2],
			headerp->shostid);
		printf("Nr of options : %d\n", headerp->nr_options);
		}
		switch(headerp->opcode)
		{
			case ICP_OP_HIT:
				puts("HIT");
				if (opt_v)
					print_hit(buf+headerp->offset);
				return 1;
				break;
			default:
				puts("MISS");
				return 0;
		}
		return;
	}

#ifdef  SUPPORT_ICPV1_4
	else if (v1_4_headerp->version == 2)
	{
		printf("Opcode           : %s\n", opcodes[v1_4_headerp->opcode]);
		printf("Version          : %d\n", v1_4_headerp->version);
		printf("Length           : %d\n", v1_4_headerp->length);
		printf("Request Number   : %d\n", v1_4_headerp->reqnum);
		printf("Host ID (hex)    : %08x\n", v1_4_headerp->shostid);
		return;
	}
#endif

	else
	{
		puts("Version not supported");
		return;
	}
}

void print_hit (char *buf)
{
	char tmp[1024];
	unsigned short len;
	unsigned short nr_ptrs;
	time_t timestamp=0;
	int i;
	int length=sizeof(icp_common_t);

	/* Get Number of pointers */ 
	memcpy(&nr_ptrs, buf,  sizeof(unsigned short));
	printf("Number of pointers: %d\n", nr_ptrs);
	buf += sizeof(unsigned short);
	length += sizeof(unsigned short);

	/* Get size of source URL */
	memcpy(&len, buf,  sizeof(unsigned short));
	printf("Source URL length : %d\n", len);
	buf += sizeof(unsigned short);
	length += sizeof(unsigned short);

	/* Get Source URL */
	strncpy(tmp, buf, len);
	tmp[len]=0; /* Add terminator */
	printf("Source URL        : %s\n", tmp);
	buf += len;
	length += len;

	/* Get pointers */
	for (i=1; i<=nr_ptrs; i++)
	{
		/* Get timestamp */
		memcpy(&timestamp, buf,  sizeof(time_t));
		buf += sizeof(time_t);
		length += sizeof(time_t);

		/* Get the size of the URL */
		memcpy(&len, buf,  sizeof(unsigned short));
		buf += sizeof(unsigned short);
		length += sizeof(unsigned short);

		/* Get the URL */
		strncpy(tmp, buf, len);
		tmp[len]=0;
		printf("Pointer URL %d     : %s [ttl = %d]\n", i, tmp, timestamp);
		buf += len;
		length += len;
	}

	return;
}
