static char rcsid[] = "$Id: cache_cf.c,v 1.32 1995/08/21 20:55:05 hardy Exp $";
/*
 *  File:         cache_cf.c
 *  Description:  cache initialization functions
 *  Author:       Chuck Neerdaaels, USC
 *  Created:      Mon May 23 1994
 *  Language:     C
 *
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <malloc.h>
#include <pwd.h>
#include <grp.h>

#include "config.h"		/* goes first */
#include "debug.h"
#include "comm.h"
#include "cache_cf.h"
#include "store.h"
#include "ttl.h"

static int max_cache_size = 16 * 1024 * 1024;	/* 16 MB */
static int cache_mem_high_water_mark = 90;	/* 90% */
static int cache_mem_low_water_mark = 60;	/* 60% */
static int cache_swap_high_water_mark = 90;	/* 90% */
static int cache_swap_low_water_mark = 60;	/* 60% */
static int max_cache_swap_size = 100 << 10;	/* 100 MB (100*1024 kbytes) */
static int gopher_max_obj_size = 2 * 1024 * 1024;	/* 2 MB */
static int http_max_obj_size = 4 * 1024 * 1024;		/* 4 MB */
static int ftp_max_obj_size = 4 * 1024 * 1024;	/* 4 MB */
#ifdef USE_WAIS_RELAY
static int wais_max_obj_size = 4 * 1024 * 1024;		/* 4 MB */
#endif
static int http_ttl = 7 * 24 * 60 * 60;		/* 1 week */
static int gopher_ttl = 7 * 24 * 60 * 60;	/* 1 week */
static int ftp_ttl = 7 * 24 * 60 * 60;	/* 1 week */
static int negative_ttl = 5 * 60;	/* 5 min */
static int read_timeout = 5 * 60;	/* 5 min */
static int default_lifetime = 15 * 60;	/* 15 min */
static int default_age_max = 3600 * 24 * 30;	/* 30 days */
static int clean_rate = 30 * 60;	/* 30 min */
static int mail_trace = 1;	/* on */
static int dns_children = 3;	/* 3 dnsserver processes */
static int max_dns_children = 32;	/* 32 dnsserver processes */
static char w_space[] = " \t\n";

int zap_disk_store = 0;		/* off, try to rebuild from disk */
char *cache_log_file = NULL;
char *cache_access_log_file = NULL;
char *cache_hierarchy_log_file = NULL;
char *cache_dir = NULL;
char *cache_mgr = NULL;
char *ftp_program = NULL;
char *dns_program = NULL;
int cache_hierarchy_log_flag = 0;
int cachelog_flag = -1;
int httpd_accel_mode = 0;
int httpd_accel_port = 0;
int httpd_accel_with_proxy = 0;
int source_ping = 1;
char *httpd_accel_host = NULL;
char *httpd_accel_prefix = NULL;
char *effective_user = NULL;
char *effective_group = NULL;
int emulate_httpd_log = 1;	/* enabled, do emulation-type logging */

#ifdef USE_WAIS_RELAY
char *wais_relay_host = NULL;
int wais_relay_port;
#endif

#ifdef APPEND_DOMAIN
char *append_domain = NULL;
int append_domain_len;
#endif				/* APPEND_DOMAIN */

stoplist *http_stoplist = NULL;
stoplist *gopher_stoplist = NULL;
stoplist *ftp_stoplist = NULL;
ip_acl *ip_access_control_list = NULL;
ip_acl *manager_ip_access_control_list = NULL;
int quick_abort = 0;		/* off by default */

void self_destruct(in_string)
char *in_string;
{
    char fatal_str[4096];

    sprintf(fatal_str, "Bungled cached.conf: %s\n", in_string);
    fatal(fatal_str);
}

int ip_acl_match(c1, c2, c3, c4, a1, a2, a3, a4)
int c1;
int c2;
int c3;
int c4;
int a1;
int a2;
int a3;
int a4;
{
    if (!((a1 == 0) || (a1 == c1)))
	return 0;
    if (!((a2 == 0) || (a2 == c2)))
	return 0;
    if (!((a3 == 0) || (a3 == c3)))
	return 0;
    if (!((a4 == 0) || (a4 == c4)))
	return 0;

    return 1;
}


ip_access_type
ip_access_check(address, list)
struct in_addr address;
ip_acl *list;
{
    int c1, c2, c3, c4;
    ip_acl *p;
    unsigned int naddr = 0;	/* network byte-order IP addr */

    naddr = htonl(address.s_addr);
    c1 = ((int) naddr & 0xff000000) >> 24;
    c2 = ((int) naddr & 0x00ff0000) >> 16;
    c3 = ((int) naddr & 0x0000ff00) >> 8;
    c4 = ((int) naddr & 0x000000ff);

    debug(10, "ip_access_check: Using %d.%d.%d.%d\n", c1, c2, c3, c4);

    if ((c1 == 127) && (c2 == 0) && (c3 == 0) && (c4 == 1))
	return IP_ALLOW;	/* always allow localhost */

    p = list;
    while (p) {
	debug(10, "ip_access_check: %d.%d.%d.%d vs %d.%d.%d.%d\n",
	      c1, c2, c3, c4, p->a1, p->a2, p->a3, p->a4);
	if (ip_acl_match(c1, c2, c3, c4, p->a1, p->a2, p->a3, p->a4))
	    return p->access;
	p = p->next;
    }

    return IP_ALLOW;
}

void addToIPACL(list, ip_str, access)
ip_acl **list;
char *ip_str;
ip_access_type access;
{
    ip_acl *p, *q;
    int a1, a2, a3, a4;

    if (!ip_str) {
	return;
    }
    if (!(*list)) {
	/* empty list */
	*list = (ip_acl *) xcalloc(1, sizeof(ip_acl));
	(*list)->next = NULL;
	q = *list;
    } else {
	p = *list;
	while (p->next)
	    p = p->next;
	q = (ip_acl *) xcalloc(1, sizeof(ip_acl));
	q->next = NULL;
	p->next = q;
    }

    /* decode ip address */
    if (strstr(ip_str, "all") || strstr(ip_str, "ALL") ||
	strstr(ip_str, "All")) {
	a1 = a2 = a3 = a4 = 0;
    } else {
	int t;
	a1 = a2 = a3 = a4 = 0;
	t = sscanf(ip_str, "%u.%u.%u.%u", &a1, &a2, &a3, &a4);
    }

    q->access = access;
    q->a1 = a1;
    q->a2 = a2;
    q->a3 = a3;
    q->a4 = a4;

}


void addToStopList(list, key)
stoplist **list;
char *key;
{
    stoplist *p, *q;

    if (!(*list)) {
	/* empty list */
	*list = (stoplist *) xcalloc(1, sizeof(stoplist));
	(*list)->key = strdup(key);
	(*list)->next = NULL;
    } else {
	p = *list;
	while (p->next)
	    p = p->next;
	q = (stoplist *) xcalloc(1, sizeof(stoplist));
	q->key = strdup(key);
	q->next = NULL;
	p->next = q;
    }
}

/* Use this #define in all the parse*() functions.  Assumes 
   ** char *token and char *line_in are defined
 */

#define GetInteger(var) \
	token = strtok(NULL, w_space); \
	if( token == (char *) NULL) \
		self_destruct(line_in); \
	if (sscanf(token, "%d", &var) != 1) \
		self_destruct(line_in);


void parseHostLine(line_in)
char *line_in;
{
    char *type, *hostname;
    int ascii_port, tcp_port, udp_port;
    char *token;

    /* Parse a cache_host line */
    if (!(hostname = strtok(NULL, w_space)))
	self_destruct(line_in);
    if (!(type = strtok(NULL, w_space)))
	self_destruct(line_in);

    GetInteger(ascii_port);
    GetInteger(udp_port);
    GetInteger(tcp_port);

    neighbors_cf_add(hostname, type, ascii_port, udp_port, tcp_port);
}

void parseMailTraceLine(line_in)
char *line_in;
{
    char *onoff;

    if (!(onoff = strtok(NULL, w_space)))
	self_destruct(line_in);

    /* toggle mail trace depending on on/off value */
    if (getMailTrace() == 0 && !strcasecmp(onoff, "on"))
	mail_trace = 1;
    else if (getMailTrace() != 0 && !strcasecmp(onoff, "off"))
	mail_trace = 0;
}


void parseSourcePingLine(line_in)
char *line_in;
{
    char *srcping;

    srcping = strtok(NULL, w_space);
    if (srcping == (char *) NULL)
	self_destruct(line_in);

    /* set source_ping, default is on. */
    if (!strcasecmp(srcping, "on"))
	source_ping = 1;
    else if (!strcasecmp(srcping, "off"))
	source_ping = 0;
    else
	source_ping = 1;
}


void parseQuickAbortLine(line_in)
char *line_in;
{
    char *abort;

    abort = strtok(NULL, w_space);
    if (abort == (char *) NULL)
	self_destruct(line_in);

    if (!strcasecmp(abort, "on") || !strcasecmp(abort, "quick"))
	quick_abort = 1;
    else if (!strcmp(abort, "off") || !strcasecmp(abort, "normal"))
	quick_abort = 0;
    else
	quick_abort = 0;

}


void parseHttpLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(http_max_obj_size);
    GetInteger(http_ttl);
    http_ttl *= 60;
    http_max_obj_size <<= 20;	/* 1<<20 == 1024*1024 */
}

void parseMemLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(max_cache_size);
    max_cache_size <<= 20;
}

void parseMemHighLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(cache_mem_high_water_mark);
}

void parseMemLowLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(cache_mem_low_water_mark);
}


void parseSwapLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(max_cache_swap_size);
    max_cache_swap_size <<= 10;
}

void parseSwapHighLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(cache_swap_high_water_mark);
}

void parseSwapLowLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(cache_swap_low_water_mark);
}


void parseGopherLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(gopher_max_obj_size);
    GetInteger(gopher_ttl);
    gopher_ttl *= 60;
    gopher_max_obj_size <<= 20;
}

void parseFtpLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(ftp_max_obj_size);
    GetInteger(ftp_ttl);
    ftp_ttl *= 60;
    ftp_max_obj_size <<= 20;
}

void parseTTLPattern(line_in)
char *line_in;
{
    char *token;
    char *pattern;
    time_t abs_ttl = 0;
    int pct_age = 0;
    time_t age_max = default_age_max;
    int i;

    token = strtok(NULL, w_space);	/* token: regex pattern */
    if (token == (char *) NULL)
	self_destruct(line_in);
    pattern = strdup(token);

    GetInteger(i);		/* token: abs_ttl */
    abs_ttl = (time_t) (i * 60);	/* convert minutes to seconds */

    token = strtok(NULL, w_space);	/* token: pct_age */
    if (token != (char *) NULL) {	/* pct_age is optional */
	if (sscanf(token, "%d", &pct_age) != 1)
	    self_destruct(line_in);
    }
    token = strtok(NULL, w_space);	/* token: age_max */
    if (token != (char *) NULL) {	/* age_max is optional */
	if (sscanf(token, "%d", &i) != 1)
	    self_destruct(line_in);
	age_max = (time_t) (i * 60);	/* convert minutes to seconds */
    }
    ttlAddToList(pattern, abs_ttl, pct_age, age_max);

    safe_free(pattern);
}

void parseNegativeLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(negative_ttl);
    negative_ttl *= 60;
}

void parseReadTimeoutLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(read_timeout);
    read_timeout *= 60;
}

void parseLifetimeLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(default_lifetime);
    default_lifetime *= 60;
}

void parseCleanRateLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(clean_rate);
    clean_rate *= 60;
}

void parseDnsChildrenLine(line_in)
char *line_in;
{
    char *token;
    GetInteger(dns_children);
}

void parseMgrLine(line_in)
char *line_in;
{
    char *token;
    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(cache_mgr);
    cache_mgr = strdup(token);
}

void parseDirLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(cache_dir);
    cache_dir = strdup(token);

}

void parseLogLine(line_in)
char *line_in;
{
    char *token;
    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(cache_log_file);
    cache_log_file = strdup(token);
    GetInteger(cachelog_flag);
}

void parseHttpdAccelLine(line_in)
char *line_in;
{
    char *token;
    char buf[1024];

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    httpd_accel_host = strdup(token);
    GetInteger(httpd_accel_port);

    sprintf(buf, "http://%s:%d", httpd_accel_host, httpd_accel_port);
    httpd_accel_prefix = strdup(buf);
    httpd_accel_mode = 1;
}

void parseHttpdAccelWithProxyLine(line_in)
char *line_in;
{
    char *proxy;

    proxy = strtok(NULL, w_space);
    if (proxy == (char *) NULL)
	self_destruct(line_in);

    /* set httpd_accel_with_proxy, default is off. */
    if (!strcasecmp(proxy, "on"))
	httpd_accel_with_proxy = 1;
    else if (!strcasecmp(proxy, "off"))
	httpd_accel_with_proxy = 0;
    else
	httpd_accel_with_proxy = 0;
}

void parseEffectiveUserLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    effective_user = strdup(token);

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;			/* group is optional */
    effective_group = strdup(token);
}

void parseAccessLogLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(cache_access_log_file);
    cache_access_log_file = strdup(token);
}

void parseHierachyLogLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(cache_hierarchy_log_file);
    cache_hierarchy_log_file = strdup(token);
    cache_hierarchy_log_flag = 1;
}

void parseFtpProgramLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(ftp_program);
    ftp_program = strdup(token);
}

void parseEmulateLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    if (!strcasecmp(token, "on") || !strcasecmp(token, "enable"))
	emulate_httpd_log = 1;
    else
	emulate_httpd_log = 0;
}

void parseDnsProgramLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(dns_program);
    dns_program = strdup(token);
}

#ifdef USE_WAIS_RELAY
void parseWAISRelayLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    safe_free(wais_relay_host);
    wais_relay_host = strdup(token);

    GetInteger(wais_relay_port);
    GetInteger(wais_max_obj_size);
    wais_max_obj_size <<= 20;
}

#endif

void parseAccessAllowLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToIPACL(&ip_access_control_list, token, IP_ALLOW);
}

void parseManagerAccessAllowLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToIPACL(&manager_ip_access_control_list, token, IP_ALLOW);
}

void parseAccessDenyLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToIPACL(&ip_access_control_list, token, IP_DENY);
}

void parseManagerAccessDenyLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToIPACL(&manager_ip_access_control_list, token, IP_DENY);
}


void parseHttpStopLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToStopList(&http_stoplist, token);
}

void parseGopherStopLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToStopList(&gopher_stoplist, token);
}
void parseFtpStopLine(line_in)
char *line_in;
{
    char *token;

    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	return;
    addToStopList(&ftp_stoplist, token);
}

#ifdef APPEND_DOMAIN
void parseAppendDomainLine(line_in)
char *line_in;
{
    char *token;

    /* parse an append_domain line */
    token = strtok(NULL, w_space);
    if (token == (char *) NULL)
	self_destruct(line_in);
    if (*token != '.')
	self_destruct(line_in);
    append_domain = strdup(token);
    append_domain_len = strlen(token);

}
#endif				/* APPEND_DOMAIN */



/* Maybe a bit heavy handed, but parser is immune to virtually every sane
   persons definition of white space */

int parseConfigFile(file_name)
char *file_name;
{
    FILE *fp = fopen(file_name, "r");
    char *token, tmp_line[BUFSIZ];
    char line_in[BUFSIZ];
    char fatal_str[4096];

    /* Initialize a few global strings, in case they aren't user defined */

    cache_log_file = NULL;
    cache_hierarchy_log_flag = 0;
    cache_access_log_file = strdup("/tmp/cache/cached.access.log");
    cache_hierarchy_log_file = strdup("/tmp/cache/cached.hierarchy.log");
    cache_dir = strdup("/tmp/cache");
    ftp_program = strdup("ftpget.pl");
    dns_program = strdup("dnsserver");

    if (fp == NULL) {
	sprintf(fatal_str, "Unable to open configuration file: %s\n", file_name);
	fatal(fatal_str);
    }
    memset(line_in, '\0', BUFSIZ);
    while (fgets(line_in, BUFSIZ, fp)) {
	if (line_in[0] == '#' || line_in[0] == '\n' || line_in[0] == '\0')
	    continue;		/* skip comments */

	/* Use tmp_line as a temporary pointer to the input line */
	/* AWC Debug */
	if (line_in[0] != '\n') {
	    strcpy(tmp_line, line_in);

	    /* Read first field, then decide how to parse */
	    /* Since we use strncmp(), the order of the cmps counts... */
	    if ((token = strtok(tmp_line, w_space))) {

		/* Parse a cache_host line */
		if (!strncmp(token, "cache_host", 10))
		    parseHostLine(line_in);

		/* Parse a cache_dir line */
		else if (!strncmp(token, "cache_dir", 9))
		    parseDirLine(line_in);

		/* Parse a cache_log line */
		else if (!strncmp(token, "cache_log", 9)) {
		    cache_log_file = strdup("/tmp/cache/cached.log");
		    parseLogLine(line_in);
		}
		/* Parse a cache_access_log line */
		else if (!strncmp(token, "cache_access_log", 16))
		    parseAccessLogLine(line_in);

		/* Parse a cache_hierarchy_log line */
		else if (!strncmp(token, "cache_hierarchy_log", 18))
		    parseHierachyLogLine(line_in);

		/* Parse a httpd_accel_with_proxy line */
		else if (!strncmp(token, "httpd_accel_with_proxy", 22))
		    parseHttpdAccelWithProxyLine(line_in);

		/* Parse a httpd_accel line */
		else if (!strncmp(token, "httpd_accel", 11))
		    parseHttpdAccelLine(line_in);

		/* Parse a cache_effective_user line */
		else if (!strncmp(token, "cache_effective_user", 20))
		    parseEffectiveUserLine(line_in);

		/* Parse a cache_mem_high line */
		else if (!strncmp(token, "cache_swap_high", 15))
		    parseSwapHighLine(line_in);

		/* Parse a cache_mem line_low */
		else if (!strncmp(token, "cache_swap_low", 14))
		    parseSwapLowLine(line_in);

		/* Parse a cache_mem_high line */
		else if (!strncmp(token, "cache_mem_high", 14))
		    parseMemHighLine(line_in);

		/* Parse a cache_mem line_low */
		else if (!strncmp(token, "cache_mem_low", 13))
		    parseMemLowLine(line_in);

		/* Parse a cache_mem line */
		/* XXX: this must be AFTER cache_mem_low, etc. */
		else if (!strncmp(token, "cache_mem", 9))
		    parseMemLine(line_in);

		/* Parse a cache_swap line */
		else if (!strncmp(token, "cache_swap", 10))
		    parseSwapLine(line_in);

		/* Parse a cache_mgr line */
		else if (!strncmp(token, "cache_mgr", 9))
		    parseMgrLine(line_in);

		/* Parse a access_allow line */
		else if (!strncmp(token, "access_allow", 12))
		    parseAccessAllowLine(line_in);

		/* Parse a access_deny line */
		else if (!strncmp(token, "access_deny", 11))
		    parseAccessDenyLine(line_in);

		/* Parse a access_allow line */
		else if (!strncmp(token, "manager_access_allow", 20))
		    parseManagerAccessAllowLine(line_in);

		/* Parse a access_deny line */
		else if (!strncmp(token, "manager_access_deny", 19))
		    parseManagerAccessDenyLine(line_in);

		/* Parse a http_stop line */
		else if (!strncmp(token, "http_stop", 9))
		    parseHttpStopLine(line_in);

		/* Parse a gopher_stop line */
		else if (!strncmp(token, "gopher_stop", 11))
		    parseGopherStopLine(line_in);

		/* Parse a ftp_stop line */
		else if (!strncmp(token, "ftp_stop", 8))
		    parseFtpStopLine(line_in);

		/* Parse a gopher protocol line */
		/* XXX: Must go after any gopher* token */
		else if (!strncmp(token, "gopher", 6))
		    parseGopherLine(line_in);

		/* Parse a http protocol line */
		/* XXX: Must go after any http* token */
		else if (!strncmp(token, "http", 4))
		    parseHttpLine(line_in);

		/* Parse a ftp protocol line */
		/* XXX: Must go after any ftp* token */
		else if (!strncmp(token, "ftp", 3))
		    parseFtpLine(line_in);

		else if (!strncmp(token, "ttl_pattern", 11))
		    parseTTLPattern(line_in);

		/* Parse a negative_ttl line */
		else if (!strncmp(token, "negative_ttl", 12))
		    parseNegativeLine(line_in);

		/* Parse a read_timeout line */
		else if (!strncmp(token, "read_timeout", 12))
		    parseReadTimeoutLine(line_in);

		/* Parse a clean_rate line */
		else if (!strncmp(token, "clean_rate", 10))
		    parseCleanRateLine(line_in);

		/* Parse a client_lifetime line */
		else if (!strncmp(token, "client_lifetime", 15))
		    parseLifetimeLine(line_in);

		/* Parse a cache_ftp_program line */
		else if (!strncmp(token, "cache_ftp_program", 17))
		    parseFtpProgramLine(line_in);

		/* Parse a cache_dns_program line */
		else if (!strncmp(token, "cache_dns_program", 17))
		    parseDnsProgramLine(line_in);

		/* Parse a cache_dns_program line */
		else if (!strncmp(token, "dns_children", 12))
		    parseDnsChildrenLine(line_in);

		/* Parse mail trace line */
		else if (!strncmp(token, "mail_trace", 10))
		    parseMailTraceLine(line_in);

		/* Parse source_ping line */
		else if (!strncmp(token, "source_ping", 11))
		    parseSourcePingLine(line_in);

		/* Parse quick_abort line */
		else if (!strncmp(token, "quick_abort", 11))
		    parseQuickAbortLine(line_in);

		/* Parse old abort_mode line - FOR BACKWARDS COMPATIBILITY */
		else if (!strncmp(token, "abort_mode", 10))
		    parseQuickAbortLine(line_in);

		/* Parse emulate_httpd_log line */
		else if (!strncmp(token, "emulate_httpd_log", 17))
		    parseEmulateLine(line_in);

#ifdef APPEND_DOMAIN
		else if (!strcmp(token, "append_domain"))
		    parseAppendDomainLine(line_in);
#endif				/* APPEND_DOMAIN */

#ifdef USE_WAIS_RELAY
		else if (!strcmp(token, "wais_relay"))
		    parseWAISRelayLine(line_in);
#endif

		/* If unknown, treat as a comment line */
		else {
		    /* do nothing */ ;
		}
	    }
	}
    }

    /* Sanity checks */
    if (default_lifetime < read_timeout) {
	printf("WARNING: client_lifetime (%d seconds) is less than read_timeout (%d seconds).\n", default_lifetime, read_timeout);
	printf("         This may cause serious problems with your cache!!!\n");
	printf("         Change your cached.conf file.\n");
	fflush(stdout);		/* print message */
    }
    if (getCacheSwapMax() < (getCacheMemMax() >> 10)) {
	printf("WARNING: cache_swap (%d kbytes) is less than cache_mem (%d bytes).\n", getCacheSwapMax(), getCacheMemMax());
	printf("         This will cause serious problems with your cache!!!\n");
	printf("         Change your cached.conf file.\n");
	max_cache_swap_size = getCacheMemMax() >> 10;
	printf("         For this run, however, cached will use %d kbytes for cache_swap.\n", max_cache_swap_size);
	fflush(stdout);		/* print message */
    }
    if (clean_rate < 60) {
	clean_rate = (30 * 60);
	printf("WARNING: clean_rate is less than one minute.\n");
	printf("         This will cause serious problems with your cache!!!\n");
	printf("         Change your cached.conf file.\n");
	printf("         For this run, however, cached will use %d minutes for clean_rate.\n", (int) (getCleanRate() / 60));
	fflush(stdout);		/* print message */
    }
    storeSanityCheck(cache_dir);


    if (dns_children < 1 || dns_children > max_dns_children) {
	printf("WARNING: dns_children was set to a bad value: %d\n", dns_children);
	printf("Setting it to the default 3.\n");
	dns_children = 3;
    }
    fclose(fp);

    return 0;
}


/* 
   Daemonize a process according to guidlines in "Advanced Programming
   For The UNIX Environment", W.R. Stevens ( Addison Wesley, 1992) - Ch. 13
 */
int daemonize()
{
    int n_openf, i;
    pid_t pid;


    if ((pid = fork()) < 0)
	return (-1);
    else if (pid != 0)
	exit(0);

    /* Child continues */
    setsid();			/* Become session leader */

    n_openf = getMaxFD();	/* Close any inherited files */
    for (i = 0; i < n_openf; i++)
	close(i);

    umask(0);			/* Clear file mode creation mask */

    return (0);
}


int check_suid()
{
    struct passwd *pwd;
    struct group *grp;

    if (geteuid() == 0) {
	/* Started as a root, check suid option */
	if (effective_user && (pwd = getpwnam(effective_user))) {

	    /* change current directory to swap space so we can get core */
	    if (chdir(cache_dir)) {
		debug(1, "Chdir Failed: Cached cannot write core file when it crash: %s\n",
		      sys_errlist[errno]);
	    }
	    if (effective_group && (grp = getgrnam(effective_group))) {
		setgid(grp->gr_gid);
	    } else {
		setgid(pwd->pw_gid);
	    }
	    setuid(pwd->pw_uid);
	}
    }
    return 0;
}

int getHttpMax()
{
    return http_max_obj_size;
}

int getHttpTTL()
{
    return http_ttl;
}

int getGopherMax()
{
    return gopher_max_obj_size;
}

int getGopherTTL()
{
    return gopher_ttl;
}

#ifdef USE_WAIS_RELAY
int getWAISMax()
{
    return wais_max_obj_size;
}
#endif

int getFtpMax()
{
    return ftp_max_obj_size;
}

int getFtpTTL()
{
    return ftp_ttl;
}

int getNegativeTTL()
{
    return negative_ttl;
}

int getCacheMemMax()
{
    return max_cache_size;
}

int getCacheMemHighWaterMark()
{
    return cache_mem_high_water_mark;
}

int getCacheMemLowWaterMark()
{
    return cache_mem_low_water_mark;
}


int getCacheSwapHighWaterMark()
{
    return cache_swap_high_water_mark;
}

int getCacheSwapLowWaterMark()
{
    return cache_swap_low_water_mark;
}

int getCacheSwapMax()
{
    return max_cache_swap_size;
}

int setCacheSwapMax(size)
int size;
{
    max_cache_swap_size = size;
    return max_cache_swap_size;
}

int getReadTimeout()
{
    return read_timeout;
}

int getClientLifetime()
{
    return default_lifetime;
}

int getCleanRate()
{
    return clean_rate;
}

int getMailTrace()
{
    return mail_trace;
}

int getSourcePing()
{
    return source_ping;
}

int getDNSChildren()
{
    return dns_children;
}

int getQuickAbort()
{
    return quick_abort;
}
