/*
 *  $Id: cache_cf.h,v 1.13 1995/08/21 06:57:55 duane Exp $
 *
 *  File:         config.h
 *  Description:  Declarations of parsing and config functions
 *  Author:       Chuck Neerdaels, USC
 *  Created:      Mon May 23 
 *  Language:     C
 *
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#ifndef _CACHE_CONFIG_H_
#define _CACHE_CONFIG_H_

#include "ansihelp.h"
#include <netinet/in.h>
#include <arpa/inet.h>

typedef struct _stoplist {
    char *key;
    struct _stoplist *next;
} stoplist;

typedef enum {
    IP_ALLOW,
    IP_DENY
} ip_access_type;



typedef struct _ip_acl {
    int a1, a2, a3, a4;
    ip_access_type access;
    struct _ip_acl *next;
} ip_acl;

extern int errno;
extern char *sys_errlist[];
extern char *ftp_program;
extern char *dns_program;
extern int httpd_accel_mode;
extern int httpd_accel_with_proxy;
extern char *httpd_accel_prefix;
extern char *cache_hierarchy_log_file;
extern int cache_hierarchy_log_flag;
extern int zap_disk_store;
extern stoplist *http_stoplist;
extern stoplist *gopher_stoplist;
extern stoplist *ftp_stoplist;
extern ip_acl *ip_access_control_list;
extern ip_acl *manager_ip_access_control_list;

#ifdef APPEND_DOMAIN
extern char *append_domain;
extern int append_domain_len;
#endif				/* APPEND_DOMAIN */

#ifdef USE_WAIS_RELAY
extern char *wais_relay_host;
extern int wais_relay_port;
#endif

extern int parseConfigFile _PARAMS((char *file_name));
extern int daemonize();
extern int check_suid();

extern int getHttpMax();
extern int getHttpTTL();
extern int getGopherMax();
extern int getGopherTTL();
extern int getFtpMax();
extern int getFtpTTL();
extern int getNegativeTTL();
extern int getCacheMemMax();
extern int getCacheMemHighWaterMark();
extern int getCacheMemLowWaterMark();
extern int getCacheSwapMax();
extern int setCacheSwapMax _PARAMS((int size));
extern int getCacheSwapHighWaterMark();
extern int getCacheSwapLowWaterMark();
extern int getReadTimeout();
extern int getClientLifetime();
extern int getCleanRate();
extern int getDNSChildren();
extern int ip_acl_match _PARAMS((int c1, int c2, int c3, int c4,
				 int a1, int a2, int a3, int a4));
extern ip_access_type ip_access_check _PARAMS((struct in_addr address,
					       ip_acl * list));

extern int getMailTrace();
extern int getSourcePing();
extern int getQuickAbort();

#endif
