/* 
 *  $Id: comm.h,v 1.26 1995/08/21 20:55:12 hardy Exp $
 *
 *  File:         comm.h
 *  Description:  Declarations for socket communication  facility.
 *  Commor:       John Noll, USC
 *  Created:      Fri Jun 18 11:55:35 1993
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#ifndef COMM_H
#define COMM_H

#include <netdb.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/param.h>		/* For MAXHOSTNAMELEN */
#include <netinet/in.h>
#include "proto.h"
#include "debug.h"
#include "ansihelp.h"

#define COMM_OK		  (0)
#define COMM_ERROR	 (-1)
#define COMM_NO_HANDLER	 (-2)
#define COMM_NOMESSAGE	 (-3)
#define COMM_TIMEOUT	 (-4)

#define COMM_BLOCKING	  (0x0)
#define COMM_NONBLOCKING  (0x1)
#define COMM_INTERRUPT    (0x2)
#define COMM_DGRAM        (0x4)

/* Select types. */
#define COMM_SELECT_READ   (0x1)
#define COMM_SELECT_WRITE  (0x2)
#define COMM_SELECT_EXCEPT (0x4)
#define COMM_SELECT_TIMEOUT (0x8)
#define COMM_SELECT_LIFETIME (0x10)

/* 
 *  CACHED_RETRIEVE_ERROR_MSG args: 
 *      $1 is URL, 
 *      $2 is URL, 
 *      $3 is protocol type string
 *      $4 is error code, 
 *      $5 is error msg, 
 *      $6 is message to user
 *      $7 is cached version
 *      $8 is cached hostname
 */

#define CACHED_RETRIEVE_ERROR_MSG "\
<TITLE>ERROR: The requested URL could not be retrieved</TITLE>\n\
<H2>ERROR: The requested URL could not be retrieved</H2>\n\
<HR>\n\
<P>\n\
While trying to retrieve the URL:\n\
<A HREF=\"%s\">%s</A>\n\
<P>\n\
The following %s error was encountered:\n\
<UL>\n\
<LI><STRONG>ERROR %d -- %s</STRONG>\n\
</UL>\n\
<P>This means that:\n\
<PRE>\n\
    %s\n\
</PRE>\n\
<P> <HR> <P>\n\
This problem was detected by the\n\
<A HREF=\"http://harvest.cs.colorado.edu/\">Harvest Object Cache</A>\n\
(Version %s) running on the host <EM>'%s'</EM>.\n\
<BR>\n"


typedef int (*PF) _PARAMS((int, caddr_t));

#ifdef __cplusplus
extern "C" {
#endif

/* Global variables */
    typedef struct fde {
	int openned;		/* 1 if openned, 0 if not. */
	int sender;		/* 1 if client, 0 if server */
	int port;		/* Local port #. */
	int (*handler) ();	/* Interrupt handler. */

	/* Select handlers. */
	caddr_t client_data;	/* App. data to associate w/ handled conn. */
	int (*read_handler) ();	/* Read  select handler. */
	caddr_t read_data;	/* App. data to associate w/ handled conn. */
	int (*write_handler) ();	/* Write select handler. */
	caddr_t write_data;	/* App. data to associate w/ handled conn. */
	int (*except_handler) ();	/* Except select handler. */
	caddr_t except_data;	/* App. data to associate w/ handled conn. */
	int (*timeout_handler) ();	/* Timeout handler. */
	time_t timeout_time;	/* Allow 1-second granularity timeouts */
	time_t timeout_delta;	/* The delta requested */
	caddr_t timeout_data;	/* App. data to associate w/ handled conn. */
	int (*lifetime_handler) ();	/* Lifetime expire handler. */
	caddr_t lifetime_data;	/* App. data to associate w/ handled conn. */

	unsigned int comm_type;
    } FD_ENTRY;

    extern FD_ENTRY *fd_table;

    extern int comm_open _PARAMS((unsigned int io_type, int port, PF));
    extern int comm_accept _PARAMS((int fd));
    extern char *comm_client _PARAMS((int fd));
    extern int comm_close _PARAMS((int fd));
    extern int comm_connect _PARAMS((int sock, char *hst, int prt));
    extern int comm_connect_addr _PARAMS((int sock, struct sockaddr_in *));
    extern char *comm_hostname _PARAMS(());
    extern char *comm_hostname_direct _PARAMS(());
	extern u_num32 comm_hostip _PARAMS(());
    extern int comm_init _PARAMS(());
    extern int comm_listen _PARAMS((int sock));
    extern char *comm_peerhost _PARAMS((int fd));
    extern int comm_peerport _PARAMS((int fd));
    extern int comm_port _PARAMS((int fd));
    extern int comm_read _PARAMS((int fd, char *buf, int size));
    extern int comm_select _PARAMS((long sec, long usec, time_t));
    extern int comm_pending _PARAMS((int fd, long sec, long usec));
    extern int comm_sethandler _PARAMS((int fd, PF, caddr_t));
    extern int comm_set_select_handler_plus_timeout _PARAMS((int fd, unsigned int type, PF, caddr_t, time_t));
    extern int comm_set_select_handler _PARAMS((int fd, unsigned int type, PF, caddr_t));
    extern int comm_get_select_handler _PARAMS((int fd, unsigned int type, PF *, caddr_t *));
    extern int comm_write _PARAMS((int fd, char *buf, int size));
    extern int comm_udp_send _PARAMS((int fd, char *host, int port, char *buf, int len));
    extern int comm_udp_sendto _PARAMS((int fd, struct sockaddr_in *, int size, char *buf, int len));
    extern int comm_udp_recv _PARAMS((int, char *, int, struct sockaddr_in *, int *));
    extern int comm_init();
#ifdef __cplusplus
}

#endif
#endif				/* COMM_H */
