/*
 *  $Id: debug.h,v 1.18 1995/08/02 17:08:56 hardy Exp $
 *
 *  File:         debug.h
 *  Description:  debug facility macros and declarations
 *  Author:       John Noll, USC
 *  Created:      Tue Jan  8 11:34:30 1991
 *  Language:     C++
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#ifndef _DEBUG_H_
#define _DEBUG_H_

#include <stdio.h>
#include "ansihelp.h"

extern int _db_level;
extern char *_db_file;
extern int _db_line;
extern int syslog_enable;
extern int stderr_enable;
extern FILE *debug_log_file;

void _db_init _PARAMS((char *, int, char *));
void _db_print _PARAMS((int, char *,...));

/* utility functions */

void *xmalloc _PARAMS((size_t));
void xfree _PARAMS((void *));
void *xrealloc _PARAMS((void *, size_t));
void *xcalloc _PARAMS((size_t, size_t));
char *xstrdup _PARAMS((char *));
char *xstrerror _PARAMS((int));

/* always define debug, but DEBUG not define set the db_level to 0 */

#define debug \
	if ((_db_level >= 0) && \
	    (_db_file = __FILE__) && \
	    (_db_line = __LINE__)) \
        _db_print

#define safe_free(x)	{ \
	debug(10, "free: %p\n", x); \
	if ((x) != NULL) { \
		free(x); \
		(x) = NULL; \
	} }

#endif				/* _DEBUG_H_ */
