static char rcsid[] = "$Id: fdstat.c,v 1.21 1995/08/21 20:57:14 duane Exp $";
/* 
 *  File:         fdstat.c
 *  Description:  File descript stat module
 *  Author:       Anawat Chankhunthod, USC
 *  Created:      Fri Jul  1 00:09:05 PDT 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <sys/param.h>		/* For MAXHOSTNAMELEN */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "debug.h"
#include "fdstat.h"

static int Biggest_FD = 0;

typedef enum {
    CLOSE, OPEN
} File_Desc_Status;

typedef struct _FDENTRY {
    File_Desc_Status status;
    File_Desc_Type type;
} FDENTRY;

static FDENTRY *fd_stat_tab;

extern int getMaxFD();

/* init fd stat module */
int fdstat_init(preopen)
int preopen;
{
    int i, max_fd = getMaxFD();

    fd_stat_tab = (FDENTRY *) xmalloc(sizeof(FDENTRY) * max_fd);
    memset(fd_stat_tab, '\0', sizeof(FDENTRY) * max_fd);
    for (i = 0; i < preopen; ++i) {
	fd_stat_tab[i].status = OPEN;
	fd_stat_tab[i].type = File;
    }

    for (i = preopen; i < max_fd; ++i) {
	fd_stat_tab[i].status = CLOSE;
	fd_stat_tab[i].type = Other;
    }

    Biggest_FD = preopen - 1;
    return 0;
}

/* call for updating the current biggest fd */
void fdstat_update(fd, status)
int fd;
File_Desc_Status status;
{
    unsigned int i;

    if (fd >= getMaxFD())
        debug(0, "Running out of file descriptors");

    if (fd < Biggest_FD) {
	/* nothing to do here */
	return;
    }
    if ((fd > Biggest_FD) && (status == OPEN)) {
	/* just update the biggest one */
	Biggest_FD = fd % getMaxFD();
	return;
    }
    if ((fd == Biggest_FD) && (status == CLOSE)) {
	/* just scan to Biggest_FD - 1 */
	for (i = Biggest_FD; i > 0; --i) {
	    if (fd_stat_tab[i].status == OPEN)
		break;
	}
	Biggest_FD = i;
	return;
    }
    if ((fd == Biggest_FD) && (status == OPEN)) {
	/* do nothing here */
	/* it could happen since some of fd are out of our control */
	return;
    }
    debug(0, "WARNING: fdstat_update: Internal inconsistency:\n");
    debug(0, "         Biggest_FD = %d, this fd = %d, status = %s\n",
	  Biggest_FD, fd, status == OPEN ? "OPEN" : "CLOSE");
    debug(0, "         fd_stat_tab[%d].status == %s\n",
	  fd, fd_stat_tab[fd].status == OPEN ? "OPEN" : "CLOSE");
    debug(0, "         fd_stat_tab[%d].type == %s\n",
	  fd, fd_stat_tab[fd].type == File ? "File" :
	  fd_stat_tab[fd].type == Socket ? "Socket" : "Other");

    return;
}


/* call when open fd */
void fdstat_open(fd, type)
int fd;
File_Desc_Type type;
{
    fd_stat_tab[fd].status = OPEN;
    fd_stat_tab[fd].type = type;
    fdstat_update(fd, OPEN);
}

/* call when close fd */
void fdstat_close(fd)
int fd;
{
    fd_stat_tab[fd].status = CLOSE;
    fdstat_update(fd, CLOSE);
}

/* return the biggest fd */
int fdstat_biggest_fd()
{
    return Biggest_FD;
}
