static char rcsid[] = "$Id: ftp.c,v 1.63 1995/08/21 21:23:17 duane Exp $";
/* 
 *  File:         ftp.c
 *  Description:  state machine for ftp retrieval protocol.  Based on John's
 *                gopher retrieval module.
 *  Author:       Anawat Chankhunthod, USC
 *  Created:      Tue May 28 10:57:11 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <sys/param.h>		/* For MAXHOSTNAMELEN. */
#include <sys/errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>		/* for WNOHANG */

#include "comm.h"
#include "store.h"
#include "stat.h"
#include "url.h"
#include "mime.h"
#include "fdstat.h"
#include "cache_cf.h"
#include "ttl.h"
#include "config.h"

static char ftpASCII[] = "A";
static char ftpBinary[] = "I";

extern int errno;
extern char *sys_errlist[];

typedef struct _Ftpdata {
    StoreEntry *entry;
    char host[MAXHOSTNAMELEN];
    char type_id;
    char request[MAX_URL];
    char *type;
    char *mime_hdr;
    FILE *pfp;
    int cpid;
    int ftp_fd;
} FtpData;

extern char *tmp_error_buf;

static FILE *ftp_open_pipe();
static void ftp_close_pipe();

/* XXX: this does not support FTP on a different port! */
int ftp_url_parser(url, host, request, user, password)
char *url;
char *host;
char *request;
char *user;
char *password;
{
    char atypebuf[MAX_URL], hostbuf[MAX_URL], *tmp;
    int t;

    /* initialize everything */
    atypebuf[0] = hostbuf[0] = '\0';
    request[0] = host[0] = user[0] = password[0] = '\0';

    t = sscanf(url, "%[a-zA-Z]://%[^/]%s", atypebuf, hostbuf, request);
    if ((t < 2) ||
      !(!strcasecmp(atypebuf, "ftp") || !strcasecmp(atypebuf, "file"))) {
	return -1;
    } else if (t == 2) {	/* no request */
	strcpy(request, "/");
    } else {
	tmp = url_convert_hex(request);		/* convert %xx to char */
	strncpy(request, tmp, MAX_URL);
	safe_free(tmp);
    }

    /* url address format is something like this:
       [ userid [ : password ] @ ] host 
       or possibly even
       [ [ userid ] [ : [ password ] ] @ ] host

       So we must try to make sense of it.  */

    /* XXX: this only support [user:passwd@]host */
    t = sscanf(hostbuf, "%[^:]:%[^@]@%s", user, password, host);
    if (t < 3) {
	strcpy(host, user);	/* no login/passwd information */
	strcpy(user, "anonymous");
	strcpy(password, "harvest@");
    }
    /* we need to convert user and password for URL encodings */
    tmp = url_convert_hex(user);
    strcpy(user, tmp);
    safe_free(tmp);

    tmp = url_convert_hex(password);
    strcpy(password, tmp);
    safe_free(tmp);

    return 0;
}

int ftpCachable(url, type, mime_hdr)
char *url;
char *type;
char *mime_hdr;
{
    stoplist *p;

    /* scan stop list */
    p = ftp_stoplist;
    while (p) {
	if (strstr(url, p->key))
	    return 0;
	p = p->next;
    }

    /* else cachable */
    return 1;
}

/* This will be called when socket lifetime is expired. */
void ftpLifetimeExpire(fd, data)
int fd;
FtpData *data;
{
    debug(4, "ftpLifeTimeExpire: FD %d: <URL:%s>\n", fd, data->entry->url);
    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
	    data->entry->url,
	    data->entry->url,
	    "FTP",
	    110,
	    "Transaction Timeout",
      "The Network/Remote site may be down or too slow. Try again later",
	    HARVEST_VERSION,
	    comm_hostname());
    storeAbort(data->entry, tmp_error_buf);
    ftp_close_pipe(data->pfp, data->cpid);
    /* do it to update fdstat and fdtable */
    comm_close(fileno(data->pfp));
    safe_free(data);
}



/* This will be called when data is ready to be read from fd.  Read until
   error or connection closed. */
int ftpReadReply(fd, data)
int fd;
FtpData *data;
{
    char buf[4096];
    int len = read(fd, buf, 4096);
    debug(5, "ftpReadReply - fd: %d read len:%d\n", fd, len);

    if (len < 0) {
	debug(1, "ftpReadReply - error reading: %s\n",
	      sys_errlist[errno]);
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"FTP",
		305,
		"Read Error.",
		"Network/Remote Site is down. Try again later.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(data->entry, tmp_error_buf);
	ftp_close_pipe(data->pfp, data->cpid);
	/* do it to update fdstat and fdtable */
	comm_close(fileno(data->pfp));
	safe_free(data);
    } else if (len == 0) {
	/* Connection closed; retrieval done. */
	data->entry->ttl = ttlSet(data->entry);
	storeComplete(data->entry);
	ftp_close_pipe(data->pfp, data->cpid);
	/* do it to update fdstat and fdtable */
	comm_close(fileno(data->pfp));
	safe_free(data);
    } else if ((data->entry->e_current_len + len) > getFtpMax()) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"FTP",
		306,
		"Object is too large.",
		"Contact your system manager for further help, or retrieve the object directly without a proxy.",
		HARVEST_VERSION,
		comm_hostname()
	    );
	storeAbort(data->entry, tmp_error_buf);
	ftp_close_pipe(data->pfp, data->cpid);
	/* do it to update fdstat and fdtable */
	comm_close(fileno(data->pfp));
	safe_free(data);
    } else if (data->entry->flag & CLIENT_ABORT_REQUEST) {
	/* append the last bit of info we get */
	storeAppend(data->entry, buf, len);
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"FTP",
		307,
		"Client Aborted",
		"Client(s) dropped connection before transmission is complete.\nObject fetching is aborted.\n",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);
    } else {
	storeAppend(data->entry, buf, len);
	comm_set_select_handler(fd, COMM_SELECT_READ, (PF) ftpReadReply, (caddr_t) data);
    }
    return 0;
}

/* This will be called when request write is complete. Schedule read of reply. */
int ftpSendComplete(fd, buf, size, errflag, data)
int fd;
char *buf;
int size;
int errflag;
FtpData *data;
{
    debug(1, "ftpSendComplete - THIS SHOULD NOT HAPPEN fd: %d size: %d errflag: %d\n",
	  fd, size, errflag);
    return 0;
}

/* This will be called when connect completes. Write request. */
int ftpSendRequest(fd, data)
int fd;
FtpData *data;
{
    debug(1, "ftpSendRequest - THIS SHOULD NOT HAPPEN fd: %d\n", fd);
    return 0;
}

int ftpStart(unusedfd, url, entry)
int unusedfd;
char *url;
StoreEntry *entry;
{
    /* Create state structure. */
    char user[MAX_URL], password[MAX_URL];
    FtpData *data = (FtpData *) xmalloc(sizeof(FtpData));

    debug(3, "FtpStart - url: %s\n", url);

    memset(data, '\0', sizeof(FtpData));	/* must init */
    data->entry = entry;

    /* Parse url. */
    user[0] = password[0] = '\0';
    if (ftp_url_parser(url, data->host, data->request, user, password)) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		entry->url,
		entry->url,
		"FTP",
		309,
		"Invalid URL syntax: Cannot parse.",
		"Please contact your system manager for further help.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(entry, tmp_error_buf);
	safe_free(data);
	return COMM_ERROR;
    }
    debug(5, "FtpStart - host:%s, request:%s, user:%s, passwd:%s\n",
	  data->host, data->request, user, password);

    /* Create socket. */
    data->pfp = ftp_open_pipe(ftp_program, data->host, data->request,
			      user, password, "r", &(data->cpid));

    if (data->pfp == NULL) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		entry->url,
		entry->url,
		"FTP",
		308,
		"Cannot connect to FTP slave process",
		"Please contact your system manager for further help.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(entry, tmp_error_buf);
	safe_free(data);
	return COMM_ERROR;
    } else {
	debug(5, "startftp - conn %d EINPROGRESS\n", data->ftp_fd);
    }

    /* Connection OK here */
    fdstat_open(fileno(data->pfp), Other);
    data->ftp_fd = fileno(data->pfp);
    setNonBlocking(data->ftp_fd);

    /* Install connection complete handler. */
    comm_set_select_handler(data->ftp_fd, COMM_SELECT_WRITE, 0, 0);
    comm_set_select_handler(data->ftp_fd, COMM_SELECT_LIFETIME, (PF) ftpLifetimeExpire,
			    (caddr_t) data);
    comm_set_select_handler(data->ftp_fd, COMM_SELECT_READ, (PF) ftpReadReply, (caddr_t) data);
    return COMM_OK;
}

/*
 *  ftp_open_pipe - This opens a pipe to the ftpget.pl command.
 *  It currently supports read-only pipes and hardcoded args.  The child
 *  process only has stdin from /dev/null, stdout to the pipe,
 *  and stderr inherited from the parent.  cpid is set to the
 *  pid of the child process or to -1 on error.  Returns a read-only
 *  FILE pointer to the read end of the pipe, or NULL on error.  
 *
 *  Allows process to make many ftp_open_pipe() calls.  -DH
 */
static FILE *ftp_open_pipe(p1, p2, p3, p4, p5, type, cpid)
char *p1, *p2, *p3, *p4, *p5, *type;
int *cpid;
{
    int pfd[2], pid, fd;
    char *transfer;

    if (p3[strlen(p3) - 1] == '/')
	transfer = ftpASCII;
    else {
	char *ext;
	ext_table_entry *e;

	if ((ext = strrchr(p3, '.')) != NULL) {
	    ext++;
	    transfer = ((e = mime_ext_to_type(ext)) &&
		      strncmp(e->mime_type, "text", 4) == 0) ? ftpASCII :
		ftpBinary;
	} else
	    transfer = ftpBinary;
    }

    *cpid = -1;			/* initialize first */

    if (type == NULL || strcmp(type, "r") != 0)
	return (NULL);		/* unsupported */

    if (pipe(pfd) < 0)
	return (NULL);

    if ((pid = fork()) < 0)
	return (NULL);

    if (pid != 0) {		/* parent */
	*cpid = pid;		/* pass child pid */
	(void) close(pfd[1]);	/* close the write pipe */
	return (fdopen(pfd[0], "r"));	/* return fluffy read pipe */
    }
    /* child */
    if ((fd = open("/dev/null", O_RDONLY, 0)) > 0)
	(void) dup2(fd, 0);	/* /dev/null -> stdin */
    dup2(pfd[1], 1);		/* stdout -> write pipe */
    /* stderr is inherited */

    /* close all file desc, and make sure we close the read pipe */
    for (fd = 3; fd < 64; fd++)
	(void) close(fd);
    (void) close(pfd[0]);

    /*
     *  Run the ftpget.pl command:
     *   p1 is the ftpget.pl program, need execlp() to use PATH
     *   p2 is the remote host
     *   p3 is the remote file
     *   transfer is "A" for ASCII and "I" for binary transfer
     *   p4 is the username
     *   p5 is the password
     */
    execlp(p1, p1, "-htmlify", "-", p2, p3, transfer, p4, p5, NULL);
    perror(p1);
    _exit(1);
}

/*
 *  ftp_close_pipe - closes the pipe opened by ftp_open_pipe.  
 *  Non-blocking.  -DH
 */
static void ftp_close_pipe(fp, cpid)
FILE *fp;
int cpid;
{
    int status;
    int ret;

    (void) fclose(fp);		/* close stdio ptr */
    ret = waitpid(cpid, &status, WNOHANG);	/* non-blocking wait */

    if (ret < 0 || ret != cpid)
	return;

    if (WIFSIGNALED(status)) {
	debug(0, "%s exited due to signal %d\n",
	      ftp_program, WTERMSIG(status));
    } else if (WIFEXITED(status) && WEXITSTATUS(status) > 0) {
	debug(5, "%s returned exit status %d\n",
	      ftp_program, WEXITSTATUS(status));
    }
}
