static char rcsid[] = "$Id: icp_lib.c,v 1.22 1995/08/03 22:36:53 hardy Exp $";
/* 
 *  File:         icp_lib.c
 *  Description:  library of icp send functions.
 *  Author:       John Noll, USC
 *                Dean Povey, University of Queensland
 *  Created:      Wed Mar 30 18:12:01 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "debug.h"
#include "icp_lib.h"
#include "proto.h"
#include "comm.h"

extern int errno;
#if !defined(__FreeBSD__)
extern char *sys_errlist[];
#endif

int icp_proto_init()
{
}

long icp_proto_reqnum()
{
}

/* Send a QUERY request to server  */
int icp_query(sock, reqnum, auth, rid, url, hostname, port, options)
int sock;
u_num32 reqnum;
u_num32 *auth;
struct in_addr rid;
char *url;
char *hostname;
int port;
unsigned char options;
{
    char *buf, *p;
    int len;
    u_num32 myAddress;
    icp_common_t header;

    len = sizeof(icp_common_t) + sizeof(struct in_addr) + strlen(url) + 1;

	if (options)
	{
		len += sizeof(unsigned char);
		header.offset  = htons(sizeof(icp_common_t) + sizeof(unsigned char));
		header.nr_options = 1;
	}
	else
	{
		header.nr_options = 0;
		header.offset  = htons(sizeof(icp_common_t));
	}
    buf = (char *) xcalloc(1, len);
    header.opcode  = ICP_OP_QUERY;
	header.shostid = comm_hostip();  /* Already in net byte order */
	header.reqnum = reqnum;
	memset(header.padding, 0, sizeof(u_num32) * 3);
	header.version = ICP_VERSION_CURRENT;
    header.length  = htons(len);
    p = buf;
    memcpy(p, &header, sizeof(header));
    p = buf + ntohs(header.offset);

	/* Copy options */
	if (options)
	{
		memcpy(buf + sizeof(icp_common_t), &options, 
			sizeof(unsigned char));
	}

    myAddress = htonl(rid.s_addr);
    memcpy(p, &myAddress, sizeof(struct in_addr));
    p += sizeof(struct in_addr);
    memcpy(p, url, strlen(url));	/* already zero filled by calloc */
    return comm_udp_send(sock, hostname, port, buf, header.length);
}

/* Send an ADD advertisment to the server  */
/* XXX this should be able to add more than one pointer */
int icp_add(sock, reqnum, auth, rid, url, pointer, hostname, port, 
	timestamp, ttl, size)
int sock;
u_num32 reqnum;
u_num32 *auth;
struct in_addr rid;
char *url;
char *pointer;
char *hostname;
int port;
u_num32 timestamp;
u_num32 ttl;
u_num32 size;
{
    char *buf, *p;
    unsigned short len;
    u_num32 myAddress;
    icp_common_t header;
	unsigned short url_len;
	unsigned char scope=10;
	unsigned char nr_items=1;

    len = sizeof(icp_common_t) +      /* Header */
		(2 *sizeof(unsigned char)) +  /* Scope and Number of items */
		(3 * sizeof(u_num32)) +       /* Size, ttl and timestamp */
		(2* sizeof(unsigned short)) + /* Source and Pointer URL lengths */
		strlen(url)  +  1 +           /* Source Url */ 
		strlen(pointer) + 1;          /* Pointer Url */

	header.nr_options = 0;
	header.offset  = htons(sizeof(icp_common_t));
    buf = (char *) xcalloc(1, len);
    header.opcode  = ICP_OP_ADD;
	header.shostid = comm_hostip();  /* Already in net byte order */
	header.reqnum = reqnum;
	memset(header.padding, 0, sizeof(u_num32) * 3);
	header.version = ICP_VERSION_CURRENT;
    header.length  = htons(len);
    p = buf;
    memcpy(p, &header, sizeof(header));
    p = buf + ntohs(header.offset);

	/* Add the data for this advertisment */
	memcpy(p, &scope, sizeof(unsigned char));
	p += sizeof(unsigned(char));

	/* Add the number of items for this advertisment */
	memcpy(p, &nr_items, sizeof(unsigned char));
	p += sizeof(unsigned(char));

	/* Add the data for this pointer */
	size = htonl(size);
	memcpy(p, &size, sizeof(u_num32));
	p += sizeof(u_num32);
	ttl = htonl(ttl);
	memcpy(p, &ttl, sizeof(u_num32));
	p += sizeof(u_num32);
	timestamp = htonl(timestamp);
	memcpy(p, &timestamp, sizeof(u_num32));
	p += sizeof(u_num32);

	/* Add the pointer and src urls */ 
	url_len = htons(strlen(url) + 1);
	memcpy(p, &url_len, sizeof(unsigned short));
	p += sizeof(unsigned short);
    memcpy(p, url, strlen(url));	/* already zero filled by calloc */
	p += htons(url_len);

	url_len = htons(strlen(pointer) + 1);
    memcpy(p, &url_len, sizeof(unsigned short));
    p += sizeof(unsigned short);
    memcpy(p, pointer, strlen(pointer));    /* already zero filled by calloc */

    return comm_udp_send(sock, hostname, port, buf, header.length);
}

/* Send an DELETE advertisment to the server  */
/* XXX this should be able to delete more than one pointer */
int icp_delete(sock, reqnum, auth, rid, url, pointer, hostname, port)
int sock;
u_num32 reqnum;
u_num32 *auth;
struct in_addr rid;
char *url;
char *pointer;
char *hostname;
int port;
{
    char *buf, *p;
    unsigned short len;
    u_num32 myAddress;
    icp_common_t header;
	unsigned short url_len;
	unsigned char scope=10;
	unsigned char nr_items=1;

    len = sizeof(icp_common_t) +      /* Header */
		(2 *sizeof(unsigned char)) +  /* Scope and Number of items */
		(2* sizeof(unsigned short)) + /* Source and Pointer URL lengths */
		strlen(url)  +  1 +           /* Source Url */ 
		strlen(pointer) + 1;          /* Pointer Url */

	header.nr_options = 0;
	header.offset  = htons(sizeof(icp_common_t));
    buf = (char *) xcalloc(1, len);
    header.opcode  = ICP_OP_DELETE;
	header.shostid = comm_hostip();  /* Already in net byte order */
	header.reqnum = reqnum;
	memset(header.padding, 0, sizeof(u_num32) * 3);
	header.version = ICP_VERSION_CURRENT;
    header.length  = htons(len);
    p = buf;
    memcpy(p, &header, sizeof(header));
    p = buf + ntohs(header.offset);

	/* Add the data for this advertisment */
	memcpy(p, &scope, sizeof(unsigned char));
	p += sizeof(unsigned(char));

	/* Add the number of items for this advertisment */
	memcpy(p, &nr_items, sizeof(unsigned char));
	p += sizeof(unsigned(char));

	/* Add the pointer and src urls */ 
	url_len = htons(strlen(url) + 1);
	memcpy(p, &url_len, sizeof(unsigned short));
	p += sizeof(unsigned short);
    memcpy(p, url, strlen(url));	/* already zero filled by calloc */
	p += htons(url_len);

	url_len = htons(strlen(pointer) + 1);
    memcpy(p, &url_len, sizeof(unsigned short));
    p += sizeof(unsigned short);
    memcpy(p, pointer, strlen(pointer));    /* already zero filled by calloc */

    return comm_udp_send(sock, hostname, port, buf, header.length);
}

/* Send a SEND object request over SOCK. */
int icp_send(sock, reqnum, auth, rid, url)
int sock;
u_num32 reqnum;
u_num32 *auth;
struct in_addr rid;
char *url;
{
    icp_common_t header;
    int len = sizeof(icp_common_t) + sizeof(struct in_addr) + strlen(url) + 1;
    char *p, *buf;

    buf = (char *) xcalloc(1, len);
    memset(&header, '\0', sizeof(icp_common_t));
    header.opcode = htons(ICP_OP_SEND);
    header.length = htons(len);
	header.offset = htons(sizeof(icp_common_t));
	header.nr_options = 0;
    p = buf;
    memcpy(p, &header, sizeof(icp_common_t));
    p += sizeof(icp_common_t);
    memcpy(p, &rid, sizeof(struct in_addr));
    p += sizeof(struct in_addr);
    memcpy(p, url, strlen(url));	/* already zero filled by xcalloc */

    write(sock, buf, len);
    return 0;
}

static int ReadDataBegin(sock, msg)
int sock;
icp_object *msg;
{
    int len, result = COMM_OK;
    icp_datab_t tmp;

    if ((len = read(sock, (char *) &tmp, 3 * sizeof(u_num32))) <= 0) {
	/* Return error; assume zero bytes read is closed connection. */
	debug(1, "icp_libReadDataBegin - error reading DATABEG header\n");
	result = COMM_ERROR;
    } else {
	char *buf, *p;
	int to_read;

	msg->ttl = ntohl(tmp.db_ttl);
	msg->timestamp = ntohl(tmp.db_ts);
	msg->object_size = ntohl(tmp.db_size);

	if (msg->object_size) {
	    msg->buf_len = msg->object_size;
	} else {
	    msg->buf_len = msg->header.length;
	}

	buf = (char *) xcalloc(1, msg->buf_len);
	to_read = (msg->header.length - sizeof(icp_common_t) -
		   3 * sizeof(u_num32));
	p = buf;

	debug(4, "ReadDataBegin - reading data size = %d\n", to_read);

	while (to_read) {
	    if ((len = read(sock, p, to_read)) < 0) {
		debug(1, "ReadDataBegin - error reading data: %s\n",
		      sys_errlist[errno]);
		result = COMM_ERROR;
	    } else {
		debug(4, "ReadDataBegin - read  %d bytes\n", len);
		to_read -= len;
		p += len;
		result = COMM_OK;
	    }
	}
	debug(4, "ReadDataBegin - total %d bytes read\n", p - buf);
	msg->data = buf;
	msg->offset += p - buf;
    }
    return result;
}

static int ReadData(sock, msg)
int sock;
icp_object *msg;
{
    int result = COMM_OK, len;
    u_num32 tmp;
    char url[ICP_MAX_URL];

    debug(3, "ReadData\n");
    if ((len = read(sock, (char *) &tmp, sizeof(u_num32))) <= 0) {
	/* Return error; assume zero bytes read is closed connection. */
	debug(1, "ReadData - error reading DATA header\n");
	result = COMM_ERROR;
    } else {
	u_num32 offset = ntohl(tmp);
	int msg_len = (msg->header.length - sizeof(icp_common_t)
		       - sizeof(u_num32));
	if (msg->buf_len < (msg->offset + msg_len)) {
	    msg->buf_len += msg_len;
	    msg->data = (char *) xrealloc(msg->data, msg->buf_len);
	}
	while (msg_len) {
	    if ((len = read(sock, msg->data + msg->offset, msg_len)) < 0) {
		/* Return error; assume zero bytes read is closed connection. */
		debug(1, "ReadData - error reading data\n");
		result = COMM_ERROR;
	    } else {
		msg->offset += len;
		msg_len -= len;
	    }
	}
    }
    return result;
}


static ReadError(sock, msg)
int sock;
icp_object *msg;
{
    int len, buf_len = (msg->header.length - sizeof(icp_common_t)
			- sizeof(unsigned short) + 1);
    unsigned short code;
    char *buf = (char *) xcalloc(1, buf_len);

    debug(1, "ReadError\n");

    if ((len = read(sock, (char *) &code, sizeof(unsigned short))) <= 0) {
	/* Return error; assume zero bytes read is closed connection. */
	debug(1, "ReadError - error reading error code\n");
    } else if ((len = read(sock, (char *) buf, buf_len)) <= 0) {
	/* Return error; assume zero bytes read is closed connection. */
	debug(1, "ReadError - error reading error message\n");
    } else {
	debug(1, "ReadError - msg: %s\n", buf);
	msg->data = buf;
    }
}

int icp_receive_data(sock, msg)
int sock;
icp_object *msg;
{
    int result = COMM_OK, done = 0, len;
    u_num32 query_host;
    char url[ICP_MAX_URL];

    while (!done && !result) {
	if ((len = read(sock, (char *) &msg->header, sizeof(icp_common_t))) <= 0) {
	    /* Return error; assume zero bytes read is closed connection. */
	    debug(1, "icp_receive_data - error reading header\n");
	    result = COMM_ERROR;	/* Will cause HandleRead to close conn. */
	} else {
	    int op = ntohs(msg->header.opcode);
	    /* Process request. */
	    if (op == ICP_OP_DATABEG) {
		debug(1, "icp_receive_data - processing ICP_OP_DATABEG\n");
		result = ReadDataBegin(sock, msg);
	    } else if (op == ICP_OP_DATA) {
		debug(1, "icp_receive_data - processing ICP_OP_DATA\n");
		result = ReadData(sock, msg);
	    } else if (op == ICP_OP_DATAEND) {
		debug(1, "icp_receive_data - processing ICP_OP_DATAEND\n");
		result = ReadData(sock, msg);
		done = 1;
	    } else if (op == ICP_OP_ERR) {
		debug(1, "icp_receive_data - processing ICP_OP_ERR\n");
		result = ReadError(sock, msg);
		done = 1;
	    } else {
		/* Should not be any other opcode. */
		debug(1, "icp_receive_data - invalid opcode recieved: %d\n", op);
	    }
/*      if (msg->offset == msg->object_size) done = 1; */
	}
    }
    return result;
}
