
/* 
 *  $Id: ipcache.h,v 1.15 1995/08/21 20:55:43 hardy Exp $
 *
 *  File:         ipcache.h
 *  Description:  ip address cache. speed up gethostbyname()
 *  Author:       Anawat Chankhunthod
 *  Created:      
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#ifndef _IPCACHE_H_
#define _IPCACHE_H_
#include <netdb.h>

/* typedef  int(*IPH) _PARAMS((int, ipcache_entry *, caddr_t)); */
typedef int (*IPH) _PARAMS((int, struct hostent *, caddr_t));

struct hostent *ipcache_gethostbyname();
int ipcache_nbgethostbyname();
void ipcache_init();
void ipcache_flush();

typedef struct _ipcache_entry {
    char *name;
    long timestamp;
    long lastref;
    long ttl;
    unsigned char addr_count;
    unsigned char alias_count;
    enum {
	CACHED, PENDING, NEGATIVE_CACHED
    } status:2;
    struct hostent entry;
    struct _ip_pending *pending_head;
    struct _ip_pending *pending_tail;
} ipcache_entry;

#endif
