static char rcsid[] = "$Id: main.c,v 1.88 1995/08/21 20:55:45 hardy Exp $";
/* 
 *  File:         main.c
 *  Description:  main loop for cache
 *  Author:       John Noll, USC
 *  Created:      Mon Dec 13 10:10:28 1993 (John Noll, USC) sfdif
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <stdio.h>
#include <limits.h>		/* for USHRT_MAX */
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <signal.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "cache_cf.h"
#include "debug.h"
#include "comm.h"
#include "icp.h"
#include "stat.h"
#include "stack.h"
#include "fdstat.h"
#include "ipcache.h"
#include "config.h"


/* WRITE_PID_FILE - tries to write a cached.pid file on startup */
#ifndef WRITE_PID_FILE
#define WRITE_PID_FILE
#endif

time_t cached_starttime = 0L;
time_t next_cleaning = 0L;
extern time_t cached_curtime;
int do_mail = 0;
int prompt = 0;
int theAsciiConnection = -1;
int theBinaryConnection = -1;
int theUdpConnection = -1;
int do_reuse = 1;
void death(), deathb(), shut_down();
int debug_level = 0;
unsigned long nconn = 0L;
#define N_UPPER_CACHES 10
int catch_signals = 1;
char *tmp_error_buf = NULL;

char *config_file;
extern char *cache_dir;
extern char *cache_log_file;
extern char *cache_access_log_file;
extern int cachelog_flag;

void raise_debug_lvl(), reset_debug_lvl();

Stack free_pkt_list;

/*
 *  send_mail_to_ucboulder() - Sends the hostname, IP number, and port
 *  number via email to the University of Colorado, Boulder.  You can
 *  avoid this code by defining 'mail_trace off' in cached.conf.
 */
void send_mail_to_ucboulder(portnum)
int portnum;
{
    char *hostname;
    struct hostent *hep;
    FILE *pipefp;
    struct in_addr cin;

    if (getMailTrace() == 0)
	return;

    debug(1, "Sending trace mail to UC-Boulder\n");

    if ((hostname = comm_hostname())
	&& (hep = ipcache_gethostbyname(hostname))
	&& (access("/bin/mail", X_OK) == 0)
      && (pipefp = popen("/bin/mail cache_tracker@cs.colorado.edu", "w"))
	&& hep->h_name && hep->h_addr_list && hep->h_addr_list[0]) {
	memcpy(&cin, hep->h_addr_list[0], hep->h_length);
	fprintf(pipefp, "To: cache_tracker@cs.colorado.edu\n");
	fprintf(pipefp, "Subject: cache trace from %s\n\n", hep->h_name);
	fprintf(pipefp, "\n\n\n%s %s %d\n.\n\n", hep->h_name, inet_ntoa(cin),
		portnum);
	fflush(pipefp);
	pclose(pipefp);
    }
}

int main(argc, argv)
int argc;
char **argv;
{
    int c, asciiPort = 3128, binaryPort = 3129, udpPort = 3130;
    int malloc_debug_level = 0;
    int debug_level_overwrite = 0;
    extern char *optarg;
    extern int optind;
    int errcount = 0;
    static int neighbors = 0;

    cached_starttime = cached_curtime = time(0L);

    /* try to use as many file descriptors as possible */
    /* System V uses RLIMIT_NOFILE and BSD uses RLIMIT_OFILE */
#if defined(HAVE_SETRLIMIT)
    {
	struct rlimit rl;

#if   defined(RLIMIT_NOFILE)
	if (getrlimit(RLIMIT_NOFILE, &rl) < 0) {
	    perror("getrlimit: RLIMIT_NOFILE");
	} else {
	    rl.rlim_cur = rl.rlim_max;	/* set it to the max */
	    if (setrlimit(RLIMIT_NOFILE, &rl) < 0) {
		perror("setrlimit: RLIMIT_NOFILE");
	    }
	}
#elif defined(RLIMIT_OFILE)
	if (getrlimit(RLIMIT_OFILE, &rl) < 0) {
	    perror("getrlimit: RLIMIT_OFILE");
	} else {
	    rl.rlim_cur = rl.rlim_max;	/* set it to the max */
	    if (setrlimit(RLIMIT_OFILE, &rl) < 0) {
		perror("setrlimit: RLIMIT_OFILE");
	    }
	}
#endif
    }
#endif

#ifdef sun
    /* set malloc option */
    /* use small block algorithm for faster allocation */
    /* grain of small block */
    mallopt(M_GRAIN, 16);
    /* biggest size that is considered a small block */
    mallopt(M_MXFAST, 4096);
    /* number of holding small block */
    mallopt(M_NLBLKS, 100);
#endif

    /* allocate storage for error messages */
    tmp_error_buf = xcalloc(1, 8192);
/*init comm module */
    comm_init();

#ifdef DAEMON
    if (daemonize()) {
	fprintf(stderr, "Error: couldn't create daemon process\n");
	exit(0);
    }
    /*  signal( SIGHUP, restart ); *//* restart if/when proc dies */
#endif				/* DAEMON */

    /* we have to init fdstat here. */
    fdstat_init(PREOPEN_FD);

    config_file = xcalloc(1, 1024);
    strncpy(config_file, "/etc/cached.conf", 1024);

    /* enable syslog by default */
    syslog_enable = 1;
    /* disable stderr debug printout by default */
    stderr_enable = 0;
    /* preinit for debug module */
    debug_log_file = stderr;
    hash_init(0);

    while ((c = getopt(argc, argv, "vCRseif:a:p:u:d:m:zh?")) != -1)
	switch (c) {
	case 'v':
	    printf("Harvest Cache: Version %s\n", HARVEST_VERSION);
	    exit(0);
	case 'C':
	    catch_signals = 0;
	    break;
	case 's':
	    syslog_enable = 0;
	    break;
	case 'e':
	    stderr_enable = 1;
	    break;
	case 'R':
	    do_reuse = 0;
	    break;
	case 'i':
	    prompt = 1;
	    break;
	case 'f':
	    strncpy(config_file, optarg, 1024);
	    break;
	case 'a':
	    asciiPort = atoi(optarg);
	    break;
	case 'p':
	    binaryPort = atoi(optarg);
	    break;
	case 'u':
	    udpPort = atoi(optarg);
	    break;
	case 'd':
	    stderr_enable = 1;
	    debug_level_overwrite = 1;
	    debug_level = atoi(optarg);
	    break;
	case 'm':
	    malloc_debug_level = atoi(optarg);
	    break;
	case 'z':
	    zap_disk_store = 1;
	    break;
	case '?':
	case 'h':
	default:
	    printf("\
Usage: cached [-Rsehvz] [-f config-file] [-d debug-level] [-[apu] port]\n\
       -e        Print messages to stderr.\n\
       -h        Print help message.\n\
       -s        Disable syslog output.\n\
       -v        Print version.\n\
       -z        Zap disk storage -- deletes all objects in disk cache.\n\
       -C        Do not catch fatal signals.\n\
       -R        Do not set REUSEADDR on port.\n\
       -f file   Use given config-file instead of /etc/cached.conf.\n\
       -d level  Use given debug-level, prints messages to stderr.\n\
       -a port	 Specify ASCII port number (default: 3128).\n\
       -p port	 Specify TCP port number (default: 3129).\n\
       -u port	 Specify UDP port number (default: 3130).\n");

	    exit(1);
	    break;
	}

    if (!neighbors) {
	neighbors_create(N_UPPER_CACHES);
	++neighbors;
    };

    if (catch_signals) {
	signal(SIGINT, shut_down);
	signal(SIGTERM, shut_down);
	signal(SIGSEGV, death);
	signal(SIGBUS, deathb);
    }
    signal(SIGPIPE, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    parseConfigFile(config_file);
    if (!debug_level_overwrite) {
	debug_level = cachelog_flag;
    }
    /* to toggle debugging */
#ifdef SIGUSR1
    signal(SIGUSR1, raise_debug_lvl);
#endif
#ifdef SIGUSR2
    signal(SIGUSR2, reset_debug_lvl);
#endif

#ifdef NO_LOGGING
    _db_init("cached", 0, cache_log_file);
#else
    _db_init("cached", debug_level, cache_log_file);
#endif

    debug(0, "Starting Harvest Cache (version %s)...\n", HARVEST_VERSION);

    /* init ipcache */
    ipcache_init();

    /* init neighbors */
    neighbors_init();


#ifdef MALLOC_DBG
    malloc_debug(malloc_debug_level);
#endif

    /* Used to store a stack of empty message packets to save time later */
    init_stack(&free_pkt_list);


    if (prompt) {
	while (1) {
	    asciiProcessInput(0, 1);
	}
    } else {
	theAsciiConnection = comm_open(COMM_NONBLOCKING, asciiPort, 0);
	if (theAsciiConnection < 0) {
	    fatal("Cannot open ascii Port\n");
	}
	comm_listen(theAsciiConnection);
	comm_set_select_handler(theAsciiConnection, COMM_SELECT_READ, asciiHandleConn, 0);

	if (!httpd_accel_mode) {
#ifdef KEEP_BINARY_CONN
	    theBinaryConnection = comm_open(COMM_NONBLOCKING, binaryPort, 0);

	    if (theBinaryConnection < 0) {
		fatal("Cannot open Binary Port\n");
	    }
	    comm_listen(theBinaryConnection);
	    comm_set_select_handler(theBinaryConnection, COMM_SELECT_READ, icpHandleTcp, 0);
#endif
	    theUdpConnection = comm_open(COMM_NONBLOCKING | COMM_DGRAM, udpPort, 0);
	    if (theUdpConnection < 0) {
		fatal("Cannot open UDP Port\n");
	    }
	    comm_set_select_handler(theUdpConnection, COMM_SELECT_READ, icpHandleUdp, 0);
	}
	icpSaveAsciiPort(asciiPort);	/*Save to find what type error msg to send */
	/* Save the port and host info for building icp urls */
	icpSetHostUrl(comm_hostname(), asciiPort, udpPort, binaryPort);

	/* Now that the fd's are open, initialize neighbor connections */
	if (!httpd_accel_mode) {	/* only if a normal cached */
	    neighbors_open(theUdpConnection);
	}
	/* do suid checking here */
	check_suid();

	/* module initialization */
	disk_init();
	stat_init(&CacheInfo, cache_access_log_file);
	storeInit(cache_dir);

#ifdef WRITE_PID_FILE
/* Tries to write the pid to cached.pid in the same directory as cached.conf */
	{
	    FILE *fp;
	    char pidfn[4096], *s;

	    memset(pidfn, '\0', sizeof(char) * 4096);
	    strcpy(pidfn, config_file);
	    s = strrchr(pidfn, '/');
	    if (s == NULL) {
		strcpy(pidfn, "/etc/cached.pid");
	    } else {
		strcpy(s, "/cached.pid");
	    }
	    fp = fopen(pidfn, "w");
	    if (fp != NULL) {
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
	    }
	}
#endif

	/* main loop */
	next_cleaning = time(0L) + getCleanRate();
	while (1) {
	    switch (comm_select((long) 60, (long) 0, next_cleaning)) {
	    case COMM_OK:
		/* do nothing */
		break;
	    case COMM_ERROR:
		errcount++;
		debug(0, "Select loop Error. Retry. %d\n", errcount);
		if (errcount == 10)
		    fatal_dump("Select Loop failed.!\n");
		break;
	    case COMM_TIMEOUT:
		/* this happens after 1 minute of idle time, or
		   when next_cleaning has arrived */
		if (1 == do_mail) {
		    do_mail = 2;
		    send_mail_to_ucboulder(asciiPort);
		}
		/* garbage collection */
		if (cached_curtime >= next_cleaning) {
		    debug(1, "Performing a garbage collection...\n");
		    storePurgeOld();
		    next_cleaning = cached_curtime + getCleanRate();
		}
		/* house keeping */
		kill_zombie();
		break;
	    default:
		debug(0, "MAIN: Internal error -- this should never happen.\n");
		break;
	    }
	}
    }
    exit(0);
}

void raise_debug_lvl()
{
    extern int _db_level;
    _db_level = 10;

#if defined(_HARVEST_SYSV_) && defined(SIGUSR1)
    signal(SIGUSR1, raise_debug_lvl);
#endif
}

void reset_debug_lvl()
{
    extern int _db_level;
    _db_level = debug_level;

#if defined(_HARVEST_SYSV_) && defined(SIGUSR2)
    signal(SIGUSR2, reset_debug_lvl);
#endif
}
