static char rcsid[] = "$Id: mime.c,v 1.13 1995/08/04 00:01:03 hardy Exp $";
/* 
 *  File:         mime.c
 *  Description:  Mime Module 
 *  Author:       Anawat Chankhunthod, USC
 *  Created:      Mon Dec 12 16:09:40 PST 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "ansihelp.h"		/* goes first */
#include "debug.h"
#include "mime.h"
#include "mime_table.h"

void mime_process(mime)
char *mime;
{
    char *ptr;

    if (!mime)
	return;

    if ((ptr = strstr(mime, "If-Modified-Since:")) != NULL) {
	/* nuke If-Modified-Since */
	char *eol;
	if ((eol = strstr(ptr, "\r\n")) == NULL &&
	    (eol = strstr(ptr, "\n")) == NULL)
	    return;		/* rare */
	eol += 2;
	strcpy(ptr, eol);
	debug(5, "mime_process: If-Modified-Since is eliminated.\n");
	debug(6, "Remain MIME is: %s\n", mime);
    }
}

int mime_refresh_request(mime)
char *mime;
{
    if (strstr(mime, "no-cache"))
	return 1;

    return 0;
}

ext_table_entry *
 mime_ext_to_type(extension)
char *extension;
{
    int i, low, high, comp;
    char ext[16], *cp;

    if (!extension || strlen(extension) >= (sizeof(ext) - 1))
	return NULL;
    strcpy(ext, extension);
    for (cp = ext; *cp; cp++)
	if (isupper(*cp))
	    *cp = tolower(*cp);
    low = 0;
    high = EXT_TABLE_LEN - 1;
    while (low <= high) {
	i = (low + high) / 2;
	if ((comp = strcmp(ext, ext_mime_table[i].name)) == 0)
	    return &ext_mime_table[i];
	if (comp > 0)
	    low = i + 1;
	else
	    high = i - 1;
    }
    return NULL;
}

/*
 *  mk_mime_hdr - Generates a MIME header using the given parameters.
 *  You can call mk_mime_hdr with a 'lmt = time(NULL) - ttl' to
 *  generate a fake Last-Modified-Time for the header.
 *  'ttl' is the number of seconds relative to the current time
 *  that the object is valid.
 *
 *  Returns the MIME header in the provided 'result' buffer, and
 *  returns non-zero on error, or 0 on success.
 */
int mk_mime_hdr(result, ttl, size, lmt, type)
char *result, *type;
int size;
time_t ttl, lmt;
{
    extern char *mkrfc850( /* time_t *t */ );
    time_t t = time(NULL), expiretime;
    char date[100];
    char expire[100];
    char last_modified_time[100];

    if (result == NULL)
	return 1;

    expiretime = t + ttl;

    date[0] = expire[0] = last_modified_time[0] = result[0] = '\0';
    strncpy(date, mkrfc850(&t), 100);
    strncpy(expire, mkrfc850(&expiretime), 100);
    strncpy(last_modified_time, mkrfc850(&lmt), 100);

    sprintf(result, "Content-Type: %s\r\nContent-Size: %d\r\nDate: %s\r\nExpires: %s\r\nLast-Modified-Time: %s\r\n", type, size, date, expire, last_modified_time);
    return 0;
}
