static char rcsid[] = "$Id: neighbors.c,v 1.35 1995/08/21 20:55:47 hardy Exp $";
/* 
 *  File:         neighbors.c
 *  Description:  Keeps track of other caches and cache clients
 *  Author:       Peter Danzig, Chuck and Anawat Chankhunthod USC
 *                Dean Povey, University of Queensland
 *  Created:      Mon May 16 14:18:36 PDT 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include "comm.h"
#include "store.h"
#include "icp.h"
#include "proto.h"
#include "neighbors.h"
#include "ipcache.h"
#include "cache_cf.h"

extern int theUdpConnection;

typedef struct {
    int max;
    int n;
    edge *edges;
    int fd;
} neighbors;

neighbors *friends;

struct neighbor_cf {
    char *host;
    char *type;
    int ascii_port;
    int udp_port;
    int tcp_port;
    struct neighbor_cf *next;
};

struct neighbor_cf *Neighbor_cf = NULL;

static icp_common_t echo_hdr;
static short echo_port;
FILE *cache_hierarchy_log;

#define SIBLING_EDGE   0
#define PARENT_EDGE    1
#define ELSEWHERE_EDGE 2 /* XXX Needs work */

edge *
 whichEdge(header, from)
icp_common_t *header;
struct sockaddr_in *from;
{
    int i;

    for (i = 0; i < friends->n; ++i)
	if (ipcache_check(friends->edges[i].host, &(from->sin_addr), 1) == 0) {
	    return (&friends->edges[i]);
	}
    return (NULL);
}

void neighbors_install(host, type, ascii_port, udp_port, tcp_port)
char *host;
char *type;
int ascii_port;
int udp_port;
int tcp_port;
{
    int i;

    debug(1, "installing a %s\n", type);

    /* Ge  a free edge entry */
    for (i = 0; (friends->edges[i].host);)
	++i;

    if (friends->n != i) {
	fprintf(stderr, "Looks like edges are out of sync\n");
    }
    friends->edges[i].ascii_port = ascii_port;
    friends->edges[i].udp_port = udp_port;
    friends->edges[i].tcp_port = tcp_port;

	/* XXX Need to add support for "unknown" type */
    if (!strncmp(type, "parent", 8))
	friends->edges[i].type = PARENT_EDGE;
    else
	friends->edges[i].type = SIBLING_EDGE;

    friends->edges[i].host = strdup(host);

    friends->n++;

}

void neighbors_open(fd)
int fd;
{
    int i;
    struct sockaddr_in our_socket_name;
    struct sockaddr_in *ap;
    struct in_addr *tmp_addr;
    struct in_addr *getAddress();
    int sock_name_length = sizeof(our_socket_name);

    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "getsockname failed??");
    }
    friends->fd = fd;

    /* open log file */
    if (cache_hierarchy_log_flag) {
	if (!(cache_hierarchy_log = fopen(cache_hierarchy_log_file, "a+"))) {
	    debug(1, "Cannot open hierachical log file\n Disable hierachical logging\n");
	    cache_hierarchy_log_flag = 0;
	}
    }
    /* Prepare neighbor connections, one at a time */
    for (i = 0; i < friends->n; ++i) {

	tmp_addr = getAddress(friends->edges[i].host);
	if (tmp_addr == (struct in_addr *) NULL) {
	    debug(1, "getAddress(%s) Failed; skipping this neighbor.\n",
		  friends->edges[i].host);
	    continue;
	}
	friends->edges[i].rtt = 1000;

	/* Prepare query packet for future use */
	friends->edges[i].header.opcode = ICP_OP_QUERY;
	friends->edges[i].header.version = ICP_VERSION_CURRENT;
	friends->edges[i].header.length = 0;
	friends->edges[i].header.reqnum = 0;
	memset(friends->edges[i].header.auth, '\0', sizeof(u_num32) * ICP_AUTH_SIZE);
	friends->edges[i].header.shostid = our_socket_name.sin_addr.s_addr; 
	friends->edges[i].header.offset = sizeof(icp_common_t);
	friends->edges[i].header.nr_options = 0;

	ap = &friends->edges[i].in_addr;
	memset(ap, '\0', sizeof(struct sockaddr_in));
	ap->sin_family = AF_INET;
	memcpy(&ap->sin_addr, tmp_addr, sizeof(ap->sin_addr));
	ap->sin_port = htons(friends->edges[i].udp_port);

	if (friends->edges[i].type == 1) {
	    debug(3, "parent_install: host %s addr %s port %d\n",
		  friends->edges[i].host, inet_ntoa(ap->sin_addr),
		  friends->edges[i].udp_port);
	} else {
	    debug(3, "neighbor_install: host %s addr %s port %d\n",
		  friends->edges[i].host, inet_ntoa(ap->sin_addr),
		  friends->edges[i].udp_port);
	}

	/* do this only the first time thru */
	if (0 == echo_hdr.opcode) {
	    struct servent *sep;

	    echo_hdr.opcode = ICP_OP_SECHO;
	    echo_hdr.version = ICP_VERSION_CURRENT;
	    echo_hdr.length = 0;
	    echo_hdr.reqnum = 0;
	    memset(echo_hdr.auth, '\0', sizeof(u_num32) * ICP_AUTH_SIZE);
	    echo_hdr.shostid = our_socket_name.sin_addr.s_addr;

	    sep = getservbyname("echo", "udp");
	    echo_port = sep ? sep->s_port : 7;
	}
    }
}

neighbors *
 neighbors_create(max)
int max;
{
    neighbors *f;

    f = (neighbors *) xcalloc(sizeof(neighbors), 1);
    f->max = max;
    f->n = 0;
    f->edges = (edge *) xcalloc(sizeof(neighbors), max + 1);
    friends = f;

    return (f);
}


int neighborsUdpPing(fdunused, url, entry)
int fdunused;
char *url;
StoreEntry *entry;
{
    int i;

    entry->e_pings_n_pings = 0;
    entry->e_pings_n_acks = 0;
    entry->e_pings_first_miss = NULL;

    for (i = 0; (i < friends->n); ++i) {

	if (((friends->edges[i].type == 0) && !(entry->flag & REFRESH_REQUEST)) ||
	    (friends->edges[i].type == 1)) {
	    debug(4, "neighborsUdpPing - url: %s to cache %s \n",
		  url, inet_ntoa(friends->edges[i].in_addr.sin_addr));

	    friends->edges[i].header.reqnum++;

	    if (friends->edges[i].udp_port == echo_port) {
		debug(4, "neighborsUdpPing: Looks like a dumb cache, send ECHO ping\n");
		icpUdpSend(friends->fd, url, &echo_hdr,
			   &friends->edges[i].in_addr, ICP_OP_SECHO, 0);
	    } else {
		icpUdpSend(friends->fd, url, &friends->edges[i].header,
			   &friends->edges[i].in_addr, ICP_OP_QUERY, 0);
	    }
	    entry->e_pings_n_pings++;
	}
    }

    /* only ping if we have neighbors */
    if (echo_hdr.opcode) {
	struct hostent *hep;
	struct sockaddr_in to_addr;
	char proto[1025], host[1025], *colon;
	int t;

	memset(&proto, 0, 1025);
	memset(&host, 0, 1025);

	t = sscanf(url, "%[^:]://%[^/]", proto, host);
	colon = strchr(host, ':');	/* host:port */
	if (colon)
	    *colon = '\0';

	if ((t == 2) && (hep = ipcache_gethostbyname(host))) {

	    if (getSourcePing()) {
		debug(6, "neighborsUdpPing: Send to original host\n");
		debug(6, "neighborsUdpPing: url=%s, proto=%s, host=%s, t=%d\n",
		      url, proto, host, t);
		to_addr.sin_family = AF_INET;
		memcpy(&to_addr.sin_addr, hep->h_addr, hep->h_length);
		to_addr.sin_port = echo_port;
		echo_hdr.reqnum = time(0);

		debug(6, "neighborsUdpPing - url: %s to url-host %s \n",
		      url, inet_ntoa(to_addr.sin_addr));

		/* send to original site */
		icpUdpSend(friends->fd, url, &echo_hdr, &to_addr, ICP_OP_SECHO, 0);
	    } else {
		debug(6, "neighborsUdpPing: Source Ping is disabled.\n");
	    }
	} else {
	    debug(6, "neighborsUdpPing - url %s: unknown/bad host!\n", url);
	}
    }
    return (i);
}

int neighborsUdpQuery(fdunused, url, buf) 
int fdunused;
char *url;
caddr_t buf;
{
    int i;
	icp_common_t *headerp=(icp_common_t *) buf;

    for (i = 0; (i < friends->n); ++i) {

	/* 
	 * NB: When doing recursive queries, we only need to query the
	 * parents.  If our siblings have something, then our parent will know
	 * about it.  This may or may not be bad for load balancing XXX
	 */
	if (friends->edges[i].type == PARENT_EDGE) {
	    debug(4, "neighborsUdpQuery - url: %s to cache %s \n",
		  url, inet_ntoa(friends->edges[i].in_addr.sin_addr));

	    if (friends->edges[i].udp_port == echo_port) {
		debug(4, "neighborsUdpQuery: Looks like a dumb cache, do nothing\n");
	    } else {
		icpUdpSend(friends->fd, url, headerp,
			   &friends->edges[i].in_addr, ICP_OP_QUERY, ICP_OPT_RECURSE);
	    }
	}
    }

    return (i);
}

/* XXX I can't bring myself to spell advertise "advertize" ;) (DGP) */
int neighborsUdpAdvertise(buf, buf_len) 
caddr_t buf;
int buf_len;
{
    int i;

    for (i = 0; (i < friends->n); ++i) {

	/* 
	 * NB: We only advertise to our parents
	 */
	if (friends->edges[i].type == PARENT_EDGE) {

	    if (friends->edges[i].udp_port == echo_port) {
	debug(4, "neighborsUdpAdvertise: Looks like a dumb cache, do nothing\n");
	    } else {
	    debug(4, "neighborsUdpAdvertise: Sending advertisment to to cache %s\n",
		  inet_ntoa(friends->edges[i].in_addr.sin_addr));

		icpUdpForward(friends->fd, buf, buf_len, &friends->edges[i].in_addr);
	    }
	}
	}

    return COMM_OK;
}


/* I should attach these records to the entry.  We take the first
   hit we get our wait until everyone misses.  The timeout handler
   call needs to nip this shopping list or call one of the misses.

   If a hit process is already started, then sobeit
 */
void neighborsUdpAck(fdunused, url, buf, from, entry)
int fdunused;
char *url;
caddr_t buf;
struct sockaddr_in *from;
StoreEntry *entry;
{
    edge *e;
	icp_common_t *header = (icp_common_t *)buf;

    debug(6,
	  "neighborsUdpAck: url=%s (%d chars), header=0x%x, from=0x%x, ent=0x%x\n",
	  url, strlen(url), header, from, entry);
    debug(6, "     hdr: opcode=%d, ver=%d, shostid=%x, len=%d, rn=0x%x\n",
	  header->opcode, header->version, header->shostid,
	  header->length, header->reqnum);
    debug(6, "     from: fam=%d, port=%d, addr=0x%x\n",
	  from->sin_family, from->sin_port, from->sin_addr.s_addr);

    /* check if someone is already fetching it */
    if (BIT_TEST(entry->flag, REQ_DISPATCHED) || (entry->ping_status != WAITING)) {
	if (entry->ping_status == DONE) {
	    debug(6, "There is already a cache/source dispatched for this object.\n");
	} else {
	    debug(6, "It is already timeout for pinging.\n");
	}
	return;
    }
    /* look up for neighbor/parent entry */
    e = whichEdge(header, from);


    debug(6, "neighborsUdpAck - url: %s to us %s \n",
	  url, e ? inet_ntoa(e->in_addr.sin_addr) : "url-host");

    /* Is it an echo from source or a dumb cache ? */
    if (header->opcode == ICP_OP_SECHO) {
	/* receive ping back from source or from non-cached cache */

	if (e) {
	    debug(6, "Got SECHO from non-cached cache:%s\n",
		  inet_ntoa(e->in_addr.sin_addr));
	    debug(6, "Might as well take advantage of it.\n");
	    if (cache_hierarchy_log_flag) {
		if (e->type == 0) {
		    fprintf(cache_hierarchy_log, "%s %s %s\n", entry->url,
			    HIER_NEIGHBOR, e->host);
		} else {
		    fprintf(cache_hierarchy_log, "%s %s %s\n", entry->url,
			    HIER_PARENT_HIT, e->host);
		}
		fflush(cache_hierarchy_log);
	    }
	    BIT_SET(entry->flag, REQ_DISPATCHED);
	    entry->ping_status = DONE;
	    getFromSource(0, entry, e);
	} else {
	    /* if we reach here, source is the one has the fastest respond. */
	    /* fetch directly from source */
	    debug(6, "Get SECHO first: Source is the fastest site to get object.\n");
	    debug(6, "Go get from the source directly.\n");

	    if (cache_hierarchy_log_flag) {
		fprintf(cache_hierarchy_log, "%s %s\n", entry->url, HIER_DIRECT);
		fflush(cache_hierarchy_log);
	    }
	    BIT_SET(entry->flag, REQ_DISPATCHED);
	    entry->ping_status = DONE;
	    getFromSource(0, entry, 0);
	}
	return;
    }
    if (header->opcode == ICP_OP_HIT) {
	/* If an edge is not found, count it as a MISS message. */
	if (!e) {
	    /* count it as a MISS message */
	    entry->e_pings_n_acks++;
	    return;
	}

	/* GOT a HIT here */
	if (cache_hierarchy_log_flag) {
	    if (e->type == 0) {
		fprintf(cache_hierarchy_log, "%s %s %s\n", entry->url, HIER_NEIGHBOR, e->host);
	    } else {
		fprintf(cache_hierarchy_log, "%s %s %s\n", entry->url, HIER_PARENT_HIT, e->host);
	    }
	    fflush(cache_hierarchy_log);
	}
	BIT_SET(entry->flag, REQ_DISPATCHED);
	entry->ping_status = DONE;

#ifdef SUPPORT_ICPV1_4
	/* Check for an old version of the ICP */
	if (header->version < ICP_VERSION_CURRENT)
	{
		debug(6, "HIT: Getting %s from host: %s\n", entry->url, e->host);
		getFromSource(0, entry, e);
		return;
	}
	else
#endif
	if (header->version != ICP_VERSION_CURRENT)
	{
		/* XXX !!! Bitch.  (Need to implement this properly) */
		debug(1, "Received invalid ICP packet from %s\n", e->host);
		return;
	}
	else
	{
		/* Process the pointers and decide which source to slurp from */
		getFromPointers(0, entry, buf);
		return;
	}
    } else if (header->opcode == ICP_OP_MISS) {
	entry->e_pings_n_acks++;

	if ((entry->e_pings_first_miss == NULL) && e && e->type == 1) {
	    entry->e_pings_first_miss = e;

	}
	if (entry->e_pings_n_acks == entry->e_pings_n_pings) {
	    BIT_SET(entry->flag, REQ_DISPATCHED);
	    entry->ping_status = DONE;
	    debug(6, "Receive MISSes from all neighbors and parents\n");
	    if (entry->e_pings_first_miss) {
		debug(6, "Getting %s from first_miss parent: %s\n", entry->url,
		      entry->e_pings_first_miss->host);

		if (cache_hierarchy_log_flag) {
		    fprintf(cache_hierarchy_log, "%s %s %s\n", entry->url, HIER_PARENT_MISS, entry->e_pings_first_miss->host);
		    fflush(cache_hierarchy_log);
		}
	    } else {
		/* I'm orphan. :( */
		debug(6, "There is no parent. Getting %s from original site\n", entry->url);
		if (cache_hierarchy_log_flag) {
		    fprintf(cache_hierarchy_log, "%s %s\n", entry->url, HIER_NO_PARENT);
		    fflush(cache_hierarchy_log);
		}
	    }

	    getFromSource(0, entry, entry->e_pings_first_miss);
	    return;
	}
    } else {
    }
}

void neighbors_cf_add(host, type, ascii_port, udp_port, tcp_port)
char *host;
char *type;
int ascii_port;
int udp_port;
int tcp_port;
{
    struct neighbor_cf *t, *u;

    t = (struct neighbor_cf *) xmalloc(sizeof(struct neighbor_cf));
    t->host = xstrdup(host);
    t->type = xstrdup(type);
    t->ascii_port = ascii_port;
    t->udp_port = udp_port;
    t->tcp_port = tcp_port;
    t->next = (struct neighbor_cf *) NULL;

    if (Neighbor_cf == (struct neighbor_cf *) NULL) {
	Neighbor_cf = t;
    } else {
	for (u = Neighbor_cf; u->next; u = u->next);
	u->next = t;
    }
}

void neighbors_init()
{
    struct neighbor_cf *t, *next;

    for (t = Neighbor_cf; t; t = next) {
	next = t->next;
	if (strncmp(t->host, comm_hostname(), MAXHOSTNAMELEN))
	    neighbors_install(t->host, t->type,
			      t->ascii_port, t->udp_port, t->tcp_port);
	xfree(t->host);
	xfree(t->type);
	xfree(t);
    }
}


/*
 * neighbors_getEdge
 *
 * Given a hostname and ports find the corresponding edge, return
 * 0 if the edge does not match any.
 */
edge *
neighbors_get(host, ascii_port, tcp_port, udp_port)
char *host;
int ascii_port;
int tcp_port;
int udp_port;
{
	int i;

	/* Search edge entries */
    for (i = 0; i < friends->n; ++i){
    if (ipcache_checkname(host, friends->edges[i].host) == 0 &&
	    friends->edges[i].ascii_port == ascii_port &&
	    friends->edges[i].tcp_port == tcp_port &&
	    friends->edges[i].udp_port == udp_port)
        return (&friends->edges[i]);
    }
    return (NULL);

}
