static char rcsid[] = "$Id: proto.c,v 1.47 1995/08/21 20:55:51 hardy Exp $";
/*
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <string.h>
#include <stdlib.h>

#include "config.h"		/* goes first */
#include "ansihelp.h"		/* goes second */
#include "debug.h"
#include "comm.h"
#include "proto.h"		/* for  caddr_t */
#include "neighbors.h"
#include "store.h"
#include "cache_cf.h"
#include "icp.h"		/* for icpStateData - must succeed store.h */

typedef struct _protodispatch_data {
    int fd;
    char *url;
    StoreEntry *entry;
} protodispatch_data;


extern char *tmp_error_buf;

int getFromPointers _PARAMS((int fd, StoreEntry *entry, caddr_t buf));
int getFromSource _PARAMS((int fd, StoreEntry * entry, edge * e));
int getFromSourceDirect _PARAMS((int fd, StoreEntry * entry));
char *choosePointer _PARAMS((caddr_t buf, edge **e));
char *parseICPUrl _PARAMS((char *url, char *host, int *ascii_port, 
	int *udp_port, int *tcp_port));

extern int httpd_accel_mode;

/* return 1 for cachable url
   return 0 for uncachable url */
int proto_cachable(url, type, mime_hdr)
char *url;
char *type;
char *mime_hdr;
{
    if (url == (char *) NULL)
	return 0;

    if (!strncasecmp(url, "http://", 7))
	return httpCachable(url, type, mime_hdr);
    if (!strncasecmp(url, "ftp://", 6))
	return ftpCachable(url, type, mime_hdr);
    if (!strncasecmp(url, "gopher://", 9))
	return gopherCachable(url, type, mime_hdr);
#ifdef USE_WAIS_RELAY
    if (!strncasecmp(url, "wais://", 7))
	return 0;
#endif

    return 1;
}

/* return default ttl for that URL */
int proto_defaultTTL(url)
char *url;
{
    if (url == (char *) NULL)
	return getHttpTTL();
    if (strncasecmp(url, "http://", 7) == 0)
	return getHttpTTL();
    if (strncasecmp(url, "gopher://", 9) == 0)
	return getGopherTTL();
    if (strncasecmp(url, "ftp://", 6) == 0)
	return getFtpTTL();
    return getHttpTTL();
}
/* called when DNS lookup is done by ipcache. */

int protoDispatchDNSHandle(fd, unused_hp, data)
int fd;
struct hostent *unused_hp;
protodispatch_data *data;
{
	/*
	 * XXX 
	 * If there are no neigbours, then we get the object from the source,
	 * otherwise we send a request to the neighbors, and install a timeout 
	 * handler (This is done in neigborsUdpPing).  If no neighbors respond 
	 * then we will get from the source (getSourceDirect). 
	 *
	 * If a neigbour responds then they will be processed in icpHandleUdp
     * and then in neighborsUdpAck
	 */

    if (proto_cachable(data->url, HTTP_OPS[data->entry->type_id], data->entry->mime_hdr)) {
	if (neighborsUdpPing(data->fd, data->url, data->entry)) {
	    comm_set_select_handler_plus_timeout(data->fd, COMM_SELECT_TIMEOUT,
						 (PF) getFromSourceDirect,
						 (caddr_t) data->entry,
						 (time_t) 2);
	    safe_free(data);
	    return 0;
	}
    }
    /* if not, go get from source */
    data->entry->ping_status = NOPING;	/* for debugging sake */
    getFromSource(data->fd, data->entry, 0);

    safe_free(data);
    return 0;
}


int protoDispatch(fd, url, entry)
int fd;
char *url;
StoreEntry *entry;
{
    protodispatch_data *data = (protodispatch_data *) xcalloc(1, sizeof(protodispatch_data));

    debug(5, "protoDispatch: %s URL: %s\n", HTTP_OPS[entry->type_id], url);

    /* Start retrieval process. */
    if (strncasecmp(url, "cache_object:", 13) == 0) {
	safe_free(data);
	return objcacheStart(fd, url, entry);
    } else if (httpd_accel_mode &&
	  strncmp(url, httpd_accel_prefix, strlen(httpd_accel_prefix)) &&
	       !httpd_accel_with_proxy) {
	/* allow only HTTP_ACCEL query only */
	safe_free(data);
	return protoNotImplemented(fd, url, entry);
    } else {
	int t;
	char garbage[1024], host[1024];
	char *colon;

	entry->ping_status = WAITING;

	memset(host, '\0', 1024);
	t = sscanf(url, "%[^:]://%[^/]", garbage, host);
	colon = strchr(host, ':');	/* host:port */
	if (colon)
	    *colon = '\0';

	data->fd = fd;
	data->url = url;
	data->entry = entry;

	/* start name look up here */
	ipcache_nbgethostbyname(host, fd, protoDispatchDNSHandle, data);
    }
    return 0;
}

protoDispatchRecursiveQuery(fd, url, buf)
int fd;
char *url;
caddr_t buf;
{
	/* 
	 * NB: We don't check if the url is cacheable as ICPv2.0 allows
	 * you to specify pointers to replicated services, this means that
	 * cgi-bin scripts are in a sense cached.
	 */
    if (neighborsUdpQuery(fd, url, buf)) 
	{
		/* 
		 * XXX This should use a smoothed RTT for timeout rather than a 
		 * hard coded value
		 * XXXX Currently 600 secs for testing
		 */
		comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT,
        	(PF) icpUdpRecursiveTimeout, (caddr_t) url, (time_t) 600);
        return 0;
	}

	/* 
	 * XXX Not exactly the same event but if there are no neighbors this
	 * has the same effect
	 */
	icpUdpRecursiveTimeout(fd, url);

	return 0;
}

/* Use to undispatch a particular url/fd from DNS pending list */
/* I have it here because the code that understand protocol/url */
/* should be here. */
int protoUndispatch(fd, url)
int fd;
char *url;
{

    debug(5, "protoUndispatch fd: %d URL: %s\n", fd, url);

    /* Start retrieval process. */
    if (strncasecmp(url, "cache_object:", 13) == 0) {
	return 0;
    } else {
	int t;
	char garbage[1024], host[1024];
	char *colon;


	memset(host, '\0', 1024);
	t = sscanf(url, "%[^:]://%[^/]", garbage, host);
	colon = strchr(host, ':');	/* host:port */
	if (colon)
	    *colon = '\0';

	/* clean up DNS pending list for this name/fd look up here */
	ipcache_unregister(host, fd);
    }
    return 0;
}

int getFromSourceDirect(fd, entry)
int fd;
StoreEntry *entry;
{
    debug(5, "in getFromSourceDirect\n");

    entry->ping_status = TIMEOUT;
    return getFromSource(fd, entry, entry->e_pings_first_miss);
}

int getFromSource(fd, entry, e)
int fd;
StoreEntry *entry;
edge *e;
{
    char *url = entry->url;
    char *type = HTTP_OPS[entry->type_id];
    char *mime_hdr = entry->mime_hdr;

    /* If this is called from our neighbor detection,
       then we have to reset the signal handler.  We
       probably need to check for a race here on a previous
       close of the client connection
     */
    if (!fd)
	fd = entry->pending[0]->fd;

    if (e) {
	debug(5, "in getFromSource URL: %s %s from %s\n", type, url, e->host);
	debug(10, "    mime_hdr: %s\n", mime_hdr);
    }
    comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT,
					 (PF) 0, (caddr_t) 0, (time_t) 0);

    if (e) {
	return proxyhttpStart(e, url, entry);
    } else if (strncasecmp(url, "http://", 7) == 0) {
	return httpStart(fd, url, type, mime_hdr, entry);
    } else if (strncasecmp(url, "gopher://", 9) == 0) {
	return gopherStart(fd, url, entry);
    } else if (strncasecmp(url, "news://", 7) == 0) {
	return protoNotImplemented(fd, url, entry);
    } else if (strncasecmp(url, "file://", 7) == 0) {
#ifndef NO_FTP_FOR_FILE
	return ftpStart(fd, url, entry);
#else
	return protoNotImplemented(fd, url, entry);
#endif
    } else if (strncasecmp(url, "ftp://", 6) == 0) {
	return ftpStart(fd, url, entry);
    } else if (strncasecmp(url, "dht://", 6) == 0) {
	return protoNotImplemented(fd, url, entry);
    } else {
	return protoNotImplemented(fd, url, entry);
    }
    return 0;
}


int 
getFromPointers(fd, entry, buf)
int fd;
StoreEntry *entry;
caddr_t buf;
{
    char *url = 0;
    char *type = HTTP_OPS[entry->type_id];
    char *mime_hdr = entry->mime_hdr;
	edge *e=0;

    /* XXX ??? If this is called from our neighbor detection,
       then we have to reset the signal handler.  We
       probably need to check for a race here on a previous
       close of the client connection
     */
    if (!fd)
	fd = entry->pending[0]->fd;

	/* Get the URL to retreive and the host to retreive it from
	 * NB: The HIT packet we received may have had multiple entries for
	 * the object we are after.  If this is the case then choosePointer will
	 * pick one at random.  If the edge returned is not null, then the pointer
	 * pointed to another cached-host.  If the url is null, then something went
	 * wrong
     */
	url = choosePointer(buf, &e);
	if (!url)
	{
		/* XXX Log an error */
		debug(1, "error choosing a pointer from the HIT packet from %s\n", 
			e->host);
		return 0;
	}

	if (!e)
	{
		/* XXX - Register a hit */
		debug(4, "HIT: %s at %s\n", entry->url, url);
	}
	else
	{
		/* XXX - Register a hit */
		debug(4, "HIT: %s at %s\n", entry->url, e->host);
	}

	/* Retreive the object */
    if (e) {
	debug(5, "in getFromPointer URL: %s %s from %s\n", type, url, e->host);
	debug(10, "    mime_hdr: %s\n", mime_hdr);
    }
    comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT,
					 (PF) 0, (caddr_t) 0, (time_t) 0);

    if (e) {
	return proxyhttpStart(e, url, entry);
    } else if (strncasecmp(url, "http://", 7) == 0) {
	return httpStart(fd, url, type, mime_hdr, entry);
    } else if (strncasecmp(url, "gopher://", 9) == 0) {
	return gopherStart(fd, url, entry);
    } else if (strncasecmp(url, "news://", 7) == 0) {
	return protoNotImplemented(fd, url, entry);
    } else if (strncasecmp(url, "file://", 7) == 0) {
#ifndef NO_FTP_FOR_FILE
	return ftpStart(fd, url, entry);
#else
	return protoNotImplemented(fd, url, entry);
#endif
    } else if (strncasecmp(url, "ftp://", 6) == 0) {
	return ftpStart(fd, url, entry);
    } else if (strncasecmp(url, "dht://", 6) == 0) {
	return protoNotImplemented(fd, url, entry);
    } else {
	return protoNotImplemented(fd, url, entry);
    }
    return 0;
}

int protoNotImplemented(fd, url, entry)
int fd;
char *url;
StoreEntry *entry;
{
    char buf[256];

    debug(2, "protoNotImplemented: Cannot retrieve URL: %s\n", url);

    buf[0] = '\0';
    if (httpd_accel_mode)
	strcpy(buf, "cached is running in HTTPD accelerator mode, so it does not allow the normal URL syntax.");
    else
	sprintf(buf, "Your URL may be incorrect: '%s'\n", url);

    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
	    entry->url,
	    entry->url,
	    "CACHE-PROTO",
	    409,
	    "Unsupported protocol",
	    buf,
	    HARVEST_VERSION,
	    comm_hostname());
    storeAbort(entry, tmp_error_buf);
    return 0;
}


/*
 * choosePointer
 *
 * Given a HIT returned from one of our neighbors, parse the packet and
 * select a pointer to retreive the object from. 
 * 
 * XXX - The pointer is currently chosen at random
 */
char *choosePointer(buf, e)
caddr_t buf;
edge **e;
{
	icp_common_t *header = (icp_common_t *)buf;
	/* icp_hit_list_t *hit_list=0; */
	unsigned short nr_ptrs;
	unsigned short size;
	caddr_t tmp;
	int i, ptr=0;
	int ascii_port=0;
	int tcp_port=0;
	int udp_port=0;
	char *url=0; 
	char *ret=0;
	char *host; /* XXX */

	/* Get the offset of the data */
	tmp = buf + header->offset;

	/* Get the number of pointers from the packet */
	memcpy(&nr_ptrs, tmp, sizeof(unsigned short));
	nr_ptrs = ntohs(nr_ptrs);
	tmp += sizeof(unsigned short);

	/* 
	 * XXX Pick a pointer to use - a random guess at this stage is probably 
	 * as good as any 
	 */
	ptr=rand()%nr_ptrs;

	/* 
	 * Get the size of the source URL - and skip over it 
	 * and the timestamp of the first pointer entry
	 */
	memcpy(&size, tmp, sizeof(unsigned short));
    size = ntohs(size);
    tmp += sizeof(unsigned short) + size + sizeof(u_num32);

	/* Get the pointer */
	for (i=0; i<ptr; i++)
	{
		/* 
		 * Get the size of the next pointer URL - and skip over it 
		 * + the timestamp of the next pointer entry
		 */
		memcpy(&size, tmp, sizeof(unsigned short));
    	size = ntohs(size);
    	tmp += sizeof(unsigned short) + size + sizeof(u_num32);
	}

	/* Get the size of the pointer to return */
	memcpy(&size, tmp, sizeof(unsigned short));
	size = ntohs(size);
    tmp += sizeof(unsigned short);

	/* 
	 * Allocate memory for the URL to return, and copy the url 
	 * NB: Size includes the null terminator 
	 */
	url = xcalloc(1, size);
	memcpy(url, tmp, size); 
	
	/*
	 * Check to see if the URL is an icp url, in which case, we need to 
	 * build an edge for the cached-host 
	 */
	if (strncasecmp(url, "icp://", 6) == 0)
	{
		/* Get the host and ports of the cached host */ 
		host = xcalloc(1, 256); /* XXX - should use a constant */
		ret = xcalloc(1, size);
		strcpy(ret, parseICPUrl(url, host, &ascii_port, &tcp_port, &udp_port));

		/* Get an edge for this cached-host */ 
		*e = neighbors_get(host, ascii_port, udp_port, tcp_port);
		if (!*e)
		{
			/* This neighbor was not found, install a new one */
			neighbors_install(host, "unknown", ascii_port, 
				udp_port, tcp_port);

			/* XXX What if this fails ?? */
			*e = neighbors_get(host, ascii_port, udp_port, tcp_port);
		}

		/* Clean up */
		xfree(url);
		xfree(host);

		return ret;
	}
	else
	{
		/* Return a null edge entry */
		*e = 0;
		return url;
	}


}


char *parseICPUrl(url, host, ascii_port, udp_port, tcp_port)
char *url;
char *host;
int *ascii_port;
int *udp_port;
int *tcp_port;
{
	int i, n=0;
	char *tmp;
	char num[12];
	int ports[3] = {3128, 3129, 3130}; /* XXX - Defaults */
	int port;

	/* Skip the icp:// part */
	tmp = url + 6;

	/* Get the host part */ 
	i = strcspn(tmp, "/:");
	strncpy(host, tmp, i); 

	/* Get the ports */
	tmp += i;
	while (*tmp != '/')
	{
		if (n>2)
		{
			debug(1, "parseICPUrl: invalid ICP url <%s>\n", url);
			return 0;
		}

		tmp++;
		i = strcspn(tmp, ":/");

		/* Get the port if it isn't a default */ 
		if (i)
		{
			strncpy(num, tmp, i);
			port=atoi(num);

			/* Check port */
			if (!port)
			{
				debug(1, "parseICPUrl: invalid ICP url <%s>\n", url);
				return 0;
			}
			else
			{
				ports[n] = port;
			}
			
		}
		n++;
		tmp+=i;
	}

	/* Get the ports */
	*ascii_port = ports[0];
	*tcp_port   = ports[1];
	*udp_port   = ports[2];

	/* Return the url */
	return tmp+1;
}
