/*
 *  $Id: proto.v1.4h,v 1.14 1995/08/02 22:39:42 hardy Exp $
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#ifndef PROTOV1_4_HEADER
#define PROTOV1_4_HEADER

#include "ansihelp.h"
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>

#define ICPV1_4_VERSION_CURRENT 2 /* Version code used for ICPv1.4 */

#define ICPV1_4_AUTH_SIZE (2)	/* size of authenticator field */
struct icpv1_4_common_s {
    unsigned char opcode;	/* opcode */
    unsigned char version;	/* version number */
    unsigned short length;	/* total length (bytes) */
    u_num32 reqnum;		/* req number (req'd for UDP) */
    u_num32 auth[ICPV1_4_AUTH_SIZE];	/* authenticator (future) */
    u_num32 shostid;		/* sender host id */
};
typedef struct icpv1_4_common_s icpv1_4_common_t;
#define ICPV1_4_COMMON_SZ (sizeof(icpv1_4_common_t))
#define ICPV1_4_HDR_SZ (sizeof(icpv1_4_common_t)+sizeof(u_num32))
#define ICPV1_4_MAX_URL (4096)
typedef enum {
    ICPV1_4_OP_INVALID,		/* to insure 0 doesn't get accidently interpreted. */
    ICPV1_4_OP_QUERY,		/* query opcode (cl->sv) */
    ICPV1_4_OP_HIT,			/* hit (cl<-sv) */
    ICPV1_4_OP_MISS,		/* miss (cl<-sv) */
    ICPV1_4_OP_ERR,			/* error (cl<-sv) */
    ICPV1_4_OP_SEND,		/* send object non-auth (cl->sv) */
    ICPV1_4_OP_SENDA,		/* send object authoritative (cl->sv) */
    ICPV1_4_OP_DATABEG,		/* first data, but not last (sv<-cl) */
    ICPV1_4_OP_DATA,		/* data middle of stream (sv<-cl) */
    ICPV1_4_OP_DATAEND,		/* last data (sv<-cl) */
    ICPV1_4_OP_SECHO,		/* echo from source (sv<-os) */
    ICPV1_4_OP_END			/* marks end of opcodes */
} icpv1_4_opcode;

#define ICPV1_4_OP_HIGHEST (ICPV1_4_OP_END - 1)		/* highest valid opcode */


/* Header for QUERY packet */
struct icpv1_4_query_s {
    u_num32 q_rhostid;		/* requestor host id */
    char *q_url;		/* variable sized URL data */
};
typedef struct icpv1_4_query_s icpv1_4_query_t;
#define ICPV1_4_QUERY_SZ (sizeof(icpv1_4_query_t))

/* Header for HIT packet */
struct icpv1_4_hit_s {
    u_num32 h_size;		/* size if known */
    char *h_url;		/* variable sized URL data */
};
typedef struct icpv1_4_hit_s icpv1_4_hit_t;
#define ICPV1_4_HIT_SZ (sizeof(icpv1_4_hit_t))

/* Header for MISS packet */
struct icpv1_4_miss_s {
    char *m_url;		/* variable sized URL data */
};
typedef struct icpv1_4_miss_s icpv1_4_miss_t;
#define ICPV1_4_MISS_SZ (sizeof(icpv1_4_miss_t))

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* Header for ERROR packet */
struct icpv1_4_error_s {
    unsigned short e_code;	/* error code */
    char *e_msg;		/* variable sized string message */
};
typedef struct icpv1_4_error_s icpv1_4_error_t;
#define ICPV1_4_ERROR_SZ (sizeof(icpv1_4_error_t))

#define ICPV1_4_ERROR_MSGLEN	256	/* max size for string, incl '\0' */

/* Error Codes - These can come back in the response packet */
typedef enum {
    ICPV1_4_ERROR_INVALID,		/* invalid (not used) */
    ICPV1_4_ERROR_BADVERS,		/* version error */
    ICPV1_4_ERROR_BADURL,		/* bad URL */
    ICPV1_4_ERROR_BADFLAGS,		/* bad flags */
    ICPV1_4_ERROR_TIMEDOUT,		/* couldn't get data */
    ICPV1_4_ERROR_ACCESS,		/* authorization problem */
    ICPV1_4_ERROR_INTERNAL		/* cache server internal err */
} icpv1_4_error_code;

/* Header for SEND packet */
struct icpv1_4_send_s {
    u_num32 s_rhostid;		/* requestor host id */
    char *s_url;		/* variable sized url */
};
typedef struct icpv1_4_send_s icpv1_4_send_t;
#define ICPV1_4_SEND_SZ (sizeof(icpv1_4_send_t))


/* Header for SENDA packet */
struct icpv1_4_senda_s {
    u_num32 sa_rhostid;		/* requestor host id */
    char *sa_url;		/* variable sized url */
};
typedef struct icpv1_4_senda_s icpv1_4_senda_t;
#define ICPV1_4_SENDA_SZ (sizeof(icpv1_4_senda_t))


/* Header for DATABEGIN packet */
struct icpv1_4_datab_s {
    u_num32 db_ttl;		/* time to live */
    u_num32 db_ts;		/* timestamp when gotten from owner */
    u_num32 db_size;		/* size of object if known */
    char *db_data;		/* variable sized data */
};
typedef struct icpv1_4_datab_s icpv1_4_datab_t;
#define ICPV1_4_DATAB_SZ (sizeof(icpv1_4_datab_t))


/* Header for DATA packet */
struct icpv1_4_data_s {
    u_num32 d_offset;		/* offset into object for d_data  */
    char *d_data;		/* variable sized data */
};
typedef struct icpv1_4_data_s icpv1_4_data_t;
#define ICPV1_4_DATA_SZ (sizeof(icpv1_4_databe_t))


/* ICPV1_4 message type. */
struct icpv1_4_message_s {
    icpv1_4_common_t header;
    union {
	icpv1_4_query_t query;
	icpv1_4_hit_t hit;
	icpv1_4_miss_t miss;
	icpv1_4_error_t error;
	icpv1_4_send_t send;
	icpv1_4_senda_t senda;
	icpv1_4_datab_t data_begin;
	icpv1_4_data_t data;	/* Shared between DATA and DATAEND. */
    } op;
};
typedef struct icpv1_4_message_s icpv1_4_message_t;
#define ICPV1_4_MESSAGE_SZ (sizeof(icpv1_4_message_t))

extern int icpv1_4_proto_errno;	/* operation errors */
extern int icpv1_4_proto_init _PARAMS((void));
extern long icpv1_4_proto_reqnum _PARAMS((void));
extern int icpv1_4_query _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, struct in_addr rid, char *url, char *hostname, int port));
extern int icpv1_4_hit _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, u_num32 size));
extern int icpv1_4_miss _PARAMS((int sock, u_num32 reqnum, u_num32 * auth));
extern int icpv1_4_error _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, unsigned short errcode, char *errstr));
extern int icpv1_4_send _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, struct in_addr rid, char *url));
extern int icpv1_4_databegin _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, u_num32 ttl, u_num32 timestamp, char *data));
extern int icpv1_4_data _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, char *data));
extern int icpv1_4_dataend _PARAMS((int sock, u_num32 reqnum, u_num32 * auth, char *data));

/* stuff that handle url in proto.c.
   They are not actually belong here but they are in proto.c.
   ATC
 */

extern int proto_cachable _PARAMS((char *url, char *type, char *mime_hdr));
extern int proto_defaultTTL _PARAMS((char *url));

#endif				/* PROTO_HEADER */
