/* 
 *  File:         recursive.c
 *  Description:  Registration of recursive requests  
 *  Author:       Dean Povey, University of Queesnland 
 *  Date;         Wed Oct 12 11:35:30 EST 1995
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>       
#include <limits.h>   
#include <sys/errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ansihelp.h"
#include "proto.h"
#include "hash.h"
#include "icp.h"
#include "recursive.h"
#include "debug.h"

static HashID request_tbl=0;

extern int urlcmp(char*, char *); /* from store.c */

/* Register a recursive request */
void recRegisterRequest(url, host, reqnum, from)
char *url;
u_num32 host;
u_num32 reqnum;
struct sockaddr_in *from;
{
	recQueryData *data=0;
	recQueryData *pending=0;

	/* Allocate memory for query data */
	data = (recQueryData *) xcalloc(1, sizeof(recQueryData));

	/* Add query data info */
	data->url = strdup(url);
	data->host = host;
	data->reqnum = reqnum;
	memcpy(&data->reply_to, from, sizeof(struct sockaddr_in));
	data->next = 0;

	/* Check to see if hash table created, if it isn't create it */
    if (request_tbl == (HashID) 0) {
        request_tbl = (HashID) hash_create(
			(int (*)(char *, char *))urlcmp, 7921); /* XXX */
    }
	else
	{
		/* Check to see if there is another pending query */ 
		pending = recGetRequest(data->url);
		if (pending)
		{
			/* Append this request to the end of the list */
			while(pending->next)
				pending = pending->next;
			pending->next = data;
			return;
		}
	}
	
	/* Insert the new request into the hash table */
    hash_insert(request_tbl, data->url, data);
	return;
}

/* Get the hosts registered for a recursive request */
recQueryData *recGetRequest(url)
char *url;
{
	hash_link *hptr = NULL;

    debug(5, "recGetRequest: looking up %s\n", url);


    if (request_tbl != (HashID) 0) {
    	if ((hptr = hash_lookup(request_tbl, url)) != NULL)
        	return hptr->item;
    }
    return NULL;
}

/* Release the memory associated with a request */
int recReleaseRequest(url)
char *url;
{
	recQueryData *entry;
	recQueryData *tmp;

	if (request_tbl == ((HashID)  0))
		return -1;

	/* Get the entry */
	entry = recGetRequest(url);
	entry = entry->next;

	/* Remove the entry from the table this'll free the first entry */
	hash_delete(request_tbl, url);

	/* Free the rest */
	while(entry)
	{
		tmp = entry;
		entry = entry->next;
		safe_free(tmp);
	}
	return 0;
	
}
