static char rcsid[] = "$Id: stat.c,v 1.79 1995/08/21 21:03:40 duane Exp $";
/* 
 *  File:         stat.c
 *  Description:  stat module for object cache
 *  Author:       Anawat Chankhunthod, USC
 *  Created:      May 12 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>		/* for lseek. should be removed later. */
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#if defined(hpux) || defined(__hpux) || defined(HPUX)
#include <sys/syscall.h>
#define getrusage(a, b)  syscall(SYS_GETRUSAGE, a, b)
#define getpagesize( )   sysconf(_SC_PAGE_SIZE)
#endif				/* hpux */

#include "store.h"
#include "stat.h"
#include "disk.h"
#include "ipcache.h"
#include "cache_cf.h"
#include "hash.h"
#include "config.h"

extern int emulate_httpd_log;

#define MIN_BUFSIZE (4096)
#define MAX_LINELEN (4096)

typedef struct _log_read_data_t {
    StoreEntry *sentry;
} log_read_data_t;

typedef struct _cached_read_data_t {
    StoreEntry *sentry;
    int fd;
} cached_read_data_t;

extern int getMaxFD(), fdstat_biggest_fd();
extern char *mkrfc850();
extern time_t cached_curtime;
extern time_t cached_starttime;
extern time_t next_cleaning;
extern int httpd_accel_mode;
extern int debug_level;
extern unsigned long nconn;
extern int hash_links_allocated;
char *stat_describe(), *mem_describe(), *ttl_describe(), *flags_describe(),
*elapsed_time();

/* process utilization information */
int stat_utilization_get(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    char tempbuf[MAX_LINELEN];
    char *buffer;
    int curlen = 0;
    int size = MIN_BUFSIZE;
    int proto_id;

    buffer = (char *) xcalloc(1, size);
    sprintf(buffer, "{\n");
    curlen += 3;

    obj->proto_stat_data[0].object_count = 0;
    obj->proto_stat_data[0].mem_usage.max = 0;
    obj->proto_stat_data[0].mem_usage.min = 0;
    obj->proto_stat_data[0].mem_usage.avg = 0;
    obj->proto_stat_data[0].mem_usage.current = 0;
    obj->proto_stat_data[0].hit = 0;
    obj->proto_stat_data[0].miss = 0;
    obj->proto_stat_data[0].refcount = 0;
    obj->proto_stat_data[0].transferbyte = 0;


    /* find the total */
    for (proto_id = 1; proto_id <= PROTOCOL_SUPPORTED; ++proto_id) {
	obj->proto_stat_data[0].object_count +=
	    obj->proto_stat_data[proto_id].object_count;

	obj->proto_stat_data[0].mem_usage.max +=
	    obj->proto_stat_data[proto_id].mem_usage.max;

	obj->proto_stat_data[0].mem_usage.min +=
	    obj->proto_stat_data[proto_id].mem_usage.min;

	obj->proto_stat_data[0].mem_usage.avg +=
	    obj->proto_stat_data[proto_id].mem_usage.avg;

	obj->proto_stat_data[0].mem_usage.current +=
	    obj->proto_stat_data[proto_id].mem_usage.current;

	obj->proto_stat_data[0].hit +=
	    obj->proto_stat_data[proto_id].hit;

	obj->proto_stat_data[0].miss +=
	    obj->proto_stat_data[proto_id].miss;

	obj->proto_stat_data[0].refcount +=
	    obj->proto_stat_data[proto_id].refcount;

	obj->proto_stat_data[0].transferbyte +=
	    obj->proto_stat_data[proto_id].transferbyte;

    }

    /* dump it */
    for (proto_id = 0; proto_id < PROTOCOL_SUPPORTED + PROTOCOL_EXTRA; ++proto_id) {
	if (obj->proto_stat_data[proto_id].hit != 0) {
	    obj->proto_stat_data[proto_id].hitratio =
		(float) obj->proto_stat_data[proto_id].hit /
		((float) obj->proto_stat_data[proto_id].hit +
		 (float) obj->proto_stat_data[proto_id].miss);
	}
	sprintf(tempbuf, "{%s %d %d %d %d %4.2f %d %d %d}\n",
		obj->proto_stat_data[proto_id].protoname,
		obj->proto_stat_data[proto_id].object_count,
		obj->proto_stat_data[proto_id].mem_usage.max,
		obj->proto_stat_data[proto_id].mem_usage.current,
		obj->proto_stat_data[proto_id].mem_usage.min,
		obj->proto_stat_data[proto_id].hitratio,
		obj->proto_stat_data[proto_id].transferrate,
		obj->proto_stat_data[proto_id].refcount,
		obj->proto_stat_data[proto_id].transferbyte);

	curlen += strlen(tempbuf) % MAX_LINELEN;
	if (curlen > size) {
	    /* it's bigger than buffer. Do realloc. */
	    size += MIN_BUFSIZE;
	    buffer = (char *) xrealloc(buffer, size);
	}
	/* XXX: This strncat is dangerous on big buffers DRH */
	strncat(buffer, tempbuf, MAX_LINELEN);

    }

    sprintf(tempbuf, "}\n");
    curlen += 2;
    strncat(buffer, tempbuf, MAX_LINELEN);

    *len = curlen;
    *buf = buffer;
    return *len;
}


/* return total bytes of all registered and known objects.
   may not reflect the retrieving object....
   something need to be done here to get more accurate cache size */
int cache_size_get(obj)
cacheinfo *obj;
{
    int size = 0;
    int proto_id;
    /* sum all size, exclude total */
    for (proto_id = 1; proto_id <= PROTOCOL_SUPPORTED + PROTOCOL_EXTRA - 1;
	 ++proto_id) {
	size += obj->proto_stat_data[proto_id].mem_usage.current;
    }
    return (size);
}

/* process general IP cache information */
int stat_general_get(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{

    stat_ipcache_get(obj, (char **) buf, len);

    return *len;
}


/* process objects list */
int stat_objects_get(obj, buf, len, vm_or_not)
cacheinfo *obj;
char **buf;
int *len;
int vm_or_not;
{
    char tempbuf[MAX_LINELEN + 1], *buffer, *space[40], *space2[40];
    int did_big_alloc = 0;
    int curlen = 0, npend, i, size = MIN_BUFSIZE, sz;
    StoreEntry *entry;
    int N = 0;

    buffer = (char *) xcalloc(1, size);
    strcpy(buffer, "{\n");
    curlen += 2;

    for (entry = storeGetFirst();
	 entry != NULL;
	 entry = storeGetNext()) {
	if (vm_or_not && (entry->mem_status == NOT_IN_MEMORY)) continue;
	if ((++N & 0xFF) == 0) {
	    cached_curtime = time(NULL);
	    debug(3, "stat_objects_get:  Processed %d objects...\n", N);
	}
	for (npend = i = 0; i < entry->pending_list_size; i++) {
	    if (entry->pending[i])
		npend++;
	}
	tempbuf[0] = '\0';
	sprintf(tempbuf, "{ %s %d %s %s %s %s %d %d %s %s }\n",
		entry->url,
		entry->object_len,
		elapsed_time(entry, entry->timestamp, space),
		flags_describe(entry),
		elapsed_time(entry, entry->lastref, space2),
		ttl_describe(entry, entry->ttl),
		npend,
		entry->refcount,
		mem_describe(entry),
		stat_describe(entry));

	sz = strlen(tempbuf);
	sz = sz % MAX_LINELEN;	/* only use MAX_LINELEN of it */
	if (curlen + sz >= size) {
	    /* it's bigger than buffer. Do realloc. */
	    /* the objects list can be **VERY** big, so be smart */
	    size += did_big_alloc ? MIN_BUFSIZE :
		(meta_data.store_entries * 120);
            if (did_big_alloc == 0)
		debug(1, "stat_objects_get:  Allocating temporary memory buffer with %d bytes to hold %d objects...\n", size, meta_data.store_entries);
	    buffer = (char *) xrealloc(buffer, size);
	    did_big_alloc = 1;
	}
	/* NEVER do multiple strncat's on big buffers! Use memcpy */
	memcpy(&buffer[curlen], tempbuf, sz);
	curlen += sz;
	buffer[curlen] = '\0';
    }
    buffer[curlen++] = '}';
    buffer[curlen++] = '\n';
    buffer[curlen] = '\0';

    *len = curlen;
    *buf = buffer;
    return *len;
}


/* process a requested object into a manager format */
int stat_get(obj, req, buf, len)
cacheinfo *obj;
char *req;
char **buf;
int *len;
{

    *len = 0;

    if (strncmp(req, "objects", strlen("objects")) == 0) {
	stat_objects_get(obj, buf, len, 0);
    } else if (strncmp(req, "vm_objects", strlen("vm_objects")) == 0) {
	stat_objects_get(obj, buf, len, 1);
    } else if (strncmp(req, "general", strlen("general")) == 0) {
	stat_general_get(obj, buf, len);
    } else if (strncmp(req, "utilization", strlen("utilization")) == 0) {
	stat_utilization_get(obj, buf, len);
    }
    return *len;
}


/* generate logfile status information */
int log_status_get(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{

    char *buffer;
    int size = MIN_BUFSIZE;

    buffer = (char *) xcalloc(1, size);
    if (obj->logfile_status == LOG_ENABLE) {
	sprintf(buffer, "{\"Logfile is Enabled. Filename: %s\"}\n", obj->logfilename);
    } else {
	sprintf(buffer, "{\"Logfile is Disabled.\"}\n");
    }
    *len = strlen(buffer);
    *buf = buffer;
    return *len;
}



/* log convert handler */
/* call for each line in file, use fileWalk routine */
int logReadHandler(fd, buf, size, data)
int fd;
char *buf;
int size;
log_read_data_t *data;
{
    char tempbuf[MAX_LINELEN];

    sprintf(tempbuf, "{%s}\n", buf);
    return storeAppend(data->sentry, tempbuf, strlen(tempbuf) % MAX_LINELEN);
}

/* log convert end handler */
/* call when a walk is completed or error. */
int logReadEndHandler(fd, errflag, data)
int fd;
int errflag;
log_read_data_t *data;
{
    char tempbuf[MAX_LINELEN];

    sprintf(tempbuf, "}\n");
    storeAppend(data->sentry, tempbuf, 2);
    storeComplete(data->sentry);
    safe_free(data);
    return 0;
}



/* start converting logfile to processed format */
int log_get_start(obj, sentry)
cacheinfo *obj;
StoreEntry *sentry;
{
    if (obj->logfile_status == LOG_DISABLE) {
	char *tmp = NULL;
	int len;

	/* Manufacture status when logging is disabled */
	log_status_get(obj, &tmp, &len);
	storeAppend(sentry, tmp, len);
	storeComplete(sentry);
    } else {
	char tmp[3];
	log_read_data_t *data;

	data = (log_read_data_t *) xmalloc(sizeof(log_read_data_t));
	memset(data, '\0', sizeof(log_read_data_t));
	data->sentry = sentry;
	strcpy(tmp, "{\n");
	storeAppend(sentry, tmp, 2);
	file_walk(obj->logfile_fd, (FILE_WALK_HD) logReadEndHandler,
	 (caddr_t) data, (FILE_WALK_LHD) logReadHandler, (caddr_t) data);
    }
    return 0;
}


/* cached convert handler */
/* call for each line in file, use fileWalk routine */
int cachedReadHandler(fd, buf, size, data)
int fd;
char *buf;
int size;
cached_read_data_t *data;
{
    char tempbuf[MAX_LINELEN];

    memset(tempbuf, '\0', MAX_LINELEN);
    sprintf(tempbuf, "{\"%s\"}\n", buf);
    return storeAppend(data->sentry, tempbuf, strlen(tempbuf) % MAX_LINELEN);
}

/* cached convert end handler */
/* call when a walk is completed or error. */
int cachedReadEndHandler(fd, errflag, data)
int fd;
int errflag;
cached_read_data_t *data;
{
    char tempbuf[3];

    strcpy(tempbuf, "}\n");
    storeAppend(data->sentry, tempbuf, 2);
    storeComplete(data->sentry);
    file_close(data->fd);
    safe_free(data);
    return 0;
}


/* start convert cached.conf file to processed format */
int cached_get_start(obj, sentry)
cacheinfo *obj;
StoreEntry *sentry;
{
    char tempbuf[3];
    cached_read_data_t *data;
    extern char *config_file;

    data = (cached_read_data_t *) xmalloc(sizeof(cached_read_data_t));
    memset(data, '\0', sizeof(cached_read_data_t));
    data->sentry = sentry;
    data->fd = file_open((char *) config_file, NULL, O_RDONLY);
    sprintf(tempbuf, "{\n");
    storeAppend(sentry, tempbuf, 2);
    file_walk(data->fd, (FILE_WALK_HD) cachedReadEndHandler, (caddr_t) data,
	      (FILE_WALK_LHD) cachedReadHandler, (caddr_t) data);
    return 0;
}


int dummyhandler(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    char *buffer = (char *) xcalloc(1, 64);

    strcpy(buffer, "{ \"Not_Implemented_yet.\"}\n");
    *len = strlen(buffer);
    *buf = buffer;
    return *len;
}

#if !defined(DO_MALLINFO) && !defined(__linux__)
#define DO_MALLINFO 1
#endif

int info_get(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    char *tod;

#if defined(HAVE_GETRUSAGE) && defined(RUSAGE_SELF)
    struct rusage rusage;
#endif
    char *buffer = (char *) xcalloc(1, MIN_BUFSIZE);
    char *line = (char *) xcalloc(1, MAX_LINELEN);

#ifdef DO_MALLINFO
    struct mallinfo mallinfo(), mp;
#endif

    strcat(buffer, "{\n");	/* start */

    sprintf(line, "{Harvest Object Cache: Version %s}\n", HARVEST_VERSION);
    strncat(buffer, line, MAX_LINELEN);

    tod = mkrfc850(&cached_starttime);

    sprintf(line, "{Start Time:\t%s}\n", tod);
    strncat(buffer, line, MAX_LINELEN);

    tod = mkrfc850(&cached_curtime);
    sprintf(line, "{Current Time:\t%s}\n", tod);
    strncat(buffer, line, MAX_LINELEN);

    /* -------------------------------------------------- */

    strcat(buffer, "{Connection information for cached:}\n");

    sprintf(line, "{\tNumber of connections:\t%lu}\n", nconn);
    strncat(buffer, line, MAX_LINELEN);

    {
	float f;
	f = cached_curtime - cached_starttime;
	sprintf(line, "{\tConnections per hour:\t%.1f}\n", f == 0.0 ? 0.0 :
		(nconn / (f / 3600)));
	strncat(buffer, line, MAX_LINELEN);
    }

    /* -------------------------------------------------- */



    strcat(buffer, "{Cache information for cached:}\n");

    sprintf(line, "{\tStorage Swap size:\t%d}\n", storeGetSwapSize());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{\tStorage Mem size:\t%d}\n", storeGetMemSize());
    strncat(buffer, line, MAX_LINELEN);

    tod = mkrfc850(&next_cleaning);
    sprintf(line, "{\tStorage Expiration at:\t%s}\n", tod);
    strncat(buffer, line, MAX_LINELEN);

#if defined(HAVE_GETRUSAGE) && defined(RUSAGE_SELF)
    strcat(buffer, "{Resource usage for cached:}\n");

    getrusage(RUSAGE_SELF, &rusage);
    sprintf(line, "{\tCPU Usage: user %d sys %d}\n{\tProcess Size: rss %d KB}\n",
	    rusage.ru_utime.tv_sec, rusage.ru_stime.tv_sec,
	    rusage.ru_maxrss * getpagesize() / 1000);
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{\tPage faults with physical i/o:\t%d}\n",
	    rusage.ru_majflt);
    strncat(buffer, line, MAX_LINELEN);

#endif

#ifdef DO_MALLINFO
    mp = mallinfo();

    strcat(buffer, "{Memory usage for cached via mallinfo():}\n");

    sprintf(line, "{\ttotal space in arena:\t%d}\n", mp.arena);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tnumber of ordinary blocks:\t%d}\n", mp.ordblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tnumber of small blocks:\t%d}\n", mp.smblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tnumber of holding blocks:\t%d}\n", mp.hblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tspace in holding block headers:\t%d}\n", mp.hblkhd);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tspace in small blocks in use:\t%d}\n", mp.usmblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tspace in free blocks:\t%d}\n", mp.fsmblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tspace in ordinary blocks in use:\t%d}\n",
	    mp.uordblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tspace in free ordinary blocks:\t%d}\n", mp.fordblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tcost of enabling keep option:\t%d}\n", mp.keepcost);
    strncat(buffer, line, MAX_LINELEN);
#ifdef LNG_MALLINFO
    sprintf(line, "{\tmax size of small blocks:\t%d}\n", mp.mxfast);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tnumber of small blocks in a holding block:\t%d}\n",
	    mp.nlblks);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tsmall block rounding factor:\t%d}\n", mp.grain);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tspace (including overhead) allocated in ord. blks:\t%d}\n"
	    ,mp.uordbytes);
    sprintf(line, "{\tnumber of ordinary blocks allocated:\t%d}\n",
	    mp.allocated);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tbytes used in maintaining the free tree:\t%d}\n",
	    mp.treeoverhead);
    strncat(buffer, line, MAX_LINELEN);

#endif

#endif				/* DO_MALLINFO */

    strcat(buffer, "{File descriptor usage for cached:}\n");

    sprintf(line, "{\tMax number of file desc available:\t%d}\n", getMaxFD());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{\tLargest file desc currently in use:\t%d}\n",
	    fdstat_biggest_fd());
    strncat(buffer, line, MAX_LINELEN);

    {
	extern int *fd_lifetime;
	int i, max_fd = getMaxFD();
	char tbuf[64];

	strcat(buffer, "{\tCurrent file desc lifetimes (clients or ascii port):}\n");
	for (i = 0; i < max_fd; i++) {
	    if (fd_lifetime[i] > 0) {
		/* the lifetime should be greater than curtime */
		sprintf(tbuf, "{\t\t(%d = %d)}\n", i, fd_lifetime[i] - cached_curtime);
		strcat(buffer, tbuf);
	    }
	}
    }


    strcat(buffer, "{Stop List:}\n");
    if (http_stoplist) {
	stoplist *p;
	p = http_stoplist;
	sprintf(line, "{\tHTTP:}\n");
	strncat(buffer, line, MAX_LINELEN);
	while (p) {
	    sprintf(line, "{\t\t%s}\n", p->key);
	    strncat(buffer, line, MAX_LINELEN);
	    p = p->next;
	}
    }
    if (gopher_stoplist) {
	stoplist *p;
	p = gopher_stoplist;
	sprintf(line, "{\tGOPHER:}\n");
	strncat(buffer, line, MAX_LINELEN);
	while (p) {
	    sprintf(line, "{\t\t%s}\n", p->key);
	    strncat(buffer, line, MAX_LINELEN);
	    p = p->next;
	}
    }
    if (ftp_stoplist) {
	stoplist *p;
	p = ftp_stoplist;
	sprintf(line, "{\tFTP:}\n");
	strncat(buffer, line, MAX_LINELEN);
	while (p) {
	    sprintf(line, "{\t\t%s}\n", p->key);
	    strncat(buffer, line, MAX_LINELEN);
	    p = p->next;
	}
    }
    strcat(buffer, "{Internal Data Structures:}\n");
    sprintf(line, "{Meta Data:}\n");
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\t\tStoreEntry %d x %d}\n", sizeof(StoreEntry),
	    meta_data.store_entries);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\t\tIPCacheEntry %d x %d}\n", sizeof(ipcache_entry),
	    meta_data.ipcache_count);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\t\tHash link  %d x %d}\n", sizeof(hash_link),
	    meta_data.hash_links = hash_links_allocated);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\t\tURL strings %d}\n", meta_data.url_strings);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\t\tHot Object Cache Items %d}\n", meta_data.hot_vm);
    strncat(buffer, line, MAX_LINELEN);
    sprintf(line, "{\tTotal Accounted %d}\n",
	    meta_data.store_entries * sizeof(StoreEntry) +
	    meta_data.ipcache_count * sizeof(ipcache_entry) +
	    meta_data.hash_links * sizeof(hash_link) +
	    meta_data.url_strings);
    strncat(buffer, line, MAX_LINELEN);

    strcat(buffer, "}\n");	/* final */

    safe_free(line);
    *len = strlen(buffer);
    *buf = buffer;
    return *len;

}


int parameter_get(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    /* be careful if an object is bigger than 4096, 
       need more malloc here */
    char *buffer = (char *) xcalloc(1, MIN_BUFSIZE);
    char *line = (char *) xcalloc(1, MAX_LINELEN);

    strcpy(buffer, "{\n");

    sprintf(line, "{CacheMemMax %d \"# Maximum Memory size of cache\"}\n",
	    getCacheMemMax());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{CacheMemHigh %d \"# High W-Mark Memory size of cache\"}\n",
	    getCacheMemHighWaterMark());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{CacheMemLow %d \"# Low W-Mark Memory size of cache\"}\n",
	    getCacheMemLowWaterMark());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{CacheSwapMax %d \"# Maximum Disk Swap size of cache\"}\n",
	    getCacheSwapMax());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{CacheSwapHigh %d \"# High W-Mark Swap size of cache\"}\n",
	    getCacheSwapHighWaterMark());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{CacheSwapLow %d \"# Low W-Mark Swap size of cache\"}\n",
	    getCacheSwapLowWaterMark());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{HttpMax %d \"# Maximum size of cachable http object\"}\n",
	    getHttpMax());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{HttpTTL %d \"# TTL for http object\"}\n", getHttpTTL());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{GopherMax %d \"# Maximum size of cachable gopher object\"}\n"
	    ,getGopherMax());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{GopherTTL %d \"# TTL for gopher object\"}\n", getGopherTTL());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{FtpMax %d \"# Maximum size of cachable ftp object\"}\n",
	    getFtpMax());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{FtpTTL %d \"# TTL for ftp object\"}\n", getFtpTTL());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{NegativeTTL %d \"# TTL for negative cache\"}\n",
	    getNegativeTTL());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{ReadTimeout %d \"# Timeout for reading from remote server\"}\n", getReadTimeout());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{ClientLifetime %d \"# Lifetime for incoming ascii port requests or outgoing clients\"}\n", getClientLifetime());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{CleanRate %d \"# Rate for periodic object expiring\"}\n",
	    getCleanRate());
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{HttpAccelMode %s \"# Is operating as an HTTP accelerator\"}\n",
	    httpd_accel_mode ? "yes" : "no");
    strncat(buffer, line, MAX_LINELEN);

    sprintf(line, "{DebugLevel %d \"# Cache debug level\"}\n",
	    debug_level);
    strncat(buffer, line, MAX_LINELEN);

    /* end of stats */
    sprintf(line, "}\n");
    strncat(buffer, line, MAX_LINELEN);


    safe_free(line);
    *len = strlen(buffer);
    *buf = buffer;
    return *len;

}


int log_append(obj, url, id, size, action, method)
cacheinfo *obj;
char *url;
char *id;
int size;
char *action;
char *method;
{
    extern char *mkhttpdlogtime();
    char temp[8192];
    time_t t = time(NULL);

    if (obj->logfile_status == LOG_ENABLE) {
	temp[0] = '\0';
	if (emulate_httpd_log)
	    sprintf(temp, "%s - - [%s] \"%s %s\" %s %d\n",
		    id, mkhttpdlogtime(&t), method, url, action, size);
	else
	    sprintf(temp, "%d %s %s %d %s\n", t, url, id, size, action);

	if (file_write(obj->logfile_fd, temp, strlen(temp),
		       obj->logfile_access, NULL, NULL) != DISK_OK) {
	    debug(1, "log_append: File write failed.\n");
	}
    }
    return 0;
}

int log_enable(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    char *buffer = (char *) xmalloc(2);

    if (obj->logfile_status == LOG_DISABLE) {
	obj->logfile_status = LOG_ENABLE;

	/* open the logfile */
	obj->logfile_fd = file_open(obj->logfilename, NULL, O_RDWR | O_CREAT);
	if (obj->logfile_fd == DISK_ERROR) {
	    debug(0, "Cannot open logfile: %s\n", obj->logfilename);
	    obj->logfile_status = LOG_DISABLE;
	}
	obj->logfile_access = file_write_lock(obj->logfile_fd);

    }
    /* at the moment, store one char to make a storage manager happy */
    buffer[0] = ' ';
    buffer[1] = '\0';
    *len = 1;
    *buf = buffer;
    return *len;
}

int log_disable(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    char *buffer = (char *) xmalloc(2);

    if (obj->logfile_status == LOG_ENABLE)
	file_close(obj->logfile_fd);

    obj->logfile_status = LOG_DISABLE;

    /* at the moment, store one char to make a storage manager happy */
    buffer[0] = ' ';
    buffer[1] = '\0';
    *len = 1;
    *buf = buffer;
    return *len;
}



int log_clear(obj, buf, len)
cacheinfo *obj;
char **buf;
int *len;
{
    char *buffer = (char *) xmalloc(2);

    /* what should be done here. Erase file ??? or move it to another name */
    /* At the moment, just erase it. */
    /* bug here need to be fixed. what if there are still data in memory. Need flush here */
    if (obj->logfile_status == LOG_ENABLE)
	file_close(obj->logfile_fd);

    unlink(obj->logfilename);

    /* reopen it anyway */
    obj->logfile_fd = file_open(obj->logfilename, NULL, O_RDWR | O_CREAT);
    if (obj->logfile_fd == DISK_ERROR) {
	debug(0, "Cannot open logfile: %s\n", obj->logfilename);
	obj->logfile_status = LOG_DISABLE;
    }
    /* at the moment, store one char to make a storage manager happy */
    buffer[0] = ' ';
    buffer[1] = '\0';
    *len = 1;
    *buf = buffer;
    return *len;
}



int proto_newobject(obj, proto_id, size)
cacheinfo *obj;
int proto_id;
int size;
{
    obj->proto_stat_data[proto_id].object_count++;
    obj->proto_stat_data[proto_id].mem_usage.current += size;
    /* pretend current size to be an average for a moment. fix later */
    if (obj->proto_stat_data[proto_id].mem_usage.current >
	obj->proto_stat_data[proto_id].mem_usage.max) {
	obj->proto_stat_data[proto_id].mem_usage.max =
	    obj->proto_stat_data[proto_id].mem_usage.current;
    }
    return 0;
}


int proto_purgeobject(obj, proto_id, size)
cacheinfo *obj;
int proto_id;
int size;
{
    obj->proto_stat_data[proto_id].object_count--;
    obj->proto_stat_data[proto_id].mem_usage.current -= size;
    return 0;

}

/* update stat for each particular protocol when an object is fetched */
int proto_touchobject(obj, proto_id, size)
cacheinfo *obj;
int proto_id;
int size;
{
    obj->proto_stat_data[proto_id].refcount++;
    obj->proto_stat_data[proto_id].transferbyte += size;
    return 0;
}

int proto_hit(obj, proto_id, size)
cacheinfo *obj;
int proto_id;
int size;
{
    obj->proto_stat_data[proto_id].hit++;
    return 0;

}

int proto_miss(obj, proto_id)
cacheinfo *obj;
int proto_id;
{
    obj->proto_stat_data[proto_id].miss++;
    return 0;
}

int proto_url_to_id(url)
char *url;
{
    if (strncmp(url, "http:", 5) == 0)
	return HTTP_ID;
    if (strncmp(url, "ftp:", 4) == 0)
	return FTP_ID;
    if (strncmp(url, "gopher:", 7) == 0)
	return GOPHER_ID;
    if (strncmp(url, "cache_object:", 13) == 0)
	return CACHEOBJ_ID;
    if (strncmp(url, "news:", 5) == 0)
	return NOTIMPLE_ID;
    if (strncmp(url, "file:", 5) == 0)
	return NOTIMPLE_ID;
    if (strncmp(url, "dht:", 4) == 0)
	return NOTIMPLE_ID;
    return NOTIMPLE_ID;
}



void stat_init(object, logfilename)
cacheinfo **object;
char *logfilename;
{
    cacheinfo *obj;
    int i;

    obj = (cacheinfo *) xmalloc(sizeof(cacheinfo));
    memset(obj, '\0', sizeof(cacheinfo));

    obj->stat_get = stat_get;
    obj->info_get = info_get;
    obj->cache_size_get = cache_size_get;

    obj->log_get_start = log_get_start;
    obj->log_status_get = log_status_get;
    obj->log_append = log_append;
    obj->log_clear = log_clear;
    obj->log_enable = log_enable;
    obj->log_disable = log_disable;
    obj->logfile_status = LOG_ENABLE;

    obj->cached_get_start = cached_get_start;

    obj->parameter_get = parameter_get;

    memcpy(obj->logfilename, logfilename, (strlen(logfilename) + 1) % 256);
    obj->logfile_fd = file_open(obj->logfilename, NULL, O_RDWR | O_CREAT);
    if (obj->logfile_fd == DISK_ERROR) {
	debug(0, "Cannot open logfile: %s\n", obj->logfilename);
	obj->logfile_status = LOG_DISABLE;
	fatal("Cannot open logfile.\n");
    }
    obj->logfile_access = file_write_lock(obj->logfile_fd);

    obj->proto_id = proto_url_to_id;
    obj->proto_newobject = proto_newobject;
    obj->proto_purgeobject = proto_purgeobject;
    obj->proto_touchobject = proto_touchobject;
    obj->proto_hit = proto_hit;
    obj->proto_miss = proto_miss;
    obj->NotImplement = dummyhandler;

    for (i = 0; i < PROTOCOL_SUPPORTED + PROTOCOL_EXTRA; ++i) {

	switch (i) {

	case TOTAL_ID:
	    strcpy(obj->proto_stat_data[i].protoname, "Total");
	    break;

	case HTTP_ID:
	    strcpy(obj->proto_stat_data[i].protoname, "HTTP");
	    break;

	case GOPHER_ID:
	    strcpy(obj->proto_stat_data[i].protoname, "GOPHER");
	    break;

	case FTP_ID:
	    strcpy(obj->proto_stat_data[i].protoname, "FTP");
	    break;

	case CACHEOBJ_ID:
	    strcpy(obj->proto_stat_data[i].protoname, "CACHEOBJ");
	    break;

	case NOTIMPLE_ID:
	default:
	    strcpy(obj->proto_stat_data[i].protoname, "UNKNOWN");
	    break;
	}

	obj->proto_stat_data[i].object_count = 0;
	obj->proto_stat_data[i].hit = 0;
	obj->proto_stat_data[i].miss = 0;
	obj->proto_stat_data[i].hitratio = 0.0;
	obj->proto_stat_data[i].transferrate = 0;
	obj->proto_stat_data[i].refcount = 0;
	obj->proto_stat_data[i].transferbyte = 0;

	obj->proto_stat_data[i].mem_usage.max = 0;
	obj->proto_stat_data[i].mem_usage.min = 0;
	obj->proto_stat_data[i].mem_usage.avg = 0;
	obj->proto_stat_data[i].mem_usage.current = 0;


    }

    *object = obj;
}

/*
   Erase new line at end of string
 */
int zapnl(s)
char *s;
{
    while (*s) {
	if (*s == '\n') {
	    *s = '\0';
	    break;
	}
	++s;
    }
    return 0;
}

char *
 stat_describe(entry)
StoreEntry *entry;
{
    if (entry->status == STORE_OK)
	return ("Done");
    else if (entry->status == STORE_PENDING)
	return ("Xfr");
    else if (entry->status == STORE_ABORTED)
	return ("Abort");
    else
	return ("?");
}

char *
 mem_describe(entry)
StoreEntry *entry;
{
    if (entry->mem_status == NOT_IN_MEMORY)
	return ("D");
    else if (entry->mem_status == SWAPPING_IN)
	return ("ld");
    else if (entry->mem_status == IN_MEMORY)
	return ("VM");
    else
	return ("?");
}


char *
 ttl_describe(entry, ttl)
StoreEntry *entry;
int ttl;
{
    int hh, mm, ss;
    static char TTL[60];

    TTL[0] = '\0';
    strcpy(TTL, "UNKNOWN");	/* sometimes the TTL isn't set below */
    ttl = ttl - (cached_curtime - entry->timestamp);
    if (ttl < 0)
	strcpy(TTL, "Expired");
    else {

	hh = ttl / 3600;
	ttl -= hh * 3600;
	mm = ttl / 60;
	ttl -= mm * 60;
	ss = ttl;

	sprintf(TTL, "% 6d:%02d:%02d", hh, mm, ss);
    }
    return (TTL);
}

char *
 elapsed_time(entry, since, TTL)
StoreEntry *entry;
int since;
char *TTL;
{
    int hh, mm, ss, ttl;

    TTL[0] = '\0';
    strcpy(TTL, "UNKNOWN");	/* sometimes TTL doesn't get set */
    ttl = cached_curtime - since;
    if (ttl < 0)
	strcpy(TTL, "Expired");
    else {
	hh = ttl / 3600;
	ttl -= hh * 3600;
	mm = ttl / 60;
	ttl -= mm * 60;
	ss = ttl;

	sprintf(TTL, "% 6d:%02d:%02d", hh, mm, ss);
    }
    return (TTL);
}


#define KEY_CHANGE (1<<15)
#define KEY_URL    (1<<14)
#define CACHABLE   (1<<11)
#define REFRESH_REQUEST   (1<<10)
#define RELEASE_REQUEST (1<<9)
#define ABORT_MSG_PENDING (1<<8)
#define DELAY_SENDING (1<<7)
#define OBJ_LOCK (1<<0)

char *
 flags_describe(entry)
StoreEntry *entry;
{
    static char FLAGS[32];
    char key_change = 'k';
    char key_url = 'u';
    char cachable = 'c';
    char refresh_request = 'r';
    char release_request = 'z';
    char abort_msg_pending = 'a';
    char delay_sending = 'd';
    char obj_lock = 'l';

    memset(FLAGS, '\0', 32);
    if (BIT_TEST(entry->flag, KEY_CHANGE))
	key_change = 'K';
    if (BIT_TEST(entry->flag, KEY_URL))
	key_url = 'U';
    if (BIT_TEST(entry->flag, CACHABLE))
	cachable = 'C';
    if (BIT_TEST(entry->flag, REFRESH_REQUEST))
	refresh_request = 'R';
    if (BIT_TEST(entry->flag, RELEASE_REQUEST))
	release_request = 'Z';
    if (BIT_TEST(entry->flag, ABORT_MSG_PENDING))
	abort_msg_pending = 'A';
    if (BIT_TEST(entry->flag, DELAY_SENDING))
	delay_sending = 'D';
    if (BIT_TEST(entry->flag, OBJ_LOCK))
	obj_lock = 'L';

    sprintf(FLAGS, "%c%c%c%c%c%c%c%c%-2d",
	    key_change,
	    key_url,
	    cachable,
	    refresh_request,
	    release_request,
	    abort_msg_pending,
	    delay_sending,
	    obj_lock,
	    entry->lock_count);

    return (FLAGS);
}
