static char rcsid[] = "$Id: stmem.c,v 1.30 1995/08/21 08:11:08 duane Exp $";
/* 
 *  File:         stmem.c
 *  Description:  Store manager <-> memory manager interface
 *  Author:       Chuck Neerdaels, USC (chuckn@rand.org)
 *  Created:      Tue Apr  5 16:51:26 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "stmem.h"

void memFree(mem)
mem_ptr mem;
{
    mem_node lastp, p = mem->head;

    if (p) {
	while (p && (p != mem->tail)) {
	    lastp = p;
	    p = p->next;
	    if (lastp) {
		safe_free(lastp->data);
		safe_free(lastp);
	    }
	}

	if (p) {
	    safe_free(p->data);
	    safe_free(p);
	}
    }
    memset(mem, '\0', sizeof(mem_ptr));		/* nuke in case ref'ed again */
    safe_free(mem);
}

void memFreeData(mem)
mem_ptr mem;
{
    mem_node lastp, p = mem->head;

    while (p != mem->tail) {
	lastp = p;
	p = p->next;
	/* AWT's hack */
	safe_free(lastp->data);

	safe_free(lastp);
    }

    if (p != NULL) {
	safe_free(p->data);
	safe_free(p);
	p = NULL;
    }
    mem->head = mem->tail = NULL;	/* detach in case ref'd */
}


/* Append incoming data. */
int memAppend(mem, data, len)
mem_ptr mem;
char *data;
int len;
{
    mem_node p;

    debug(6, "memAppend: len %d\n", len);

    p = (mem_node) xcalloc(1, sizeof(Mem_Node));
    p->next = NULL;
    p->len = len;
    p->data = xcalloc(1, len);
    memcpy(p->data, data, len);

    if (!mem->head) {
	mem->head = mem->tail = p;
    } else {
	mem->tail->next = p;
	mem->tail = p;
    }

    return len;
}

#define min(x,y) ((x)<(y)? (x) : (y))

int memCompact(mem)
mem_ptr mem;
{
    return (0);
}

int memGrep(mem, string, nbytes)
mem_ptr mem;
char *string;
int nbytes;
{
    mem_node p = mem->head;
    char *str_i, *mem_i;
    int i = 0, blk_idx = 0, state, goal;

    debug(6, "memGrep: looking for %s in less than %d bytes.\n",
	  string, nbytes);

    if (!p)
	return 0;

    str_i = string;
    mem_i = p->data;
    state = 1;
    goal = strlen(string);

    while (i < nbytes) {
	if (tolower(*mem_i++) == tolower(*str_i++))
	    state++;
	else {
	    state = 1;
	    str_i = string;
	}

	/* Return offset of byte beyond the matching string */
	if (state == goal)
	    return (i + 1);

	i++;
	blk_idx++;

	if (blk_idx >= p->len) {
	    if (p->next) {
		p = p->next;
		mem_i = p->data;
		blk_idx = 0;
	    } else
		break;
	}
    }
    return 0;
}

int memCopy(mem, offset, buf, size)
mem_ptr mem;
int offset;
char *buf;
int size;
{
    mem_node p = mem->head;
    int t_off = 0;
    int bytes_to_go = size;
    char *ptr_to_buf;
    int bytes_from_this_packet = 0;
    int bytes_into_this_packet = 0;

    debug(6, "memCopy: offset %d: size %d\n", offset, size);

    if (size <= 0)
	return size;

    /* Seek our way into store */
    while ((t_off + p->len) < offset) {
	t_off += p->len;
	p = p->next;
    }

    /* Start copying begining with this block until
       we're satiated */

    bytes_into_this_packet = offset - t_off;
    bytes_from_this_packet = min(bytes_to_go,
				 p->len - bytes_into_this_packet);

    memcpy(buf, p->data + bytes_into_this_packet, bytes_from_this_packet);
    bytes_to_go -= bytes_from_this_packet;
    ptr_to_buf = buf + bytes_from_this_packet;
    p = p->next;

    while (p && bytes_to_go > 0) {
	if (bytes_to_go > p->len) {
	    memcpy(ptr_to_buf, p->data, p->len);
	    ptr_to_buf += p->len;
	    bytes_to_go -= p->len;
	} else {
	    memcpy(ptr_to_buf, p->data, bytes_to_go);
	    bytes_to_go -= bytes_to_go;
	}
	p = p->next;
    }

    return size;
}


/* Do whatever is necessary to begin storage of new object */
mem_ptr memInit()
{
    mem_ptr new = (mem_ptr) xcalloc(1, sizeof(Mem_Hdr));

    new->tail = new->head = NULL;

    new->mem_free = memFree;
    new->mem_free_data = memFreeData;
    new->mem_append = memAppend;
    new->mem_copy = memCopy;
    new->mem_grep = memGrep;

    return new;
}
