static char rcsid[] = "$Id: store.c,v 1.116 1995/08/21 23:59:15 duane Exp $";
/* 
 *  File:         store.c
 *  Description:  Storage manager for the Cache
 *  Author:       John Noll, Anawat Chankhunthod, USC 
 *                Dean Povey, University of Queensland
 *  Created:      Sun Apr  3 16:51:26 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <time.h>
#include <syslog.h>
#include <string.h>
#include "ansihelp.h"		/* goes first */
#include "store.h"
#include "url.h"
#include "stat.h"
#include "disk.h"
#include "cache_cf.h"
#include "hash.h"
#include "debug.h"
#include "comm.h" /* For comm_hostname() */
#include "icp.h"

extern time_t cached_curtime;

#define MAX_SWAP_FILE 		(64*1024)
#define SWAP_BUF 		(64*1024)
#define FATAL_BUF_SIZE 		1024
#define SWAP_DIRECTORIES 	100
#define HOT_OBJECT_HIGH        (store_mem_high / 20000)
#define HOT_OBJECT_LOW         (store_mem_low / 20000)
#define MAX_POINTERS (64*1024)

#ifdef METADATA_IN_SWAPFILE
typedef struct _swap_hdr {
    time_t expires;
    char url[MAX_URL];
} swap_hdr;

#endif

extern char *sys_errlist[];
int urlcmp _PARAMS((char *s1, char *s2));

char *HTTP_OPS[] =
{"GET", "POST", "HEAD", ""};

/* Now, this table is inaccessible to outsider. They have to use a method
   to access a value in internal storage data structure. */
HashID table = 0;
HashID ptr_table = 0; /* Hash table for storing pointers */


/* initializtion flag */
static int store_initialized = 0;

/* current memory storage size */
static unsigned long store_mem_size = 0;
static unsigned long store_mem_high = 0;
static unsigned long store_mem_low = 0;


/* current file name, swap file, use number as a filename */
static unsigned long swapfileno = 0;
static int store_swap_size = 0;	/* kilobytes !! */
static unsigned long store_swap_high = 0;
static unsigned long store_swap_low = 0;
#ifndef METADATA_IN_SWAPFILE
static int swaplog_fd = 0;
static int swaplog_lock;
static int swaplog_tmp_fd = 0;
FILE *swaplog_stream;
FILE *swaplog_tmp_stream;
#endif

/* current swap file path */
static char swappath[MAX_FILE_NAME_LEN];

/* counter for uncachable objects */
static int uncache_count = 0;
static int keychange_count = 0;

/* key temp buffer */
static char key_temp_buffer[MAX_URL];

/* free whole entry */
void storeFreeEntry(e)
StoreEntry *e;
{
    meta_data.store_entries--;

    if (e == (StoreEntry *) NULL)
	return;

    debug(5, "storeFreeEntry: Freeing %s\n", e->url);

    if (e->mem_status != NOT_IN_MEMORY) {
	store_mem_size -= e->e_current_len;
	debug(8, "storeFreeEntry: Freeing %d in-memory bytes\n",
	      e->e_current_len);
	debug(8, "storeFreeEntry: store_mem_size = %d\n", store_mem_size);
	if (e->data) {
	    e->data->mem_free(e->data);
	    e->data = NULL;
	    --meta_data.hot_vm;
	}
    }
    meta_data.url_strings -= strlen(e->url);
    safe_free(e->url);
    if (!(e->flag & KEY_URL))
	safe_free(e->key);
    safe_free(e->mime_hdr);
    safe_free(e->e_abort_msg);
    safe_free(e->pending);
    safe_free(e);
}

/* free only data buffer, let the rest of structure stay. 
   For Negative cache purpose. */
void storeFreeEntryData(e)
StoreEntry *e;
{
    debug(5, "storeFreeEntryData: Freeing data-buffer only %s\n", e->url);
    store_mem_size -= e->e_current_len;
    debug(8, "storeFreeEntryData: Freeing %d in-memory bytes\n",
	  e->e_current_len);
    debug(8, "storeFreeEntryData: store_mem_size = %d\n", store_mem_size);
    e->object_len = 0;
    e->e_current_len = 0;
    if (e->data) {
	e->data->mem_free(e->data);
	e->data = NULL;
    }
}


/* get rid of memory copy of the object */
void storePurgeMem(e)
StoreEntry *e;
{
    debug(5, "storePurgeMem: Freeing memory-copy of %s\n", e->url);
    /* free up memory data */
    if (e->data) {
	e->data->mem_free(e->data);
	e->data = NULL;
	--meta_data.hot_vm;
    }
    store_mem_size -= e->object_len;
    debug(8, "storePurgeMem: Freeing %d in-memory bytes\n",
	  e->object_len);
    debug(8, "storePurgeMem: store_mem_size = %d\n", store_mem_size);
    e->mem_status = NOT_IN_MEMORY;
    e->e_current_len = 0;
    /* free up pending list table */
    safe_free(e->pending);
    e->pending_list_size = 0;
}

/* lock the object for reading, start swapping in if necessary */
int storeLockObject(e)
StoreEntry *e;
{
    e->lock_count++;
    debug(5, "storeLockObject: locks %d: <URL:%s>\n", e->lock_count, e->url);
    BIT_SET(e->flag, OBJ_LOCK);

    if ((e->mem_status == NOT_IN_MEMORY) &&
	(e->swap_status != SWAP_OK) &&
	(e->status != STORE_PENDING)) {
	/* where is it ???? */
	debug(1, "storeLockObject: Conflicting status: NOT_IN_MEMORY && NO_SWAP: <URL:%s>.", e->url);
	return -1;
    }
    e->lastref = time(0);
    e->refcount++;

    if ((e->mem_status == NOT_IN_MEMORY) &&
	(e->swap_status == SWAP_OK)) {
	/* object is in disk and no swapping daemon running. Bring it in. */
	storeSwapInStart(e);
    }
    return 0;
}

/* unlock object, return -1 if object get released after unlock
   otherwise lock_count */

int storeUnlockObject(e)
StoreEntry *e;
{
    if (e->lock_count > 0)
	e->lock_count--;
    if (e->lock_count == 0) {
	BIT_RESET(e->flag, OBJ_LOCK);
	/* Do it here to get accurate byte count. */
	CacheInfo->proto_touchobject(CacheInfo, proto_url_to_id(e->url),
				     e->object_len);

	if (e->flag & RELEASE_REQUEST) {
	    if (e->status != STORE_PENDING) {
		storeRelease(e);
		return -1;
	    }
	} else if (e->flag & ABORT_MSG_PENDING) {
	    /* replace content with abort message */
	    storeFreeEntryData(e);
	    e->data = memInit();
	    storeAppend(e, e->e_abort_msg, strlen(e->e_abort_msg));
	    e->object_len = e->e_current_len = strlen(e->e_abort_msg);
	    BIT_RESET(e->flag, ABORT_MSG_PENDING);
	}
    }
    return (int) e->lock_count;

}

/* Lookup an object in the cache.
   return just a reference to object, don't start swapping in yet. */
StoreEntry *storeGet(url)
char *url;
{
    hash_link *hptr = NULL;

    debug(5, "storeGet: looking up %s\n", url);

    if (table != (HashID) 0) {
    if ((hptr = hash_lookup(table, url)) != NULL)
        return hptr->item;
    }
    return NULL;
}

char *storeGenerateKey(url, request_type_id)
char *url;
int request_type_id;
{
    debug(5, "storeGenerateKey: type=%d %s\n", request_type_id, url);
    if (request_type_id == REQUEST_OP_POST) {
	sprintf(key_temp_buffer, "/post/%s", url);
	return key_temp_buffer;
    }
    if (request_type_id == REQUEST_OP_HEAD) {
	sprintf(key_temp_buffer, "/head/%s", url);
	return key_temp_buffer;
    }
    return url;
}

/* Add a new object to the cache. */
StoreEntry *
 storeAdd(url, type, mime_hdr, cachable, html_request, request_type_id)
char *url;
char *type;
char *mime_hdr;
int cachable;
int html_request;
int request_type_id;
{
    char key[MAX_URL + 16];
    static StoreEntry *e;

    debug(5, "storeAdd: %s\n", url);

    meta_data.store_entries++;
    meta_data.url_strings += strlen(url);

    if (meta_data.hot_vm > HOT_OBJECT_HIGH)
	storeGetMemSpace(0, 1);
    e = (StoreEntry *) xcalloc(1, sizeof(StoreEntry));
    e->url = strdup(url);
    e->key = NULL;
    e->flag = 0;
    e->type_id = request_type_id;
    if (mime_hdr) {
	e->mime_hdr = strdup(mime_hdr);
	if (mime_refresh_request(mime_hdr))
	    BIT_SET(e->flag, REFRESH_REQUEST);
	else
	    BIT_RESET(e->flag, REFRESH_REQUEST);
    }
    if (cachable) {
	BIT_SET(e->flag, CACHABLE);
	BIT_RESET(e->flag, RELEASE_REQUEST);
    } else {
	BIT_RESET(e->flag, CACHABLE);
	/*after a lock is release, it will be released by storeUnlock */
	BIT_SET(e->flag, RELEASE_REQUEST);
    }

    if (html_request)
	BIT_SET(e->flag, REQ_HTML);
    else
	BIT_RESET(e->flag, REQ_HTML);

    e->status = STORE_PENDING;
    e->mem_status = NOT_IN_MEMORY;
    e->swap_status = NO_SWAP;
    e->swap_file_number = -1;
    e->lock_count = 0;
    BIT_RESET(e->flag, KEY_CHANGE);
    BIT_RESET(e->flag, CLIENT_ABORT_REQUEST);
    e->data = memInit();
    meta_data.hot_vm++;
    e->refcount = 0;
    e->timestamp = e->lastref = time(0);
    if (e->flag & CACHABLE) {
	if (request_type_id == REQUEST_OP_GET) {
	    e->key = e->url;
	    BIT_SET(e->flag, KEY_URL);
	} else {
	    e->key = strdup(storeGenerateKey(e->url, request_type_id));
	    BIT_RESET(e->flag, KEY_URL);
	}
    } else {
	/* prepend a uncache count number to url for a key */
	key[0] = '\0';
	sprintf(key, "/%d/%s", uncache_count, url);
	uncache_count++;
	e->key = strdup(key);
	BIT_RESET(e->flag, KEY_URL);
    }

    /* allocate pending list */
    e->pending_list_size = MIN_PENDING;
    e->pending = (struct pentry **) xcalloc(e->pending_list_size,
					    sizeof(struct pentry *));

    if (table == (HashID) 0) {
	table = (HashID) hash_create(urlcmp, 7921);
	if (table == -1)
	{
		/* XXX Log an error */
	}
    }
    hash_insert(table, e->key, e);
    return e;
}

/* Add a new object to the cache with empty memory copy and pointer to disk
   use to rebuild store from disk. */
StoreEntry *storeAddDiskRestore(url, swap_filename, size, ttl)
char *url;
char *swap_filename;
int size;
int ttl;
{
    static StoreEntry *e;

    debug(5, "StoreAddDiskRestore: <URL:%s>: size %d: ttl %d: swapfile %s.\n",
	  url, size, ttl, swap_filename);

    meta_data.store_entries++;
    meta_data.url_strings += strlen(url);

    e = (StoreEntry *) xcalloc(1, sizeof(StoreEntry));
    e->url = strdup(url);
    e->key = NULL;
    e->flag = 0;
    e->type_id = REQUEST_OP_GET;
    e->mime_hdr = NULL;
    BIT_SET(e->flag, CACHABLE);
    BIT_RESET(e->flag, RELEASE_REQUEST);
    BIT_SET(e->flag, REQ_HTML);
    e->status = STORE_OK;
    e->mem_status = NOT_IN_MEMORY;
    e->swap_status = SWAP_OK;
    e->swap_file_number = parse_file_number(swap_filename);
    e->object_len = size;
    e->e_current_len = 0;
    e->lock_count = 0;
    BIT_RESET(e->flag, KEY_CHANGE);
    BIT_RESET(e->flag, CLIENT_ABORT_REQUEST);
    e->data = NULL;
    e->refcount = 0;
    e->timestamp = e->lastref = time(0);
    e->ttl = ttl;

    e->key = e->url;
    BIT_SET(e->flag, KEY_URL);

    /* allocate pending list */
    e->pending_list_size = MIN_PENDING;
    e->pending = (struct pentry **) xcalloc(e->pending_list_size,
					    sizeof(struct pentry *));
    if (!table) {
	table = (HashID) hash_create(urlcmp, 7921);
    }
    hash_insert(table, e->key, e);
    return e;
}

/* Register interest in an object currently being retrieved. */
int storeRegister(e, fd, handler, data)
StoreEntry *e;
int fd;
PIF handler;
caddr_t data;
{
    PendingEntry *pe = (PendingEntry *) xmalloc(sizeof(PendingEntry));
    int old_size, i, j;

    debug(5, "storeRegister: fd=%d %s\n", fd, e->url);

    memset(pe, '\0', sizeof(PendingEntry));
    pe->fd = fd;
    pe->handler = handler;
    pe->data = data;

    /* 
     *  I've rewritten all this pendings stuff so that num_pending goes
     *  away, and to fix all of the 'array bounds' problems we were having.
     *  It's now a very simple array, with any NULL slot empty/avail.
     *  If something needs to be added and there are no empty slots,
     *  it'll grow the array.
     */
    /* find an empty slot */
    for (i = 0; i < e->pending_list_size; i++)
	if (e->pending[i] == NULL)
	    break;

    if (i == e->pending_list_size) {
	/* grow the array */
	struct pentry **tmp;

	old_size = e->pending_list_size;

	/* set list_size to an appropriate amount */
	e->pending_list_size += MIN_PENDING;

	/* allocate, and copy old pending list over to the new one */
	tmp = xcalloc(e->pending_list_size, sizeof(struct pentry *));
	for (j = 0; j < old_size; j++)
	    tmp[j] = e->pending[j];

	/* free the old list and set the new one */
	safe_free(e->pending);
	e->pending = tmp;

	debug(10, "storeRegister: grew pending list to %d for slot %d.\n",
	      e->pending_list_size, i);
    }
    e->pending[i] = pe;
    return 0;
}

/* remove handler assoicate to that fd from store pending list */
/* return number of successfully free pending entries */
int storeUnregister(e, fd)
StoreEntry *e;
int fd;
{
    int i;
    int freed = 0;

    /* walk the entire list looking for matched fd */
    for (i = 0; i < e->pending_list_size; i++) {
	if (e->pending[i] && (e->pending[i]->fd == fd)) {
	    /* found the match fd */
	    safe_free(e->pending[i]);
	    e->pending[i] = NULL;
	    freed++;
	}
    }

    return freed;
}

/* Call handlers waiting for  data to be appended to E. */
static void InvokeHandlers(e)
StoreEntry *e;
{
    int i;

    /* walk the entire list looking for valid handlers */
    for (i = 0; i < e->pending_list_size; i++) {
	if (e->pending[i] && e->pending[i]->handler) {
	    /* 
	     *  Once we call the handler, it is no longer needed 
	     *  until the write process sends all available data 
	     *  from the object entry. 
	     */
	    (e->pending[i]->handler) (e->pending[i]->fd, e, e->pending[i]->data);
	    safe_free(e->pending[i]);
	    e->pending[i] = NULL;
	}
    }
}

/* Append incoming data from a primary server to an entry. */
int storeAppend(e, data, len)
StoreEntry *e;
char *data;
int len;
{
    /* validity check -- sometimes it's called with bogus values */
    if (e == NULL || e->data == NULL) {
	debug(0, "storeAppend: Invalid StoreEntry (%s), aborting...\n",
	      e == NULL ? "NULL entry" : "NULL data entry");
	return -1;
    }
    debug(5, "storeAppend: appending %d bytes for %s\n", len, e->url);

#ifdef DW_QUESTIONABLE_CODE
    if (e->e_current_len + len > store_mem_high) {
	debug(0, "storeAppend: object length about to exceed high water mark!\n");
	/*      NOTE: this just gets called repeatedly.  Need to do     */
	/*      something smarter, like call storeAbort()?  -DW         */
	InvokeHandlers(e);
	return -1;
    }
#endif

    /* get some extra storage if needed */
    if (storeGetMemSpace(len, 0) < 0) {
	debug(0, "storeAppend: Unable to allocate %d bytes, aborting...\n",
	      len);
	/* AWT: should we invokehandler here ? */
	/* InvokeHandlers(e); */
	return -1;
    }
    store_mem_size += len;
    debug(8, "storeAppend: growing store_mem_size by %d\n", len);
    debug(8, "storeAppend: store_mem_size = %d\n", store_mem_size);

    (void) e->data->mem_append(e->data, data, len);
    e->e_current_len += len;
    debug(8, "storeAppend: e_current_len = %d\n", e->e_current_len);

    if ((e->status != STORE_ABORTED) && !(e->flag & DELAY_SENDING))
	InvokeHandlers(e);

    return 0;
}


/* return full name to swapfile */
char *storeSwapFullPath(fn, fullpath)
int fn;
char *fullpath;
{
    static char fullfilename[MAX_FILE_NAME_LEN];

    if (fullpath) {
	sprintf(fullpath, "%s/%02d/%d", swappath, fn % SWAP_DIRECTORIES, fn);
	return fullpath;
    }
    fullfilename[0] = '\0';
    sprintf(fullfilename, "%s/%02d/%d", swappath, fn % SWAP_DIRECTORIES, fn);
    return fullfilename;
}

/* swapping in handle */
int storeSwapInHandle(fd, buf, len, flag, e, offset)
int fd;
char *buf;
int len;
int flag;
StoreEntry *e;
int offset;
{
    debug(2, "storeSwapInHandle: <URL:%s>\n", e->url);

    if ((flag < 0) && (flag != DISK_EOF)) {
	debug(0, "storeSwapInHandle: SwapIn failure (err code = %d).\n", flag);
	e->mem_status = NOT_IN_MEMORY;
	return -1;
    }
    debug(5, "storeSwapInHandle: e->swap_offset   = %d\n", e->swap_offset);
    debug(5, "storeSwapInHandle: len              = %d\n", len);
    debug(5, "storeSwapInHandle: e->e_current_len = %d\n", e->e_current_len);
    debug(5, "storeSwapInHandle: e->object_len    = %d\n", e->object_len);

    e->swap_offset += len;
    storeAppend(e, buf, len);

    if (e->e_current_len < e->object_len) {
	/* some more data to swap in, reschedule */
	file_read(e->swap_fd, e->e_swap_buf, SWAP_BUF, e->swap_offset,
		  (FILE_READ_HD) storeSwapInHandle, (caddr_t) e);
    } else {
	/* complete swapping in */
	e->mem_status = IN_MEMORY;
	safe_free(e->e_swap_buf);
	file_close(e->swap_fd);
	debug(5, "storeSwapInHandle: SwapIn complete: <URL:%s> from %s.\n",
	      e->url, storeSwapFullPath(e->swap_file_number, NULL));
    }
    return 0;
}

/* start swapping in */
int storeSwapInStart(e)
StoreEntry *e;
{
    int fd;
#ifdef METADATA_IN_SWAPFILE
    int storeSwapInHandleHdr();
#endif

    /* sanity check! */
    if (e->swap_file_number < 0) {
	debug(0, "storeSwapInStart: No filename? <URL:%s>\n", e->url);
	e->swap_fd = -1;
	return -1;
    }
    e->swap_fd = fd = file_open(storeSwapFullPath(e->swap_file_number, NULL),
				NULL, O_RDONLY);
    if (fd <= 0) {
	debug(0, "storeSwapInStart: Unable to open swapfile: %s\n",
	      storeSwapFullPath(e->swap_file_number, NULL));
	return -1;
    }
    debug(5, "storeSwapInStart: initialized swap file '%s' for <URL:%s>\n",
	  storeSwapFullPath(e->swap_file_number, NULL), e->url);

    e->data = memInit();
    ++meta_data.hot_vm;

    e->mem_status = SWAPPING_IN;
    e->swap_offset = 0;

    e->e_swap_buf = (char *) xcalloc(1, SWAP_BUF);

#ifdef METADATA_IN_SWAPFILE
    /* read the swapfile header */
    file_read(e->swap_fd, e->e_swap_buf, sizeof(swap_hdr), 0,
	      (FILE_READ_HD) storeSwapInHandleHdr, (caddr_t) e);
    return 0;
}

int storeSwapInHandleHdr(fd, buf, len, flag, e, offset)
int fd;
char *buf;
int len;
int flag;
StoreEntry *e;
int offset;
{
    debug(2, "storeSwapInHandleHdr: <URL:%s>\n", e->url);

    if ((flag < 0) && (flag != DISK_EOF)) {
	debug(0, "storeSwapInHandleHdr: SwapIn failure (err code = %d).\n",
	      flag);
	e->mem_status = NOT_IN_MEMORY;
	return -1;
    }
    if (len != sizeof(swap_hdr)) {
	debug(0, "storeSwapInHandleHdr: expected %d bytes, got %d.\n",
	      sizeof(swap_hdr), len);
	return -1;
    }
    debug(5, "storeSwapInHandleHdr: e->swap_offset   = %d\n", e->swap_offset);
    debug(5, "storeSwapInHandleHdr: len              = %d\n", len);
    debug(5, "storeSwapInHandleHdr: e->e_current_len = %d\n", e->e_current_len);
    debug(5, "storeSwapInHandleHdr: e->object_len    = %d\n", e->object_len);

    /*  OK, now we have the swap_hdr in e->e_swap_buf.  But we already  */
    /*  know everything it contains, so don't do anything with it.      */
    /*  Now schedule the real storeSwapInHandle().                      */

    e->swap_offset += len;

#endif

    /* start swapping daemon */
    file_read(e->swap_fd, e->e_swap_buf, SWAP_BUF, e->swap_offset,
	      (FILE_READ_HD) storeSwapInHandle, (caddr_t) e);
    return 0;
}

void storeSwapOutHandle(fd, flag, e)
int fd;
int flag;
StoreEntry *e;
{
    char msgbuf[MAX_URL + 256];
    char filename[MAX_FILE_NAME_LEN];

    debug(5, "storeSwapOutHandle: Called for %s\n", e->url);

    storeSwapFullPath(e->swap_file_number, filename);

    if (flag < 0) {
	debug(1, "storeSwapOutHandle: SwapOut failure (err code = %d).\n", flag);
	e->swap_status = NO_SWAP;
	safe_free(e->e_swap_buf);
	file_close(fd);
	if (e->swap_file_number != -1)
	    unlink(filename);	/* remove it */

	if (flag == DISK_NO_SPACE_LEFT) {
	    /* reduce the swap_size limit to the current size. */
	    setCacheSwapMax(store_swap_size);
	    store_swap_high = (long) (((float) getCacheSwapMax() *
		     (float) getCacheSwapHighWaterMark()) / (float) 100);
	    store_swap_low = (long) (((float) getCacheSwapMax() *
		      (float) getCacheSwapLowWaterMark()) / (float) 100);
	}
	return;
    }
    debug(6, "storeSwapOutHandle: e->swap_offset    = %d\n", e->swap_offset);
    debug(6, "storeSwapOutHandle: e->e_swap_buf_len = %d\n", e->e_swap_buf_len);
    debug(6, "storeSwapOutHandle: e->object_len     = %d\n", e->object_len);
    debug(6, "storeSwapOutHandle: store_swap_size   = %dk\n", store_swap_size);

    e->swap_offset += e->e_swap_buf_len;
    store_swap_size += ((e->e_swap_buf_len + 1023) >> 10);	/* round up */
    if (e->swap_offset >= e->object_len) {	/* swapping complete */
	e->swap_status = SWAP_OK;
	safe_free(e->e_swap_buf);
	file_close(e->swap_fd);
	debug(5, "storeSwapOutHandle: SwapOut complete: <URL:%s> to %s.\n",
	      e->url, storeSwapFullPath(e->swap_file_number, NULL));

#ifndef METADATA_IN_SWAPFILE

#ifdef EXPIRES_IN_CACHELOG
	sprintf(msgbuf, "FILE: %s URL: %s %ld\n",
	    filename,
	    e->url,
	    cached_curtime + e->ttl);
#else /* EXPIRES_IN_CACHELOG */
	sprintf(msgbuf, "FILE: %s URL: %s\n", filename, e->url);
#endif /* EXPIRES_IN_CACHELOG */
	file_write(swaplog_fd, msgbuf, strlen(msgbuf), swaplog_lock, NULL, NULL);
#endif /* METADATA_IN_SWAPFILE */

	return;
    }
    /* write some more data, reschedule itself. */
    storeCopy(e, e->swap_offset, SWAP_BUF, e->e_swap_buf, &(e->e_swap_buf_len));
    file_write(e->swap_fd, e->e_swap_buf, e->e_swap_buf_len, e->e_swap_access,
	       storeSwapOutHandle, e);
    return;

}


/* start swapping object to disk */
int storeSwapOutStart(e)
StoreEntry *e;
{
    int i, fd;
    struct stat buf;
    char swapfilename[MAX_FILE_NAME_LEN];
#ifdef METADATA_IN_SWAPFILE
    swap_hdr hdr;
#endif


    /* avoid for duplicate filename to in-used file. */
    i = 0;
    do {
	do {
	    swapfileno = (++swapfileno) % MAX_SWAP_FILE;
	    storeSwapFullPath(swapfileno, swapfilename);
	    i++;
	} while (file_get_fd(swapfilename) != DISK_FILE_NOT_FOUND);
    } while ((stat(swapfilename, &buf) >= 0) && (i < MAX_SWAP_FILE));

    if (i == MAX_SWAP_FILE) {
	debug(0, "storeSwapOutStart: Too many swap files (> %d).\n", i);
	return -1;
    }
    /* open a file */
    fd = file_open(swapfilename, NULL, O_RDWR | O_CREAT);
    if (fd < 0) {
	debug(0, "storeSwapOutStart: Unable to open swapfile %s \n", swapfilename);
	return -1;
    }
    e->swap_fd = fd;
    debug(5, "storeSwapOutStart: Begin SwapOut <URL:%s> to %s.\n",
	  e->url, swapfilename);

    e->swap_file_number = swapfileno;
    if ((e->e_swap_access = file_write_lock(e->swap_fd)) < 0) {
	debug(0, "storeSwapOutStart: Unable to write to swapfile: %s\n",
	      swapfilename);
	return -1;
    }
    e->swap_status = SWAP_IN_PROGRESS;
    e->swap_offset = 0;
    e->e_swap_buf = (char *) xcalloc(1, SWAP_BUF);
    e->e_swap_buf_len = 0;
#ifdef METADATA_IN_SWAPFILE
    hdr.expires = cached_curtime + e->ttl;
    memset(hdr.url, '\0', MAX_URL);
    strcpy(hdr.url, e->url);

    storeGetSwapSpace(e->object_len);

    file_write(e->swap_fd, (char *) &hdr, sizeof(hdr), e->e_swap_access,
	       storeSwapOutHandle, e);
#else
    storeCopy(e, 0, SWAP_BUF, e->e_swap_buf, &(e->e_swap_buf_len));

    /* start swapping daemon */
    file_write(e->swap_fd, e->e_swap_buf, e->e_swap_buf_len, e->e_swap_access,
	       storeSwapOutHandle, e);
#endif
    return 0;
}

#ifdef METADATA_IN_SWAPFILE

#include <dirent.h>
void storeRebuildFromDisk()
{
    struct _list_t {
	char *data;
	struct _list_t *next;
    };
    struct _list_t *dir_head = NULL;
    struct _list_t *dir_tail = NULL;
    struct _list_t *dir_tmp = NULL;
    DIR *dirp = NULL;
    struct dirent *d;
    struct stat sb;
    swap_hdr hdr;
    int fd;
    char pathname[MAXPATHLEN];
    time_t start, r;
    int objcount = 0;
    StoreEntry *e = NULL;

    debug(1, "Rebuilding storage from disk image in %s\n", swappath);
    start = time(NULL);

    dir_tmp = (struct _list_t *) xmalloc(sizeof(struct _list_t));
    dir_tmp->data = xstrdup(swappath);
    dir_tmp->next = NULL;
    dir_head = dir_tail = dir_tmp;

    while (dir_head) {
	cached_curtime = time(NULL);
	if ((dirp = opendir(dir_head->data)) == (DIR *) NULL) {
	    perror(dir_head->data);
	    continue;
	}
	debug(2, "storeRebuildFromDisk: Reading %s\n", dir_head->data);

	while (d = readdir(dirp)) {
	    if (d->d_name[0] == '.')
		continue;
	    sprintf(pathname, "%s/%s", dir_head->data, d->d_name);
	    debug(4, "storeRebuildFromDisk: Checking %s\n", pathname);
	    if (stat(pathname, &sb) < 0)	/* NOTE: stat(2) sucks.  We */
		continue;	/* need a better way to ID  */
	    if (S_ISDIR(sb.st_mode)) {	/* swap files vs. dirs -DW  */
		dir_tmp = (struct _list_t *) xmalloc(sizeof(struct _list_t));
		dir_tmp->data = xstrdup(pathname);
		dir_tmp->next = NULL;
		dir_tail->next = dir_tmp;
		dir_tail = dir_tmp;
		continue;
	    }
	    if (!S_ISREG(sb.st_mode))
		continue;
	    if (sb.st_size < sizeof(swap_hdr))
		continue;
	    if ((fd = open(pathname, O_RDONLY)) < 0) {
		unlink(pathname);
		continue;
	    }
	    if (read(fd, &hdr, sizeof(swap_hdr)) != sizeof(swap_hdr)) {
		close(fd);
		unlink(pathname);
		continue;
	    }
	    close(fd);
	    if (hdr.expires < cached_curtime) {
		debug(3, "storeRebuildFromDisk: Expired %s\n", hdr.url);
		unlink(pathname);
		continue;
	    }
	    if (e = storeGet(hdr.url)) {
		debug(3, "storeRebuildFromDisk: Duplicate %s\n", hdr.url);
		if (hdr.expires > (e->ttl + e->lastref))
		    storeRelease(e);
		else {
		    unlink(pathname);
		    continue;
		}
	    }
	    debug(4, "storeRebuildFromDisk: Adding (TTL=%ld) %s\n",
		  hdr.expires - cached_curtime, hdr.url);
	    store_swap_size += ((sb.st_size + 1023) >> 10);	/* round up */
	    storeAddDiskRestore(hdr.url,
				pathname,
				sb.st_size - sizeof(swap_hdr),
				hdr.expires - cached_curtime);
	    CacheInfo->proto_newobject(CacheInfo,
				       CacheInfo->proto_id(hdr.url),
				       sb.st_size - sizeof(swap_hdr));
	    objcount++;
	}

	closedir(dirp);
	dir_tmp = dir_head->next;
	xfree(dir_head->data);
	xfree(dir_head);
	dir_head = dir_tmp;
    }

    r = time(NULL) - start;
    debug(1, "Finished rebuilding storage from disk image.\n");
    debug(1, "  Loaded %d objects.\n", objcount);
    debug(1, "  Took %d seconds (%6.1lf objects/sec).\n",
	  r > 0 ? r : 0, (double) objcount / (r > 0 ? r : 1));
}

#else				/* METADATA_IN_SWAPFILE */

/* recreate meta data from disk image in swap directory */
void storeRebuildFromDisk()
{
    int objcount = 0;
    int linecount = 0;
    char line_in[4096];
    char swapfile[1024];
    char url[MAX_URL];
    char tmp[MAX_URL];
    char filename_scan[MAX_URL];
    int tmp_int1, tmp_int2;
    StoreEntry *e;
    struct stat buf;
    time_t start, stop, r;
#ifdef EXPIRES_IN_CACHELOG
    time_t expires;
#endif


    debug(1, "Rebuilding storage from disk image in %s\n", swappath);
    start = time(NULL);

    /* go to top of the file */
    (void) lseek(swaplog_fd, 0L, SEEK_SET);

    /* check if / at the end of path */
    tmp[0] = '\0';
    sprintf(tmp, "%s", swappath);
    if (tmp[strlen(tmp) - 1] == '/') {
	tmp[strlen(tmp) - 1] = '\0';
    }
    /* create scan pattern for file name */
    filename_scan[0] = '\0';
    sprintf(filename_scan, "%s/%%d/%%d", tmp);

    memset(line_in, '\0', 4096);
    while (fgets(line_in, 4096, swaplog_stream)) {

	if ((linecount++ & 0x7F) == 0)	/* update timestamp */
	    cached_curtime = time(NULL);

	debug(10, "line_in: %s", line_in);
	if ((line_in[0] == '\0') || (line_in[0] == '\n') ||
	    (line_in[0] == '#'))
	    continue;		/* skip bad lines */

	url[0] = swapfile[0] = '\0';

#ifdef EXPIRES_IN_CACHELOG
	if (sscanf(line_in, "FILE: %[^ ] URL: %[^\n ] %ld", swapfile, url, &expires) != 3)
#else
	if (sscanf(line_in, "FILE: %[^ ] URL: %[^\n ]", swapfile, url) != 2)
#endif
	    continue;

#ifdef EXPIRES_IN_CACHELOG
	if (expires < cached_curtime) {
	    debug(3, "storeRebuildFromDisk: Expired: <URL:%s>\n", url);
	    unlink (swapfile);
	    continue;
	}
#endif

	if (stat(swapfile, &buf) < 0) {
	    debug(3, "storeRebuildFromDisk: Cached file missing: <URL:%s>: %s: %s.\n", url, swapfile, strerror(errno));
	    continue;
	}
	debug(10, "storeRebuildFromDisk: Cached file exists: <URL:%s>: %s\n",
	      url, swapfile);

	if (sscanf(swapfile, filename_scan, &tmp_int1, &tmp_int2) != 2) {
	    debug(1, "storeRebuildFromDisk: swapfileno sscanf failed on '%s'\n",
		  swapfile);
	    continue;
	}
	/* use the lastest file_no as swapfileno */
	/* swapping out check for file existing  */
	/* and avoid the duplicate anyway */
	swapfileno = tmp_int2;

	/* file is still there */
	objcount++;

	/* update store_swap_size */
	store_swap_size += ((buf.st_size + 1023) >> 10);

	if (e = storeGet(url)) {
	    debug(6, "storeRebuildFromDisk: Found a duplicate for <URL:%s> -- replacing with newer copy.\n", url);
	    fprintf(swaplog_tmp_stream, "FILE: %s URL: %s\n",
		    swapfile, url);
	    e->swap_file_number = swapfileno;
	    e->object_len = buf.st_size;
	    e->e_current_len = 0;
#ifdef EXPIRES_IN_CACHELOG
	    e->ttl = expires - cached_curtime;
#endif
	} else {
	    /* create it */
	    fprintf(swaplog_tmp_stream, "FILE: %s URL: %s\n",
		    swapfile, url);
	    storeAddDiskRestore(url, swapfile, buf.st_size,
#ifdef EXPIRES_IN_CACHELOG
				expires - cached_curtime);
#else
				proto_defaultTTL(url));
#endif
	    CacheInfo->proto_newobject(CacheInfo,
				       CacheInfo->proto_id(url),
				       buf.st_size);
	}
    }

    fflush(swaplog_tmp_stream);
    stop = time(NULL);
    r = stop - start;
    debug(1, "Finished rebuilding storage from disk image.\n");
    debug(1, "  Read %d lines from previous logfile.\n", linecount);
    debug(1, "  Loaded %d objects.\n", objcount);
    debug(1, "  Took %d seconds (%6.1lf objects/sec).\n",
	  r > 0 ? r : 0, (double) objcount / (r > 0 ? r : 1));
}

#endif				/* METADATA_IN_SWAPFILE */


/* return current swap size in kilo-bytes */
int storeGetSwapSize()
{
    return store_swap_size;
}

/* return current swap size in bytes */
int storeGetMemSize()
{
    return store_mem_size;
}


/* Complete transfer into the local cache.  */
int storeComplete(e)
StoreEntry *e;
{
    debug(5, "storeComplete: <URL:%s>\n", e->url);

    /* XXX There should be some way to determine object_len independently. */
    if (e->object_len == 0) {
	e->object_len = e->e_current_len;
	/* Handlers won't have deduced end of object if  object_len is 0; call,
	   so they can send DATAEND. */
	InvokeHandlers(e);
    }
    e->lastref = e->timestamp = time(0);
    e->status = STORE_OK;
    e->mem_status = IN_MEMORY;
    e->swap_status = NO_SWAP;
    /* start writing it to disk, exclude cache_object */
    if ((strncmp("cache_obj", e->url, strlen("cache_obj")) != 0) &&
	(e->flag & CACHABLE) && !(e->flag & RELEASE_REQUEST)) {
	storeSwapOutStart(e);
    }
    /* free up incoming MIME */
    safe_free(e->mime_hdr);
    CacheInfo->proto_newobject(CacheInfo, CacheInfo->proto_id(e->url),
			       e->object_len);
    if (e->flag & RELEASE_REQUEST)
	storeRelease(e);
}


int storeAbort(e, msg)
StoreEntry *e;
char *msg;
{
    char mime_hdr[300];
    char abort_msg[2000];

    debug(6, "storeAbort: <URL:%s>\n", e->url);
    e->ttl = getNegativeTTL();
    e->status = STORE_ABORTED;
    e->mem_status = IN_MEMORY;
    /* No DISK swap for negative cached object */
    e->swap_status = NO_SWAP;
    e->lastref = e->timestamp = time(0);

    mk_mime_hdr(mime_hdr, (time_t) e->ttl, 6 + strlen(msg), (time_t) time(NULL), "text/html");
    if (msg) {
	/* This can run off the end here. Be careful */
	if ((strlen(msg) + strlen(mime_hdr) + 50) < 2000) {
	    sprintf(abort_msg, "HTTP/1.0 400 Cache Detected Error\r\n%s\r\n\r\n%s", mime_hdr, msg);
	} else {
	    debug(0, "storeAbort: WARNING: Must increase msg length!");
	}
	storeAppend(e, abort_msg, strlen(abort_msg));
	e->e_abort_msg = strdup(abort_msg);
	BIT_SET(e->flag, ABORT_MSG_PENDING);
    }
    e->object_len = e->e_current_len;

    /* Call handlers so they can report error. */
    InvokeHandlers(e);

    if (e->flag & RELEASE_REQUEST) {
	storeRelease(e);
	return 0;
    }
    return 0;

}

/* get the first entry in the storage */
StoreEntry *storeGetFirst()
{
    hash_link *first;

    if (table == (HashID) 0)
	return NULL;
    first = hash_first(table);
    return first ? ((StoreEntry *) first->item) : NULL;
}


/* get the next entry in the storage for a given search pointer */
StoreEntry *storeGetNext()
{
    hash_link *next;

    if (table == (HashID) 0)
	return NULL;
    next = hash_next(table);
    return next ? ((StoreEntry *) next->item) : NULL;
}



/* walk through every single entries in the storage and invoke a given routine */
int storeWalkThrough(proc, data)
int (*proc) _PARAMS((StoreEntry * e, caddr_t data));
caddr_t data;
{
    StoreEntry *e;
    int count = 0;

    for (e = storeGetFirst(); e; e = storeGetNext())
	count += proc(e, data);
    return count;
}


/* compare an object timestamp and see if ttl is expired. Free it if so. */
/* return 1 if it expired, 0 if not */
int removeOldEntry(e, data)
StoreEntry *e;
caddr_t data;
{
    time_t curtime = *((time_t *) data);

    debug(5, "removeOldEntry: Checking: %s\n", e->url);
    debug(6, "removeOldEntry:   *       curtime: %8ld\n", curtime);
    debug(6, "removeOldEntry:   *  e->timestamp: %8ld\n", e->timestamp);
    debug(6, "removeOldEntry:   * time in cache: %8ld\n",
	  curtime - e->timestamp);
    debug(6, "removeOldEntry:   *  time-to-live: %8ld\n", e->ttl);

    if ((curtime - e->timestamp > e->ttl) && (e->status != STORE_PENDING)) {
	return (storeRelease(e) == 0 ? 1 : 0);
    }
    return 0;
}


/* free up all ttl-expired objects */
void storePurgeOld()
{
    int n;

    cached_curtime = time(0);
    debug(3, "storePurgeOld: Begin purging TTL-expired objects\n");
    n = storeWalkThrough(removeOldEntry, (caddr_t) & cached_curtime);
    debug(3, "storePurgeOld: Done purging TTL-expired objects.\n");
    debug(3, "storePurgeOld: %d objects expired\n", n);
}



/* find the least recent used entry in the storage. 
   The implementation is still very inefficient, better data structure
   is needed here. */

StoreEntry *storeLRU()
{
    StoreEntry *e, *lru = NULL;
    time_t long earliesttime = time(0);

    for (e = storeGetFirst(); e; e = storeGetNext()) {
	/* be sure to check status PENDING */

	if ((e->lastref < earliesttime) &&
	    (e->status != STORE_PENDING) &&
	    !(e->flag & OBJ_LOCK)) {
	    lru = e;
	    earliesttime = lru->lastref;
	}
    }

    return (lru);

}


#define MEM_LRUSCAN_BLOCK 16
#define MEM_MAX_HELP 5
/* Clear Memory storage to accommodate the given object len */
int storeGetMemSpace(size, check_vm_number)
int size;
int check_vm_number;
{
    static int n_help = 0;
    StoreEntry *LRU, *e;
    StoreEntry **LRU_list;
    int LRU_list_size = meta_data.store_entries;
    int n_scanned = 0;
    int n_pending = 0;
    int n_expired = 0;
    int n_aborted = 0;
    int n_purged = 0;
    int n_released = 0;
    int n_items, i;
    int compareLastRef();
    int n_inmem = 0;		/* extra debugging */
    int n_cantpurge = 0;	/* extra debugging */
    int mem_cantpurge = 0;	/* extra debugging */

    if (!check_vm_number && ((store_mem_size + size) < store_mem_high))
	return 0;

    LRU_list = (StoreEntry **) xcalloc(LRU_list_size, sizeof(StoreEntry *));

    for (n_items = 0, e = storeGetFirst(); e; e = storeGetNext()) {
	if (n_items >= LRU_list_size) {
	    /* have to realloc LRU_list */
	    LRU_list_size += MEM_LRUSCAN_BLOCK;
	    LRU_list = (StoreEntry **) realloc(LRU_list, LRU_list_size * sizeof(StoreEntry *));
	}
	n_scanned++;

	if (e->mem_status != IN_MEMORY)
	    continue;

	n_inmem++;

	if (e->status == STORE_PENDING) {
	    n_pending++;
	    continue;
	}
	if ((cached_curtime - e->timestamp > e->ttl) && (e->swap_status != SWAP_IN_PROGRESS)) {
	    debug(2, "storeGetMemSpace: Expired: %s\n", e->url);
	    n_expired++;
	    storeRelease(e);
	    continue;
	}
	if ((e->swap_status == SWAP_OK) &&
	    !(e->flag & OBJ_LOCK)) {
	    LRU_list[n_items++] = e;
	} else if ((e->status == STORE_ABORTED) &&
		   (e->swap_status == NO_SWAP) &&
		   !(e->flag & OBJ_LOCK)) {
	    n_aborted++;
	    LRU_list[n_items++] = e;
	} else {
	    n_cantpurge++;
	    mem_cantpurge += e->e_current_len;
	    debug(5, "storeGetMemSpace: Can't purge %7d bytes: %s\n",
		  e->e_current_len, e->url);
	    if (e->swap_status != SWAP_OK)
		debug(5, "storeGetMemSpace: --> e->swap_status != SWAP_OK\n");
	    if (e->flag & OBJ_LOCK)
		debug(5, "storeGetMemSpace: --> e->flag & OBJ_LOCK\n");
	}
    }
    debug(2, "storeGetMemSpace: Current size:     %7d bytes\n", store_mem_size);
    debug(2, "storeGetMemSpace: High W Mark:      %7d items\n", store_mem_high);
    debug(2, "storeGetMemSpace: Low W Mark:       %7d items\n", store_mem_low);
    debug(2, "storeGetMemSpace: Entry count:      %7d items\n", meta_data.store_entries);
    debug(2, "storeGetMemSpace: Scanned:          %7d items\n", n_scanned);
    debug(2, "storeGetMemSpace: In memory:        %7d items\n", n_inmem);
    debug(2, "storeGetMemSpace: Hot vm count:     %7d items\n", meta_data.hot_vm);
    debug(2, "storeGetMemSpace: Expired:          %7d items\n", n_expired);
    debug(2, "storeGetMemSpace: Negative Cached:  %7d items\n", n_aborted);
    debug(2, "storeGetMemSpace: Can't purge:      %7d items\n", n_cantpurge);
    debug(2, "storeGetMemSpace: Can't purge size: %7d bytes\n", mem_cantpurge);
    debug(2, "storeGetMemSpace: Sorting:          %7d items in LRU_list...\n", n_items);
    qsort((char *) LRU_list, n_items, sizeof(e), compareLastRef);

    /* Kick LRU out until we have enough memory space */

    if (check_vm_number) {
	/* look for vm slot */
	for (i = 0; (i < n_items) && (meta_data.hot_vm > HOT_OBJECT_LOW); ++i) {
	    if (LRU = LRU_list[i])
		if ((LRU->status == STORE_ABORTED) && (LRU->swap_status == NO_SWAP)) {
		    n_released++;
		    storeRelease(LRU);
		} else {
		    n_purged++;
		    storePurgeMem(LRU);
		}
	}
    } else {
	/* look for space */
	for (i = 0; (i < n_items) && ((store_mem_size + size) > store_mem_low); ++i) {
	    if (LRU = LRU_list[i])
		if ((LRU->status == STORE_ABORTED) && (LRU->swap_status == NO_SWAP)) {
		    n_released++;
		    storeRelease(LRU);
		} else {
		    n_purged++;
		    storePurgeMem(LRU);
		}
	}
    }

    safe_free(LRU_list);

    debug(2, "storeGetMemSpace: After freeing size: %7d bytes\n", store_mem_size);
    debug(2, "storeGetMemSpace: Purged:             %7d items\n", n_purged);
    debug(2, "storeGetMemSpace: Released:           %7d items\n", n_released);

    if ((store_mem_size + size) < store_mem_high)
	return 0;

    /* The else case: */
    debug(0, "storeGetMemSpace: Help! Nothing to free with %d allocated.\n",
	  store_mem_size + size);
    if (++n_help > MEM_MAX_HELP) {
	fatal_dump("storeGetMemSpace: Repeated failures. Exiting");
    }
    return -1;
}

int compareLastRef(e1, e2)
StoreEntry **e1, **e2;
{
    if (!e1 || !e2)
	abort();

    if ((*e1)->lastref > (*e2)->lastref)
	return (1);

    if ((*e1)->lastref < (*e2)->lastref)
	return (-1);

    return (0);
}

#define SWAP_LRUSCAN_BLOCK 16
#define SWAP_MAX_HELP 5
/* Clear Swap storage to accommodate the given object len */
int storeGetSwapSpace(size)
int size;
{
    static int swap_help = 0;
    StoreEntry *LRU, *e;
    StoreEntry **LRU_list;
    int LRU_list_count = 0;
    int scanned = 0;
    int removed = 0;
    int expired = 0;
    int pending = 0;
    int locked = 0;
    int locked_size = 0;
    int i;
    int LRU_cur_size = meta_data.store_entries;


    if (store_swap_size + ((size + 1023) >> 10) <= store_swap_high)
	return 0;

    LRU_list = (StoreEntry **) xcalloc(LRU_cur_size, sizeof(StoreEntry *));

    debug(1, "storeGetSwapSpace: Need %d bytes...\n", size);

    /* begin to scan for LRU and kick out any expired object if found */
    for (LRU_list_count = 0, e = storeGetFirst(); e; e = storeGetNext()) {

	scanned++;

	if (LRU_list_count >= LRU_cur_size) {
	    /* have to realloc  */
	    LRU_cur_size += SWAP_LRUSCAN_BLOCK;
	    LRU_list = (StoreEntry **) xrealloc((char *) LRU_list,
				    LRU_cur_size * sizeof(StoreEntry *));
	}
	if (e->status == STORE_PENDING) {
	    pending++;
	    continue;
	}
	if (cached_curtime - e->timestamp > e->ttl) {
	    debug(2, "storeGetSwapSpace: Expired: <%s>\n", e->url);
	    /* just call release. don't have to check for lock status.
	       storeRelease will take care of that and set a pending flag
	       if it's still locked. */
	    expired++;
	    storeRelease(e);
	    continue;
	}
	/* now see if it's ok to kick out */
	if (!(e->flag & OBJ_LOCK) || (e->swap_status != SWAP_IN_PROGRESS)) {
	    /* ok add it to a candidate list */
	    LRU_list[LRU_list_count++] = e;
	} else {
	    debug(2, "storeGetSwapSpace: Can't purge %7d bytes: <%s>\n",
		  e->e_current_len, e->url);
	    if (e->flag & OBJ_LOCK) {
		debug(2, "                   e->flag & OBJ_LOCK\n");
	    }
	    if (e->swap_status == SWAP_IN_PROGRESS) {
		debug(2, "                   e->swap_status == SWAPP_IN_PROGRESS\n");
	    }
	    locked++;
	    locked_size += e->e_current_len;
	}

    }

    /* end of candidate selection */
    debug(2, "storeGetSwapSpace: Current Size:   %7d kbytes\n", store_swap_size);
    debug(2, "storeGetSwapSpace: High W Mark:    %7d kbytes\n", store_swap_high);
    debug(2, "storeGetSwapSpace: Low W Mark:     %7d kbytes\n", store_swap_low);
    debug(2, "storeGetSwapSpace: Entry count:    %7d items\n", meta_data.store_entries);
    debug(2, "storeGetSwapSpace: Scanned:        %7d items\n", scanned);
    debug(2, "storeGetSwapSpace: Expired:        %7d items\n", expired);
    debug(2, "storeGetSwapSpace: Locked:         %7d items\n", locked);
    debug(2, "storeGetSwapSpace: Locked Space:   %7d bytes\n", locked_size);
    debug(2, "storeGetSwapSpace: LRU candidate:  %7d items\n", LRU_list_count);

    /* sort LRU candidate list */
    qsort((char *) LRU_list, LRU_list_count, sizeof(e), compareLastRef);


    /* Kick LRU out until we have enough swap space */
    for (i = 0;
	 (i < LRU_list_count) &&
	 (store_swap_size + ((size + 1023) >> 10) > store_swap_low);
	 ++i) {

	if ((LRU = LRU_list[i]) != NULL) {
	    if (storeRelease(LRU) == 0) {
		removed++;
	    } else {
		debug(2, "storeGetSwapSpace: Help! Can't remove objects. <%s>\n",
		      LRU->url);
	    }
	} else {
	    break;
	}

    }
    debug(2, "storeGetSwapSpace: After Freeing Size:   %7d kbytes\n", store_swap_size);

    /* free the list */
    safe_free(LRU_list);

    if ((store_swap_size + ((size + 1023) >> 10) > store_swap_high)) {
	debug(0, "storeGetSwapSpace: Help! Nothing to free with %d Kbyte allocated\n",
	      store_swap_size);
	debug(0, "                   Ask for %d bytes\n", size);
	++swap_help;
	if (swap_help > SWAP_MAX_HELP) {
	    /* dump core for tracing back */
	    fatal_dump("storeGetSwapSpace: Repeat Errors. Exiting.");
	    return -1;
	}
    }
    /* else */
    debug(1, "storeGetSwapSpace: Finished, %d objects removed.\n", removed);
    return 0;
}


/* release an object from a cache */
/* return 0 when success. */
int storeRelease(e)
StoreEntry *e;
{
    StoreEntry *result = NULL;
    StoreEntry *head_result = 0;
    hash_link *hptr;

    debug(5, "storeRelease: Releasing: %s\n", e->url);

    if (table == (HashID) 0) {
	return -1;
    }
    if ((hptr = hash_lookup(table, e->key)) == NULL) {
	debug(5, "storeRelease: Not Found: %s\n", e->url);
	return -1;
    }
    result = hptr->item;

    if (result != e) {
	debug(1, "storeRelease: Duplicated entry found? <URL:%s>\n",
	      result->url ? result->url : "NULL");
	/*      Don't return here, we have to do something.     */
	/*      Otherwise storeGetSwapSpace() loops forever.    */
	/* return -1; */
    }
    if (e->flag & OBJ_LOCK) {
	BIT_SET(e->flag, RELEASE_REQUEST);
	return -1;
    }
    if (e->type_id == REQUEST_OP_GET) {
	/* check if coresponding HEAD object exists. */
	hash_link *head_table_entry;
	head_table_entry = hash_lookup(table, storeGenerateKey(e->url, REQUEST_OP_HEAD));
	if (head_table_entry) {
	    head_result = head_table_entry->item;
	    if (head_result) {
		/* recursive call here to free up /head/ */
		storeRelease(head_result);
	    }
	}
    }
    debug(3, "storeRelease: Release object key: %s \n", e->key);

    if (e->swap_status == SWAP_OK) {
	(void) unlink(storeSwapFullPath(e->swap_file_number, NULL));
	store_swap_size -= (e->object_len + 1023) >> 10;
    } else if (e->swap_status == SWAP_IN_PROGRESS) {
	/* swapping is in progress. Stop it. Clean up file descriptor */
	store_swap_size -= ((e->swap_offset + 1023) >> 10);	/* round up */
	safe_free(e->e_swap_buf);
	file_close(e->swap_fd);
    }
    CacheInfo->proto_purgeobject(CacheInfo,
				 CacheInfo->proto_id(e->url),
				 e->object_len);
    hash_delete_link(table, hptr);
    storeFreeEntry(e);
    return 0;
}


/* store change key */
void storeChangeKey(e)
StoreEntry *e;
{
    StoreEntry *result = 0;
    char key[MAX_URL + 32];

    if (!e)
	return;

    if (table != (HashID) 0) {
	hash_link *table_entry = hash_lookup(table, e->key);
	if (table_entry)
	    result = table_entry->item;
	if (result == e) {
	    hash_delete_link(table, table_entry);

	    key[0] = '\0';
	    sprintf(key, "/x%d/%s", keychange_count++, e->key);
	    if (!(result->flag & KEY_URL))
		safe_free(result->key);
	    result->key = strdup(key);

	    hash_insert(table, e->key, e);
	    BIT_SET(result->flag, KEY_CHANGE);
	    BIT_RESET(result->flag, KEY_URL);
	} else {
	    debug(1, "storeChangeKey: Key is not unique for key: %s\n", e->key);
	}
    }
}

/* return if the current key is the original one. */
int storeOriginalKey(e)
StoreEntry *e;
{
    if (!e)
	return 1;

    return !(e->flag & KEY_CHANGE);
}

/* return 1 if a store entry is locked */
int storeEntryLocked(e)
StoreEntry *e;
{
    if (!e)
	return 0;

    /* if store is pending, object is considered locked too. */
    return ((e->flag & OBJ_LOCK) || (e->status == STORE_PENDING));
}


int storeCopy(e, stateoffset, maxSize, buf, size)
StoreEntry *e;
int stateoffset;
int maxSize;
char *buf;
int *size;
{
    int available_to_write = e->e_current_len - stateoffset;

    *size = (available_to_write >= maxSize) ?
	maxSize : available_to_write;

    debug(6, "storeCopy: avail_to_write=%d, store_offset=%d\n",
	  *size, stateoffset);

    if (*size > 0)
	(void) e->data->mem_copy(e->data, stateoffset, buf, *size);

    return *size;
}



/*
 * Go through the first 300 bytes of MIME header of a cached object, returning
 * fields that match.
 */
char *
 storeMatchMime(e, mimehdr, buf, buf_len)
StoreEntry *e;
char *mimehdr;
char *buf;
int buf_len;
{
    int i;

    int offset = storeGrep(e, mimehdr, 300);

    buf[0] = '\0';

    if (offset) {
	storeCopy(e, offset, buf_len, buf, &buf_len);
	for (i = 0; i < buf_len; ++i) {
	    if (buf[i] == '\r') {
		buf[i] = ' ';
	    }			/* strip \r */
	    if (buf[i] == '\n') {
		buf[i] = '\0';
		break;
	    }
	}
	buf[buf_len - 1] = '\0';	/* always terminate at the end */
    }
    return (buf);
}

int storeGrep(e, string, nbytes)
StoreEntry *e;
char *string;
int nbytes;
{
    if (e && e->data && (nbytes > 0))
	return e->data->mem_grep(e->data, string, nbytes);

    return 0;
}


int storeEntryValidToSend(e)
StoreEntry *e;
{
    if ((time(0) - e->timestamp < e->ttl) ||
	(e->status == STORE_PENDING))
	return 1;
    else
	return 0;
}

int storeInit(cache_dir)
char *cache_dir;
{
    char tmpbuf[FATAL_BUF_SIZE];
    char swaplog_file[MAX_FILE_NAME_LEN];
    char swaplog_tmp_file[MAX_FILE_NAME_LEN];
    int statret;
    int directory_created = 0;
    struct stat buf;

    debug(10, "storeInit: Creating swap space in %s\n", cache_dir);
    strncpy(swappath, cache_dir, MAX_FILE_NAME_LEN);
    statret = stat(cache_dir, &buf);
    if (statret < 0) {
	/* we need to create a directory for swap file here. */
	if (mkdir(cache_dir, 0777) < 0) {
	    sprintf(tmpbuf,
	    "Couldn't create %s for storage swap area.\n %s\n", cache_dir,
		    sys_errlist[errno]);
	    fatal(tmpbuf);
	}
	statret = stat(cache_dir, &buf);
	if (statret < 0) {
	    sprintf(tmpbuf, "Couldn't open %s for storage swap area.\n", cache_dir);
	    fatal(tmpbuf);
	}
	directory_created = 1;

    }
    if (zap_disk_store) {
	debug(1, "Zapping all objects on disk storage.\n");
	/* This could be dangerous, second copy of cache can destroy the existing
	   swap files of the previous cache. We may use rc file do it. */
	tmpbuf[0] = '\0';
	sprintf(tmpbuf, "rm -rf %s/{log,[0-9]*}", cache_dir);
	system(tmpbuf);
	debug(1, "Zapping objects on disk DONE.\n");
    }
#ifndef METADATA_IN_SWAPFILE
    sprintf(swaplog_file, "%s/log", cache_dir);

    swaplog_fd = file_open(swaplog_file, NULL, O_RDWR | O_CREAT);
    if (swaplog_fd < 0) {
	sprintf(tmpbuf, "Cannot open swap logfile: %s\n", swaplog_file);
	fatal(tmpbuf);
    }
    swaplog_stream = fdopen(swaplog_fd, "a+");
    if (!swaplog_stream) {
	sprintf(tmpbuf, "Cannot open a stream for swap logfile: %s\n", swaplog_file);
	fatal(tmpbuf);
    }
    swaplog_lock = file_write_lock(swaplog_fd);

    sprintf(swaplog_tmp_file, "%s/log_tmp", cache_dir);

    swaplog_tmp_fd = file_open(swaplog_tmp_file, NULL, O_RDWR | O_TRUNC | O_CREAT);
    if (swaplog_tmp_fd < 0) {
	sprintf(tmpbuf, "Cannot open swap tmp logfile: %s\n", swaplog_tmp_file);
	fatal(tmpbuf);
    }
    swaplog_tmp_stream = fdopen(swaplog_tmp_fd, "a+");
    if (!swaplog_tmp_stream) {
	sprintf(tmpbuf, "Cannot open a stream for swap tmp logfile: %s\n", swaplog_tmp_file);
	fatal(tmpbuf);
    }
    if (!(directory_created || zap_disk_store)) {
	storeRebuildFromDisk();

	/* rename tmp_log to log */
	file_close(swaplog_fd);
	unlink(swaplog_file);
	link(swaplog_tmp_file, swaplog_file);

	swaplog_fd = file_open(swaplog_file, NULL, O_RDWR | O_CREAT);
	if (swaplog_fd < 0) {
	    sprintf(tmpbuf, "Cannot reopen swap logfile: %s\n", swaplog_file);
	    fatal(tmpbuf);
	}
	swaplog_stream = fdopen(swaplog_fd, "a+");
	if (!swaplog_stream) {
	    sprintf(tmpbuf, "Cannot reopen a stream for swap logfile: %s\n", swaplog_file);
	    fatal(tmpbuf);
	}
    }
    /* close tmp_log file here. Possibly remove if it is not linked to
       the real one. */
    file_close(swaplog_tmp_fd);
    unlink(swaplog_tmp_file);
#else
    if (!(directory_created || zap_disk_store))
	storeRebuildFromDisk();
#endif


    /* Create Small directories in swap */
    if (directory_created || zap_disk_store) {
	int i;
	char name[1024];
	for (i = 0; i < SWAP_DIRECTORIES; i++) {
	    sprintf(name, "%s/%02d", cache_dir, i);
	    if (mkdir(name, 0755) < 0) {
		if (errno != EEXIST) {
		    sprintf(tmpbuf, "Couldn't creat %s for storage swap area.\n",
			    name);
		    fatal(tmpbuf);
		}
	    }
	}
    }
    store_mem_high = (long) (((float) getCacheMemMax() *
		      (float) getCacheMemHighWaterMark()) / (float) 100);
    store_mem_low = (long) (((float) getCacheMemMax() *
		       (float) getCacheMemLowWaterMark()) / (float) 100);

    store_swap_high = (long) (((float) getCacheSwapMax() *
		     (float) getCacheSwapHighWaterMark()) / (float) 100);
    store_swap_low = (long) (((float) getCacheSwapMax() *
		      (float) getCacheSwapLowWaterMark()) / (float) 100);

    store_initialized = 1;
    return 0;
}

/* 
 *  storeSanityCheck - verify that all swap storage areas exist, and
 *  are writable; otherwise, force -z.
 */
void storeSanityCheck(cache_dir)
char *cache_dir;
{
    char name[4096], tmpbuf[4096];
    int i;

    for (i = 0; i < SWAP_DIRECTORIES; i++) {
	sprintf(name, "%s/%02d", cache_dir, i);
	if (access(name, W_OK)) {
	    sprintf(tmpbuf, "WARNING: Cannot write to %s for storage swap area.\nForcing a *full restart* (e.g., cached -z)...", name);
	    if (syslog_enable)
		syslog(LOG_ALERT, tmpbuf);
	    fprintf(stderr, "cached: %s\n", tmpbuf);
	    zap_disk_store = 1;
	    return;
	}
    }
}

int urlcmp(url1, url2)
char *url1, *url2;
{
    if (!url1 || !url2)
	death();
    return (strcmp(url1, url2));
}

int parse_file_number(s)
char *s;
{
    int len;


    for (len = strlen(s); (len >= 0); --len) {
	if (s[len] == '/') {
	    return (atoi(&s[len + 1]));
	}
    }
    debug(1, "parse_file_number: Could not determine the swap file number from %s.\n", s);
    return (0);
}

/* XXX --- The following code is VERY unstable */
/*
 * XXX Stolen from hash.c - this whole method of looking up the pointer
 * cache should be revamped using proper hashing method.  At the moment, the
 * algorithim is hash the url, then sequential search from there.
 */
static int hash_url(s)
char *s;
{
    unsigned int i, j, n;

    j = strlen(s);
    for (i = j / 2, n = 0; i < j; i++)
    n ^= 271 * (unsigned) s[i];
    i = n ^ (j * 271);
    return i % MAX_POINTERS;
}

/* Lookup an object in the cache.
   return just a reference to object, don't start swapping in yet. */
StorePtrEntry *storeGetPointer(url)
char *url;
{
    hash_link *hptr = NULL;

    debug(5, "storeGetPointer: looking up %s\n", url);

    if (ptr_table != (HashID) 0) {
    if ((hptr = hash_lookup(ptr_table, url)) != NULL)
        return hptr->item;
    }
    return NULL;
}

StorePtrEntry *storeCheckCache(url)
char *url;
{
	StoreEntry *entry;
	StorePtrEntry *tmp;
	int i;
	int n;

	/* Check local storage */
	entry = storeGet(url);

	if(entry)
	{
		/* Build pseudo pointer entry to return */
		/* 
		 * XXX - Need to free the ptr entry if local is set
		 */
		tmp = (StorePtrEntry *) xcalloc(1, sizeof(StorePtrEntry));
		tmp->url = strdup(url);
		tmp->nr_ptrs = 1;
		tmp->object_len = entry->object_len;
		tmp->timestamp = entry->timestamp;
		tmp->local = 1;
		tmp->ptr = (StorePointer *) xcalloc(1, sizeof(StorePointer));
		tmp->ptr->ptr_url = xcalloc(strlen(icpGetHostUrl()) + strlen(url) 
				+ 10, 1);
		sprintf(tmp->ptr->ptr_url,"icp://%s/%s", icpGetHostUrl(), 
			url);
		tmp->ptr->ttl = entry->ttl;
		
		return tmp;
	}
	else
	{
		return storeGetPointer(url); 
	}
}

int storeAddPointer(url, pointer, ttl, timestamp, object_len)
char *url;
char *pointer;
u_num32 ttl;
u_num32 timestamp;
int object_len;
{
	StorePtrEntry *entry;
	StorePointer *tmp=0;
	StorePointer *tmp_prev=0;
	StorePointer *new=0;

	/* See if pointer exists */
	entry = storeGetPointer(url);

	/* 
	 * If the entry already exists, check the timestamps. 
	 * If they are the same, add the pointer, if the one we are
	 * trying to add is less, then discard it.  If the one we are 
	 * adding is newer, then delete everything else
	 *
	 * XXX Should probably check for a different size.
	 * XXX Some logging should happen in here
	 *
	 */
	if (entry)
	{
		if (timestamp == entry->timestamp)
		{
			/* Increment number of pointers */ 
			entry->nr_ptrs++;	
		}
		else if (timestamp < entry->timestamp)
		{
			/* Discard, too old */
			return 0;
		}
		else
		{
			/* Delete the old pointers */
			tmp = entry->ptr;
			while(tmp)
			{
				tmp_prev = tmp;
				tmp = tmp->next;
				
				/* Free mem associated with this entry */
				safe_free(tmp_prev->ptr_url);
				safe_free(tmp_prev);
			}

			entry->nr_ptrs = 1;
			entry->ptr = 0;
		}

	}

	else
	{
		/* Create the pointer entry */
    	entry = (StorePtrEntry *) xcalloc(1, sizeof(StorePtrEntry));

		/* NB: Unset fields should be 0 term'd */
    	entry->url = strdup(url);
    	entry->nr_ptrs = 1;
		entry->timestamp = timestamp;
		entry->object_len = object_len;

		/* Add the pointer entry */
    	if (ptr_table == (HashID) 0) {
    		ptr_table = (HashID) hash_create(urlcmp, 7921); /* XXX */
    	}

		if (ptr_table == -1)
		{
			/* XXX Log an error */
		}
    	hash_insert(ptr_table, entry->url, entry);
	}

	/* Add the pointer */
    new = (StorePointer *) xcalloc(1, sizeof(StorePointer));
    new->ptr_url = strdup(pointer);
    new->ttl = ttl;
	new->next = 0;

	/* Zero the hit data (Will have to rebuilt anyway) */
	entry->hit_data = 0;
	entry->hit_data_len = 0;

	if (entry->ptr)
	{
		/* find the end of the list */
		tmp = entry->ptr;
		while(tmp->next)
		{
			tmp = tmp->next;
		}

		/* Add to the end of the list */
		tmp->next = new;
	}
	else
	{
		entry->ptr = new;
	}
	

	return 1;

}

int storeDelPointer(url, pointer)
char *url;
char *pointer;
{
	StorePtrEntry *entry=0;
	StorePointer *tmp=0;
	StorePointer *tmp_last=0;

	/* See if pointer entry exists for this url */
    entry = storeGetPointer(url);

    if (entry)
    {
        /* Pointer exists see if we can find the entry to delete */
		tmp = entry->ptr;
		while(tmp)
		{
			/* 
			 * XXX Note, should make sure that the pointer aare equiv, not
			 * just equal strings, as potentially they could be different
			 * due to case in hostname and omission of default ports etc.
			 */
			if (strcmp(tmp->ptr_url, pointer)==0)
			{
			    /* Remove from the linked list */
				if (tmp_last)
				{
					tmp_last->next = tmp->next;
				}
				else
				{
					entry->ptr = tmp->next;
				}

				/* Decrement number of pointers */
				entry->nr_ptrs--;

				/* Free the memory associated with this pointer */
				safe_free(tmp->ptr_url);
				safe_free(tmp);

				/* Delete entry if there are no pointers left */
				if (!entry->nr_ptrs)
				{
					/* Delete this hash entry */
					hash_delete(ptr_table, entry->url);
				}
				return 1;
			}
			else
			{
				/* Move to the next item */
				tmp_last = tmp;
				tmp = tmp->next;
			}
		}
		return 0;
    }
	else
	{
		/* Pointer doesn't exist */
		/* XXX log this */
		return 0;
	}
}
