static char rcsid[] = "$Id: wais.c,v 1.18 1995/08/21 20:56:14 hardy Exp $";
/*
 *  File:         wais.c
 *  Description:  state machine for wais retrieval protocol (just open a
 *                connection to a wais gateway, like the CERN waisd).
 *                Based on John's gopher retrieval module.
 *  Author:       Edward Moy, Xerox PARC
 *  Created:      Tue Jun 20 13:07:42 PDT 1995
 *  Language:     C
 *
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include "config.h"
#ifdef USE_WAIS_RELAY
#include <sys/param.h>		/* For MAXHOSTNAMELEN. */
#include <sys/errno.h>
#include <stdlib.h>
#include <string.h>

#include "comm.h"
#include "store.h"
#include "stat.h"
#include "neighbors.h"
#include "url.h"
#include "ipcache.h"
#include "cache_cf.h"

extern int errno;
extern char *sys_errlist[];
extern char *dns_error_message;

typedef struct _waisdata {
    StoreEntry *entry;
    char host[MAXHOSTNAMELEN];
    int port;
    char *type;
    char *mime_hdr;
    char type_id;
    char request[MAX_URL];
} WAISData;

extern char *tmp_error_buf;

int wais_url_parser(url, host, port, request)
char *url;
char *host;
int *port;
char *request;
{
    strcpy(host, wais_relay_host);
    *port = wais_relay_port;
    strcpy(request, url);

    return 0;
}

/* This will be called when timeout on read. */
void waisReadReplyTimeout(fd, data)
int fd;
WAISData *data;
{
    debug(4, "waisReadReplyTimeout: Timeout on %d\n url: %s\n", fd, data->entry->url);
    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
	    data->entry->url,
	    data->entry->url,
	    "WAIS",
	    103,
	    "Read timeout",
	    "The Network/Remote site may be down. Try again later",
	    HARVEST_VERSION,
	    comm_hostname());
    storeAbort(data->entry, tmp_error_buf);
    comm_set_select_handler(fd, COMM_SELECT_READ, 0, 0);
    comm_close(fd);
    safe_free(data);
}

/* This will be called when socket lifetime is expired. */
void waisLifetimeExpire(fd, data)
int fd;
WAISData *data;
{
    debug(4, "waisLifeTimeExpire: FD %d: <URL:%s>\n", fd, data->entry->url);
    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
	    data->entry->url,
	    data->entry->url,
	    "WAIS",
	    110,
	    "Transaction Timeout",
      "The Network/Remote site may be down or too slow. Try again later",
	    HARVEST_VERSION,
	    comm_hostname());
    storeAbort(data->entry, tmp_error_buf);
    comm_set_select_handler(fd, COMM_SELECT_READ, 0, 0);
    comm_close(fd);
    safe_free(data);
}




/* This will be called when data is ready to be read from fd.  Read until
   error or connection closed. */
void waisReadReply(fd, data)
int fd;
WAISData *data;
{
    char buf[4096];
    int len = read(fd, buf, 4096);
    debug(5, "waisReadReply - fd: %d read len:%d\n", fd, len);

    if (len < 0) {
	debug(1, "waisReadReply - error reading errno %d: %s\n",
	      errno, sys_errlist[errno]);
	if (errno == ECONNRESET) {
	    /* Connection reset by peer */
	    /* consider it as a EOF */
	    data->entry->ttl = 0;

	    sprintf(tmp_error_buf, "\n<p>Warning: The Remote Server sent RESET at the end of transmission.\n");
	    storeAppend(data->entry, tmp_error_buf, strlen(tmp_error_buf));
	    storeComplete(data->entry);
	} else {
	    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		    data->entry->url,
		    data->entry->url,
		    "WAIS",
		    105,
		    "Read error",
		    "Network/Remote site is down. Try again later.",
		    HARVEST_VERSION,
		    comm_hostname());
	    storeAbort(data->entry, tmp_error_buf);
	}
	comm_close(fd);
	safe_free(data);
    } else if (len == 0) {
	/* Connection closed; retrieval done. */
	data->entry->ttl = 0;
	storeComplete(data->entry);
	comm_close(fd);
	safe_free(data);
    } else if ((data->entry->e_current_len + len) > getWAISMax()) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"WAIS",
		106,
		"Object is too large",
		"Contact your system manager for further help, or retrieve this object directly without a proxy.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);
    } else {
	storeAppend(data->entry, buf, len);
	comm_set_select_handler(fd, COMM_SELECT_READ, (PF) waisReadReply, (caddr_t) data);
	comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT, (PF) waisReadReplyTimeout,
				       (caddr_t) data, getReadTimeout());
    }
}

/* This will be called when request write is complete. Schedule read of
   reply. */
void waisSendComplete(fd, buf, size, errflag, data)
int fd;
char *buf;
int size;
int errflag;
WAISData *data;
{
    debug(5, "waisSendComplete - fd: %d size: %d errflag: %d\n",
	  fd, size, errflag);
    if (errflag) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"WAIS",
		101,
		"Cannot connect to the original site",
		"The remote site may be down.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);
    } else {
	/* Schedule read reply. */
	comm_set_select_handler(fd, COMM_SELECT_READ, (PF) waisReadReply, (caddr_t) data);
	comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT, (PF) waisReadReplyTimeout,
				       (caddr_t) data, getReadTimeout());

    }
    safe_free(buf);		/* Allocated by waisSendRequest. */
}

/* This will be called when connect completes. Write request. */
void waisSendRequest(fd, data)
int fd;
WAISData *data;
{
#define CR '\015'
#define LF '\012'
    int len = strlen(data->request) + 4;
    char *buf;

    debug(5, "waisSendRequest - fd: %d\n", fd);

    if (data->type)
	len += strlen(data->type);
    if (data->mime_hdr)
	len += strlen(data->mime_hdr);

    buf = (char *) xcalloc(1, len + 1);

    if (data->mime_hdr)
	sprintf(buf, "%s %s %s%c%c", data->type, data->request,
		data->mime_hdr, CR, LF);
    else
	sprintf(buf, "%s %s%c%c", data->type, data->request, CR, LF);
    debug(6, "waisSendRequest - buf:%s\n", buf);
    icpWrite(fd, buf, len, 30, waisSendComplete, data);
}

int waisStart(unusedfd, url, type, mime_hdr, entry)
int unusedfd;
char *url;
char *type;
char *mime_hdr;
StoreEntry *entry;
{
    /* Create state structure. */
    int sock, status;
    WAISData *data = (WAISData *) xmalloc(sizeof(WAISData));

    memset(data, '\0', sizeof(WAISData));
    data->entry = entry;

    debug(3, "waisStart - url:%s, type:%s\n", url, type);
    debug(4, "            header: %s\n", mime_hdr);

    /* Parse url. */
    (void) wais_url_parser(url, data->host, &data->port, data->request);
    data->type = type;
    data->mime_hdr = mime_hdr;

    /* Create socket. */
    sock = comm_open(COMM_NONBLOCKING, 0, 0);

    /* check if IP is already in cache. It must be. 
       It should be done before this route is called. 
       Otherwise, we cannot check return code for connect. */
    if (!ipcache_gethostbyname(data->host)) {
	debug(4, "waisstart: Called without IP entry in ipcache. OR lookup failed.\n");
	comm_close(sock);
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		entry->url,
		entry->url,
		"WAIS",
		102,
		"DNS name lookup failure",
		dns_error_message,
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(entry, tmp_error_buf);
	safe_free(data);
	return COMM_ERROR;
    }
    /* Open connection. */
    if ((status = comm_connect(sock, data->host, data->port))) {
	if (status != EINPROGRESS) {
	    comm_close(sock);
	    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		    entry->url,
		    entry->url,
		    "WAIS",
		    101,
		    "Cannot connect to the original site",
		    "The remote site may be down.",
		    HARVEST_VERSION,
		    comm_hostname());
	    storeAbort(entry, tmp_error_buf);
	    safe_free(data);
	    return COMM_ERROR;
	} else {
	    debug(5, "waisStart - conn %d EINPROGRESS\n", sock);
	}
    }
    /* Install connection complete handler. */
    comm_set_select_handler(sock, COMM_SELECT_LIFETIME,
			    (PF) waisLifetimeExpire, (caddr_t) data);
    comm_set_select_handler(sock, COMM_SELECT_WRITE,
			    (PF) waisSendRequest, (caddr_t) data);
    return COMM_OK;
}
#endif
